# broccoli

<b>B</b>eautiful <b>Ro</b>bot <b>C</b>++ <b>Co</b>de <b>Li</b>brary

[![pipeline status](https://gitlab.lrz.de/AMCode/broccoli/badges/develop/pipeline.svg)](https://gitlab.lrz.de/AMCode/broccoli/commits/develop)
[![coverage report](https://gitlab.lrz.de/AMCode/broccoli/badges/develop/coverage.svg)](https://gitlab.lrz.de/AMCode/broccoli/commits/develop)

You can find the API documentation of the latest release [here](http://cobra.amm.mw.tum.de/broccoli/). This link is only accessible via the AM network.

## Introduction

Broccoli is a generic C++ library for common algorithms used in robotics. Although originating from the source code of the Lola project it is not restricted to humanoid robots and may be used for any kind of robot.

## Dependencies

Broccoli requires at least cmake 3.8 and a C++11-compatible compiler. Broccoli is a header-only library and is written to be as generic as possible.
Therefore, parts of broccoli do not require any external dependencies.

### External Libraries

Some broccoli components rely on

* **Eigen** for linear algebra
* **SDL2** for device i/o
* **zlib** for log file compression

### Development Resources

The following resources are recommended for developers:

- **doxygen** for API documentation generation
- **llvm-cov/llvm-profdata** for coverage reports
- **clang-tidy** to analyze/fix the code

## CMake Configuration

The following options can be passed to CMake:

| Name | Description | Default Value |
|---|---|---|
| OPTION_BUILD_EXAMPLES | Build example applications | OFF |
| OPTION_BUILD_DOC | Enables documentation generation with doxygen | OFF |
| OPTION_BUILD_TESTS | Enables unit testing | OFF |
| OPTION_ENABLE_COVERAGE | Enables coverage detection for unit tests | ON |
| OPTION_CLANG_TIDY_FIX | Clang-tidy will always check the code if it is available on your system. If you enable this option, the code will be fixed automatically. | OFF |

## Targets

- **tests**: Builds the integrated unit tests if testing is enabled
- **coverage**: Creates coverage information on the unit tests. Coverage is output to console and a html report in the build directory.
- **doc**: (Re-)triggers generation of doxygen documentation, if OPTION_BUILD_DOCS is ON
- **update-googletest**: Updates the downloaded version of Google Test
- **update-eigen**: Updates the downloaded version of the Eigen library

## Using Broccoli in CMake Projects

Broccoli exports the target ```eat::broccoli```, which can be imported using ```find_package()```. By adding
the target via ```target_link_libraries``` to your user target, all include directories are set automatically:

```cmake
cmake_minimum_required (VERSION 3.8)
project (myproject)

find_package (broccoli 0.0.1 REQUIRED)

add_executable (example example.cpp)
target_link_libraries (example eat::broccoli)
```

You may also let broccoli pull in the dependencies to the external libraries via cmake components:
```cmake
find_package (broccoli 0.0.1 REQUIRED COMPONENTS eigen sdl zlib)
```
This way broccoli automatically searches for these libraries and adds include directories/linker settings to
the broccoli target. By adding eat::broccoli to your targets, these will automatically be configured for the external libraries.

**Overview on broccoli components:**

| Name | Description | Compiler Definition |
|---|---|---|
| eigen | Search for Eigen3 library and add include directories to target (header-only) | HAVE_EIGEN3
| sdl | Search for SDL2 library and add include directories / link libraries to target | HAVE_SDL2 |
| zlib | Search for zlib library and add include directories / link libraries to target | HAVE_ZLIB |

If you want to add external dependencies by yourself, be sure to add the corresponding compiler definitions to
activate corresponding code in broccoli.

### Broccoli as an External Project

Instead of installing broccoli to the system directories you may use CMake's ExternalProjects feature. This involves creating a file ```broccoli.CMakeLists.txt.in```, which is then used to pull and install broccoli during CMake configuration time.

Add the file ```broccoli.CMakeLists.txt.in``` to your ```cmake``` directory. You may specify any branch or release tag name to get a certain version of broccoli:
```cmake
cmake_minimum_required(VERSION 2.8.2)

project(broccoli-download NONE)
include(ExternalProject)

ExternalProject_Add(
    broccoli
    GIT_REPOSITORY      https://gitlab.lrz.de/AMCode/broccoli.git
    GIT_TAG             atlantic
    PREFIX              "${CMAKE_BINARY_DIR}/broccoli"
    INSTALL_DIR         "${CMAKE_BINARY_DIR}/broccoli"
    CMAKE_ARGS          "-DCMAKE_INSTALL_PREFIX=<INSTALL_DIR>"
    TEST_COMMAND        ""
    UPDATE_DISCONNECTED 1
    STEP_TARGETS        update
)
```

Then add some code to your ```CMakeLists.txt``` file to pull and install broccoli as external project:
```cmake
if (NOT broccoli_DIR)
    # configure and install broccoli (if not already done by super-project)
    configure_file(${CMAKE_SOURCE_DIR}/cmake/broccoli.CMakeLists.txt.in ${CMAKE_BINARY_DIR}/broccoli/download/CMakeLists.txt)
    execute_process(COMMAND ${CMAKE_COMMAND} -G "${CMAKE_GENERATOR}" .
        RESULT_VARIABLE result
        WORKING_DIRECTORY ${CMAKE_BINARY_DIR}/broccoli/download)
    if(result)
       message(FATAL_ERROR "CMake step for broccoli failed: ${result}")
    endif()
    execute_process(COMMAND ${CMAKE_COMMAND} --build .
        RESULT_VARIABLE result
        WORKING_DIRECTORY ${CMAKE_BINARY_DIR}/broccoli/download)
    if(result)
        message(FATAL_ERROR "Build step for broccoli failed: ${result}")
    endif()

    add_custom_target(update-broccoli
        COMMAND ${CMAKE_COMMAND} --build . --target broccoli-update
        COMMAND ${CMAKE_COMMAND} --build . --target all
        WORKING_DIRECTORY ${CMAKE_BINARY_DIR}/broccoli/download)

    set(broccoli_DIR ${CMAKE_BINARY_DIR}/broccoli)
endif()
```

This will install broccoli to your build directory on first cmake run. Furthermore, an additional target ```update-broccoli``` can be used to update the broccoli sources from the git repository.
To use broccoli in your CMake project just use find_package() and the imported target as usual.
