/*
 * This file is part of broccoli.
 * Copyright (C) 2019 Chair of Applied Mechanics, Technical University of Munich
 * https://www.amm.mw.tum.de/
 */

#pragma once // Load this file only once

// This module requires Eigen library
#ifdef HAVE_EIGEN3

#include <Eigen/Dense>
#include <assert.h>
#include <cmath>
#include <stdint.h>
#include <vector>

namespace broccoli {
namespace core {
    //! Extensions to the standard and Eigen's math library
    namespace math {
        //! Specification of result types.
        /*! This enum is used to given more detailed information about the result of an algorithm and which error/warning may have occured. */
        enum class Result : uint8_t {
            UNKNOWN = 0, //!< Unknown result (this should **never** be returned by a function)
            SUCCESS, //!< Algorithm was successful
            ERROR_DIMENSION_MISMATCH, //!< An **error** occured: dimensions do not match
            ERROR_OVERFLOW, //!< An **error** occured: overflow
            WARNING_CLOSE_TO_SINGULAR, //!< A **warning** occured: close to singularity
            RESULT_COUNT //!< Total count of elements
        };

        //! Returns string representation of the given result type
        static inline std::string resultString(const Result& result)
        {
            // Check result
            switch (result) {
            case Result::UNKNOWN:
                return "UNKNOWN";
            case Result::SUCCESS:
                return "SUCCESS";
            case Result::ERROR_DIMENSION_MISMATCH:
                return "ERROR_DIMENSION_MISMATCH";
            case Result::ERROR_OVERFLOW:
                return "ERROR_OVERFLOW";
            case Result::WARNING_CLOSE_TO_SINGULAR:
                return "WARNING_CLOSE_TO_SINGULAR";
            default:
                return "UNKNOWN";
            }
        }

        //! Factorial of a non-negative integer
        /*!
         * Computes the factorial of a natural integer as
         * \f[ n! = 1\cdot 2\cdot 3\cdot \dots = \prod_{k=1}^n k\f]
         * \param [in] integer The input variable (as unsigned integer)
         * \param [out] result Pointer to flag in which the result of the algorithm should be stored
         * \return The factorial of the input value
         */
        static inline uint64_t factorial(const uint64_t& integer, Result* const result = nullptr)
        {
            // Check for overflow
            if (result != nullptr) {
                if (integer > 20) {
                    *result = Result::ERROR_OVERFLOW;
                    return 0;
                } else
                    *result = Result::SUCCESS;
            }

            // Use look-up-table for speed-up
            if (integer == 0 || integer == 1)
                return 1ULL;
            else if (integer == 2)
                return 2ULL;
            else if (integer == 3)
                return 6ULL;
            else if (integer == 4)
                return 24ULL;
            else if (integer == 5)
                return 120ULL;
            else if (integer == 6)
                return 720ULL;
            else if (integer == 7)
                return 5040ULL;
            else if (integer == 8)
                return 40320ULL;
            else if (integer == 9)
                return 362880ULL;
            else if (integer == 10)
                return 3628800ULL;
            else
                return integer * factorial(integer - 1);
        }

        //! Compute coefficient tuples required by Faà di Bruno's formula
        /*!
         * Computes all tuples \f$ (m_1,\ m_2,\ \dots,\ m_n)\f$ of nonnegative integers satisfying \f$ 1\cdot m_1 + 2\cdot m_2 + \dots + n\cdot m_n = n\f$
         * with \f$ n = \f$ \p derivationOrder
         * \param [in] derivationOrder Order of required derivation (\f$n\f$)
         * \param [out] result Pointer to flag in which the result of the algorithm should be stored
         * \return Unsorted list of tuples \f$ (m_1,\ m_2,\ \dots,\ m_n)\f$
         */
        static inline std::vector<std::vector<uint64_t>> FaaDiBrunoTuples(const uint64_t& derivationOrder, Result* const result = nullptr)
        {
            // Initialize return value
            std::vector<std::vector<uint64_t>> tuples;

            // Check for overflow
            if (result != nullptr) {
                if (derivationOrder > 15) {
                    *result = Result::ERROR_OVERFLOW;
                    return tuples;
                } else
                    *result = Result::SUCCESS;
            }

            // Use look-up-table for speed-up
            if (derivationOrder == 1) {
                // Initialize tuples
                tuples.resize(1);
                for (size_t i = 0; i < tuples.size(); i++)
                    tuples[i] = std::vector<uint64_t>(derivationOrder, 0);

                // Set entries
                tuples[0][0] = 1;
            } else if (derivationOrder == 2) {
                // Initialize tuples
                tuples.resize(2);
                for (size_t i = 0; i < tuples.size(); i++)
                    tuples[i] = std::vector<uint64_t>(derivationOrder, 0);

                // Set entries
                tuples[0][0] = 2;
                tuples[1][1] = 1;
            } else if (derivationOrder == 3) {
                // Initialize tuples
                tuples.resize(3);
                for (size_t i = 0; i < tuples.size(); i++)
                    tuples[i] = std::vector<uint64_t>(derivationOrder, 0);

                // Set entries
                tuples[0][0] = 3;
                tuples[1][0] = 1;
                tuples[1][1] = 1;
                tuples[2][2] = 1;
            } else if (derivationOrder == 4) {
                // Initialize tuples
                tuples.resize(5);
                for (size_t i = 0; i < tuples.size(); i++)
                    tuples[i] = std::vector<uint64_t>(derivationOrder, 0);

                // Set entries
                tuples[0][0] = 4;
                tuples[1][0] = 2;
                tuples[1][1] = 1;
                tuples[2][1] = 2;
                tuples[3][0] = 1;
                tuples[3][2] = 1;
                tuples[4][3] = 1;
            } else if (derivationOrder == 5) {
                // Initialize tuples
                tuples.resize(7);
                for (size_t i = 0; i < tuples.size(); i++)
                    tuples[i] = std::vector<uint64_t>(derivationOrder, 0);

                // Set entries
                tuples[0][0] = 5;
                tuples[1][0] = 3;
                tuples[1][1] = 1;
                tuples[2][0] = 1;
                tuples[2][1] = 2;
                tuples[3][0] = 2;
                tuples[3][2] = 1;
                tuples[4][1] = 1;
                tuples[4][2] = 1;
                tuples[5][0] = 1;
                tuples[5][3] = 1;
                tuples[6][4] = 1;
            } else {
                // Count of all possible tuple candidates: (count of possible values)^(element count) = (n+1)^n
                size_t countOfValues = derivationOrder + 1;
                size_t tupleCandidateCount = countOfValues;
                for (size_t i = 1; i < derivationOrder; i++)
                    tupleCandidateCount *= countOfValues;
                tuples.reserve(tupleCandidateCount); // Allocate memory for maximum speed

                // Initialize tuple candidate
                std::vector<uint64_t> tupleCandidate;
                tupleCandidate.resize(derivationOrder);

                // Brute-force: compute all possible combinations with each m within [0, n] then check constraint
                for (size_t i = 0; i < tupleCandidateCount; i++) {
                    // Compute element values
                    /* Idea:
                     * -----
                     * Consider the candidate index "i" as number to the base of "n+1" (all possible values for the m's). Then we can obtain
                     * the single "digits" (m's) in the same way as we would to get every digit of a number in the dezimal system (base 10).
                     */
                    size_t number = i;
                    size_t base = derivationOrder + 1;
                    for (size_t j = 0; j < derivationOrder; j++) {
                        tupleCandidate[j] = number % base;
                        number /= base;
                    }

                    // Check constraint
                    uint64_t sum = 0;
                    for (size_t j = 1; j <= derivationOrder; j++)
                        sum += j * tupleCandidate[j - 1];
                    if (sum == derivationOrder) {
                        // Constraint is satisfied -> add candidate to tuple list
                        tuples.push_back(tupleCandidate);
                    }
                }
            }

            // Pass back computed tuples
            return tuples;
        }

        //! Solve tridiagonal system of equations.
        /*!
         * Solves a linear system of equations \f$ A \cdot X = B \f$ where \f$ A \f$ is tridiagonal by using the efficient "Thomas algorithm". The right hand side \f$ B\f$ (and thus also \f$ X \f$) may have more than one column.
         *
         * (*Algorithm according to Quarteroni, "Numerical Mathematics", Springer Berlin Heidelberg, 2007, 2nd edition, DOI:[10.1007/b98885](https://www.doi.org/10.1007/b98885), p.93ff*)
         *
         * \f[
         * \underbrace{\left[\begin{array}{ccccccc} d_1 & u_1 & & & & & 0 \\ l_2 & d_2 & u_2 & & & & \\ & \ddots & \ddots & \ddots & & & \\ & & l_i & d_i & u_i & & \\ & & & \ddots & \ddots & \ddots & \\ & & & & l_{n-1} & d_{n-1} & u_{n-1} \\ 0 & & & & & l_n & d_n \end{array}\right]}_A
         * \cdot
         * \underbrace{\left[\begin{array}{c} x_1 \\ x_2 \\ \vdots \\ x_i \\ \vdots \\ x_{n-1} \\ x_n \end{array}\right]}_X
         * =
         * \underbrace{\left[\begin{array}{c} b_1 \\ b_2 \\ \vdots \\ b_i \\ \vdots \\ b_{n-1} \\ b_n \end{array}\right]}_B
         * \f]
         *
         * \param [in] lowerDiagonalOfA Lower diagonal elements \f$ l_i \f$ of matrix \f$ A \f$ (\f$ n \f$ elements with the first element beeing zero)
         * \param [in] diagonalOfA Diagonal elements \f$ d_i \f$ of matrix \f$ A \f$ (\f$ n \f$ elements)
         * \param [in] upperDiagonalOfA Upper diagonal elements \f$ u_i \f$ of matrix \f$ A \f$ (\f$ n \f$ elements with the last element beeing zero)
         * \param [in] B Right hand side \f$ B \in \mathbb{R}^{n \times m}\f$
         * \param [out] X Solution \f$ X \in \mathbb{R}^{n \times m}\f$
         * \param [out] result Pointer to flag in which the result of the algorithm should be stored
         * \return `true` on success, `false` otherwise
         */
        static inline bool solveTriDiagonalSystemOfEquations(const Eigen::VectorXd& lowerDiagonalOfA, const Eigen::VectorXd& diagonalOfA, const Eigen::VectorXd& upperDiagonalOfA, const Eigen::MatrixXd& B, Eigen::MatrixXd& X, Result* const result = nullptr)
        {
            // Assume success
            if (result != nullptr)
                *result = Result::SUCCESS;

            // Check dimensions
            int n = diagonalOfA.rows();
            int m = B.cols();
            if (lowerDiagonalOfA.rows() != n || upperDiagonalOfA.rows() != n || B.rows() != n) {
                if (result != nullptr)
                    *result = Result::ERROR_DIMENSION_MISMATCH;
                return false;
            }

            // Resize solution
            X.resize(n, m);

            // Forward sweep
            Eigen::VectorXd H(n - 1); // Modified upper diagonal of A (temporary helper)
            Eigen::MatrixXd P(n, m); // Modified right hand side B (temporary helper)
            double denominator = diagonalOfA(0);
            double absoluteDenominator = fabs(denominator); // Absolute value of denominator
            double minimumDenominator = absoluteDenominator; // Minimum (absolute) value of denominator
            double maximumDenominator = absoluteDenominator; // Maximum (absolute) value of denominator
            if (n > 1)
                H(0) = upperDiagonalOfA(0) / denominator;
            P.block(0, 0, 1, m) = B.block(0, 0, 1, m) / denominator;
            for (int i = 1; i < n; i++) {
                denominator = (diagonalOfA(i) - lowerDiagonalOfA(i) * H(i - 1));
                absoluteDenominator = fabs(denominator);
                if (absoluteDenominator < minimumDenominator)
                    minimumDenominator = absoluteDenominator;
                if (absoluteDenominator > maximumDenominator)
                    maximumDenominator = absoluteDenominator;
                if (i < n - 1)
                    H(i) = upperDiagonalOfA(i) / denominator;
                P.block(i, 0, 1, m) = (B.block(i, 0, 1, m) - lowerDiagonalOfA(i) * P.block(i - 1, 0, 1, m)) / denominator;
            }

            // Check for singularity (use ratio between minimum and maximum absolute value of denominator as estimate for the condition number)
            if (maximumDenominator == 0 || minimumDenominator / maximumDenominator < 1e-15 /* <-- typical machine precision for 64bit-double */)
                // Send warning to user
                if (result != nullptr)
                    *result = Result::WARNING_CLOSE_TO_SINGULAR;

            // Backwards-substitution
            X.block(n - 1, 0, 1, m) = P.block(n - 1, 0, 1, m);
            for (int i = n - 2; i >= 0; i--)
                X.block(i, 0, 1, m) = P.block(i, 0, 1, m) - H(i) * X.block(i + 1, 0, 1, m);

            // Success
            return true;
        }

        //! Solve block-tridiagonal system of equations.
        /*!
         * Solves a linear system of equations \f$ A \cdot X = B \f$ where \f$ A \f$ is block-tridiagonal (square blocks). The template parameter \p BlockTypeA specifies the type of the blocks in \f$A\f$. The template parameter \p BlockTypeXB
         * specifies the type of the blocks in \f$ X \f$ and \f$ B \f$. Both template parameters \p BlockTypeA and \p BlockTypeXB have to be of type `Eigen::Matrix`.
         *
         * (*Algorithm according to Mund, "An algorithm for the interpolation of functions using quintic splines", Journal of Computational and Applied Mathematics, 1975, Vol. 1, Nr. 4, p.279-288, DOI:[10.1016/0771-050X(75)90020-0](https://www.doi.org/10.1016/0771-050X(75)90020-0)*)
         *
         * \f[
         * \underbrace{\left[\begin{array}{ccccccc} D_1 & U_1 & & & & & 0 \\ L_2 & D_2 & U_2 & & & & \\ & \ddots & \ddots & \ddots & & & \\ & & L_i & D_i & U_i & & \\ & & & \ddots & \ddots & \ddots & \\ & & & & L_{n-1} & D_{n-1} & U_{n-1} \\ 0 & & & & & L_n & D_n \end{array}\right]}_A
         * \cdot
         * \underbrace{\left[\begin{array}{c} X_1 \\ X_2 \\ \vdots \\ X_i \\ \vdots \\ X_{n-1} \\ X_n \end{array}\right]}_X
         * =
         * \underbrace{\left[\begin{array}{c} B_1 \\ B_2 \\ \vdots \\ B_i \\ \vdots \\ B_{n-1} \\ B_n \end{array}\right]}_B
         * \f]
         *
         * \warning For numerical stability the matrix \f$ A \f$ should (but does not have to) be block-diagonally dominant, i.e. \f$ ||(D_i)^{-1}|| \cdot (||L_i|| + ||U_i||) <= 1 \f$ for all \f$ i = 1,\ \dots, n \f$
         *
         * \param [in] lowerDiagonalOfA Lower diagonal blocks \f$ L_i \f$ of matrix \f$ A \f$ (\f$ n \f$ elements with the first element as zero-matrix)
         * \param [in] diagonalOfA Diagonal blocks \f$ D_i \f$ of matrix \f$ A \f$ (\f$ n \f$ elements)
         * \param [in] upperDiagonalOfA Upper diagonal blocks \f$ U_i \f$ of matrix \f$A\f$ (\f$ n \f$ elements with the last element as zero-matrix)
         * \param [in] B Right hand side (\f$ n \f$ blocks of type \p BlockTypeXB)
         * \param [out] X Solution (\f$ n \f$ blocks of type \p BlockTypeXB)
         * \param [out] result Pointer to flag in which the result of the algorithm should be stored
         * \return `true` on success, `false` otherwise
         */
        template <class BlockTypeA, class BlockTypeXB>
        static inline bool solveBlockTriDiagonalSystemOfEquations(const std::vector<BlockTypeA, Eigen::aligned_allocator<BlockTypeA>>& lowerDiagonalOfA, const std::vector<BlockTypeA, Eigen::aligned_allocator<BlockTypeA>>& diagonalOfA, const std::vector<BlockTypeA, Eigen::aligned_allocator<BlockTypeA>>& upperDiagonalOfA, const std::vector<BlockTypeXB, Eigen::aligned_allocator<BlockTypeXB>>& B, std::vector<BlockTypeXB, Eigen::aligned_allocator<BlockTypeXB>>& X, Result* const result = nullptr)
        {
            // Assume success
            if (result != nullptr)
                *result = Result::SUCCESS;

            // Check dimensions
            size_t n = diagonalOfA.size();
            if (lowerDiagonalOfA.size() != n || upperDiagonalOfA.size() != n || B.size() != n) {
                if (result != nullptr)
                    *result = Result::ERROR_DIMENSION_MISMATCH;
                return false;
            }

            // Resize solution
            X.resize(n);

            // Forward elimination
            std::vector<BlockTypeA, Eigen::aligned_allocator<BlockTypeA>> H; // Modified upper diagonal of A (temporary helper)
            H.resize(n - 1);
            std::vector<BlockTypeXB, Eigen::aligned_allocator<BlockTypeXB>> P; // Modified right hand side B (temporary helper)
            P.resize(n);
            Eigen::ColPivHouseholderQR<BlockTypeA> QR(diagonalOfA[0]); // Compute QR decomposition of D1
            double absoluteDeterminant = QR.absDeterminant(); // The absolute value of the determinant
            double minimumDeterminant = absoluteDeterminant; // The minimum observed value for absoluteDeterminant
            double maximumDeterminant = absoluteDeterminant; // The maximum observed value for absoluteDeterminant
            if (n > 1)
                H[0] = QR.solve(upperDiagonalOfA[0]);
            P[0] = QR.solve(B[0]);
            for (size_t i = 1; i < n; i++) {
                QR.compute(diagonalOfA[i] - lowerDiagonalOfA[i] * H[i - 1]);
                absoluteDeterminant = QR.absDeterminant();
                if (absoluteDeterminant < minimumDeterminant)
                    minimumDeterminant = absoluteDeterminant;
                if (absoluteDeterminant > maximumDeterminant)
                    maximumDeterminant = absoluteDeterminant;
                if (i < n - 1)
                    H[i] = QR.solve(upperDiagonalOfA[i]);
                P[i] = QR.solve(B[i] - lowerDiagonalOfA[i] * P[i - 1]);
            }

            // Check for singularity (use ratio between minimum and maximum absolute value of block-determinant as estimate for the condition number)
            if (maximumDeterminant == 0 || minimumDeterminant / maximumDeterminant < 1e-15 /* <-- typical machine precision for 64bit-double */)
                // Send warning to user
                if (result != nullptr)
                    *result = Result::WARNING_CLOSE_TO_SINGULAR;

            // Backwards-substitution
            X[n - 1] = P[n - 1];
            for (int i = n - 2; i >= 0; i--)
                X[i] = P[i] - H[i] * X[i + 1];

            // Success
            return true;
        }

        //! Compute natural logarithm of an arbitrary quaternion
        /*!
         * (*Extension of Eigen's functionality for quaternions*)
         *
         * \attention The norms of the quaternion and of the 3 components (x, y, z) have to be non-zero.
         *
         * \param [in] quaternion The quaternion to compute the natural logarithm for
         * \param [out] result Pointer to flag in which the result of the algorithm should be stored
         * \return The natural logarithm of the given quaternion as quaternion by itself.
         */
        static inline Eigen::Quaterniond quaternionNaturalLogarithm(const Eigen::Quaterniond& quaternion, Result* const result = nullptr)
        {
            // Assume success
            if (result != nullptr)
                *result = Result::SUCCESS;

            /*! Explanation:
              * ------------
              * Given a quaternion \f$q = \left[w,\ x,\ y,\ z\right]^T = \left[w,\ v\right]^T\f$ with \f$v = \left[x,\ y,\ z\right]^T\f$ the natural
              * logarithm of \f$q\f$ is given as
              * \f[ \ln(q) = \left[\ln(||q||),\ \frac{v}{||v||} * \arccos(\frac{w}{||q||})\right]^T\f]
             */

            // Initialize return value
            Eigen::Quaterniond naturalLogarithm(1, 0, 0, 0);

            // Compute norm of q
            double normq = quaternion.norm();

            // Compute norm of vector part
            double normv = sqrt(quaternion.x() * quaternion.x() + quaternion.y() * quaternion.y() + quaternion.z() * quaternion.z());

            // Avoid division by zero
            if (normq != 0 && normv != 0) {
                // Insert to formula
                naturalLogarithm.w() = log(normq);
                naturalLogarithm.vec() = quaternion.vec() / normv * acos(quaternion.w() / normq);
            } else {
                // Avoid division by zero -> send warning
                if (result != nullptr)
                    *result = Result::WARNING_CLOSE_TO_SINGULAR;
            }

            // Pass back result
            return naturalLogarithm;
        }

        //! Implements the Euclidean algorithm to find the greatest common divisor of two integer numbers.
        /*!
         * This follows the following recursive computation law for numbers \f$a, b \ge 0\f$
         * \f[gcd(a, a) = a\f]
         * \f[gcd(a, b) = gcd(a - b, b) \; \text{if} a > b\f]
         * \f[gcd(a, b) = gcd(a, b - a) \; \text{if} b > a\f]
         *
         * \tparam T An integer type
         * \param a A value >= 0
         * \param b A value >= 0
         * \return The greatest common divisor of a and b
         */
        template <typename T>
        static inline typename std::enable_if<std::numeric_limits<T>::is_integer, T>::type greatestCommonDivisor(T a, T b)
        {
            assert(a >= 0);
            assert(b >= 0);

            if (a == 0)
                return b;

            if (b == 0)
                return a;

            // The Euclidean algorithm is recursive
            // We know we have the result when the difference between a and b is zero
            while (true) {
                if (a == b) {
                    return a;
                }

                if (a > b) {
                    a = a - b;
                } else {
                    b = b - a;
                }
            }
        }

        //! Compute the least common multiple of two integer numbers
        /*!
         * \tparam T An integer type
         * \param a First value
         * \param b Second value
         * \return The least common multiple of a and b.
         */
        template <typename T>
        static inline typename std::enable_if<std::numeric_limits<T>::is_integer, T>::type leastCommonMultiple(const T& a, const T& b)
        {
            if (a == 0 && a == b)
                return 0;

            return (std::abs<T>(a) / greatestCommonDivisor(a, b)) * std::abs<T>(b);
        }

    } // namespace math
} // namespace core
} // namespace broccoli

#endif // HAVE_EIGEN3
