/*
 * This file is part of broccoli.
 * Copyright (C) 2019 Chair of Applied Mechanics, Technical University of Munich
 * https://www.amm.mw.tum.de/
 */

#pragma once // Load this file only once

#include "../io/encoding.hpp"
#include <stdint.h>
#include <string>

namespace broccoli {
namespace curve {
    //! Specification of curve interpolation methods
    enum class CurveInterpolationMethod : uint8_t {
        POLYNOMIAL_CONSTANT = 0, //!< Set coefficients to fit a polynomial of order zero (constant) with given function value.
        POLYNOMIAL_LINEAR, //!< Set coefficients to fit a polynomial of order one (linear) with given function values at \f$ x=0 \f$ and \f$ x=1 \f$.
        POLYNOMIAL_CUBIC_FIRST_DERIVATIVES, //!< Set coefficients to fit a polynomial of order three (cubic) with given function values and first order derivatives at \f$ x=0 \f$ and \f$ x=1 \f$.
        POLYNOMIAL_CUBIC_SECOND_DERIVATIVES, //!< Set coefficients to fit a polynomial of order three (cubic) with given function values and second order derivatives at \f$ x=0 \f$ and \f$ x=1 \f$.
        POLYNOMIAL_QUINTIC_FIRST_AND_SECOND_DERIVATIVES, //!< Set coefficients to fit a polynomial of order five (quintic) with given function values, first and second order derivatives at \f$ x=0 \f$ and \f$ x=1 \f$.
        CURVEINTERPOLATIONMETHOD_COUNT //!< Total count of elements
    };

    //! Returns the string representation of the given curve interpolation method
    static inline std::string curveInterpolationMethodString(const CurveInterpolationMethod& method)
    {
        // Check method
        switch (method) {
        case CurveInterpolationMethod::POLYNOMIAL_CONSTANT:
            return "POLYNOMIAL_CONSTANT";
        case CurveInterpolationMethod::POLYNOMIAL_LINEAR:
            return "POLYNOMIAL_LINEAR";
        case CurveInterpolationMethod::POLYNOMIAL_CUBIC_FIRST_DERIVATIVES:
            return "POLYNOMIAL_CUBIC_FIRST_DERIVATIVES";
        case CurveInterpolationMethod::POLYNOMIAL_CUBIC_SECOND_DERIVATIVES:
            return "POLYNOMIAL_CUBIC_SECOND_DERIVATIVES";
        case CurveInterpolationMethod::POLYNOMIAL_QUINTIC_FIRST_AND_SECOND_DERIVATIVES:
            return "POLYNOMIAL_QUINTIC_FIRST_AND_SECOND_DERIVATIVES";
        default:
            return "UNKNOWN";
        }
    }

    //! Specification of result types of curve algorithms
    enum class CurveResult : uint8_t {
        UNKNOWN = 0, //!< Unknown result (this should **never** happen)
        SUCCESS, //!< Algorithm was successful
        ERROR_INVALID_PARAMETERS, //!< An **error** occured: invalid input parameters
        ERROR_DEGREE_TOO_LOW, //!< An **error** occured: the degree of the underlying polynomial is too low
        ERROR_NOTIMPLEMENTED, //!< An **error** occured: the desired algorithm is **not** yet implemented
        CURVERESULT_COUNT //!< Total count of elements
    };

    //! Returns the string representation of the given result type for curve algorithms
    static inline std::string curveResultString(const CurveResult& result)
    {
        // Check result
        switch (result) {
        case CurveResult::UNKNOWN:
            return "UNKNOWN";
        case CurveResult::SUCCESS:
            return "SUCCESS";
        case CurveResult::ERROR_INVALID_PARAMETERS:
            return "ERROR_INVALID_PARAMETERS";
        case CurveResult::ERROR_DEGREE_TOO_LOW:
            return "ERROR_DEGREE_TOO_LOW";
        case CurveResult::ERROR_NOTIMPLEMENTED:
            return "ERROR_NOTIMPLEMENTED";
        default:
            return "UNKNOWN";
        }
    }

    //! Abstract representation of a curve (underlying function is not specified yet)
    /*!
     * Class acts as a prototype and declares all functions and members a curve should provide.
     *
     * The underlying function is given with respect to `position` which is **not** the time in general.
     * Instead one should parametrize the function in a way, such that `position` \f$\in [0, 1]\f$ relates to
     * the area of interest. This implies, that one has to be careful when using the \ref evaluate
     * function to get the time derivative (use chain rule!).
     */
    class Curve {
    public:
        //! Specification of possible underlying function types
        enum class FunctionType : uint8_t {
            POLYNOMIAL = 0, //!< Polynomial functions (\f$a\,x^2 + b\,x + c = \dots\f$)
            TRIGONOMETRIC, //!< Trigonometric functions (\f$\sin(\dots), \cos(\dots), \dots\f$)
            EXPONENTIAL //!< Exponential functions (\f$e^{\dots}\f$)
        };

        //! Returns the string representation of the given function type
        static std::string functionTypeString(const FunctionType& type)
        {
            // Check type
            switch (type) {
            case FunctionType::POLYNOMIAL:
                return "POLYNOMIAL";
            case FunctionType::TRIGONOMETRIC:
                return "TRIGONOMETRIC";
            case FunctionType::EXPONENTIAL:
                return "EXPONENTIAL";
            default:
                return "UNKNOWN";
            }
        }

        //! Constructor
        Curve()
        {
        }

        //! Destructor
        virtual ~Curve()
        {
        }

        //! Comparison operator: **equality**
        virtual bool operator==(const Curve& reference) const
        {
            (void)reference; // Unused

            // No own members -> always equal
            return true;
        }

        //! Comparison operator: **inequality**
        virtual bool operator!=(const Curve& reference) const { return !(*this == reference); }

        //! Evaluation of (derivative of) underlying function
        /*!
         * \warning The derivative is given with respect to the \p position (interpolation parameter), **not** time in general!
         *
         * \param [in] position Position to evaluate (derivative) at (should - but has not to be - within [0, 1])
         * \param [in] derivationOrder Order of the derivation (use 0 for base function, i.e. no derivation)
         * \return Value (of derivative) at specified position
         */
        virtual double evaluate(const double& position, const unsigned int& derivationOrder = 0) const = 0;

        //! Interpolate coefficients of underlying function to fit specific parameters (through-points, derivatives, ...)
        /*!
         * A detailed description on how to set the \p parameters is given below at the corresponding section.
         * \param [in] parameters Interpolation parameters (meaning varies depending on chosen method - for details see corresponding sections below)
         * \param [in] method Interpolation method to be used (has to be compatible with curve type)
         * \param [out] result Pointer to flag in which the result of the algorithm should be stored
         * \return `true` on success, `false` otherwise
         */
        virtual bool interpolate(const std::vector<double>& parameters, const CurveInterpolationMethod& method, CurveResult* const result = nullptr)
        {
            (void)parameters; // Unused
            (void)method; //Unused

            // Virtual function -> has to be overwritten by derived class -> base class method is not implemented!
            if (result != nullptr)
                *result = CurveResult::ERROR_NOTIMPLEMENTED;
            return false;
        }

        // Helpers
        // -------
        //! Returns degree of polynomial if underlying function is of type POLYNOMIAL and 0 otherwise
        static unsigned int degree() { return 0; }

        // Encoding
        // --------
        //! Encode member data as XML element and add it to the specified stream
        /*!
         * \param [in,out] stream The stream to which the member data should be appended to.
         * \param [in] XMLIndentationLevel Level of indentation in XML format (used to add spaces).
         * \param [in] XMLTabWhiteSpaces Count of whitespaces corresponding to a tab.
         * \param [in] numericFormat C-format specifier for numeric values.
         * \return Count of appended characters (elements in stream).
         */
        virtual io::encoding::CharacterStreamSize encodeToXML(io::encoding::CharacterStream& stream, const size_t& XMLIndentationLevel, const size_t& XMLTabWhiteSpaces, const std::string& numericFormat = "%.7g") const = 0;
    };
} // namespace curve
} // namespace broccoli
