/*
 * This file is part of broccoli.
 * Copyright (C) 2019 Chair of Applied Mechanics, Technical University of Munich
 * https://www.amm.mw.tum.de/
 */

#pragma once // Load this file only once

// This module requires Eigen library
#ifdef HAVE_EIGEN3

#include "../core/math.hpp"
#include "QuaternionCurve.hpp"
#include <Eigen/Dense>
#include <array>
#include <math.h>

namespace broccoli {
namespace curve {
    //! Class abstracting quaternion cuves using <b>S</b>pherical <b>L</b>inear int<b>ERP</b>olation (SLERP)
    /*! \warning This class is restricted to **unit-quaternions**! */
    class QuaternionSLERPCurve : public QuaternionCurve {
    public:
        //! Specialized constructor
        /*!
         * \param [in] beginQuaternion The quaternion at the beginning of the curve.
         * \param [in] endQuaternion The quaternion at the end of the curve.
         */
        QuaternionSLERPCurve(const Eigen::Quaterniond& beginQuaternion, const Eigen::Quaterniond& endQuaternion)
            : m_controlPoints{ { beginQuaternion, endQuaternion } }
        {
        }

        //! Default constructor
        /*! Initializes as \f$ q_0 = q_1 = [w=1,\, x=0,\, y=0,\, z=0]^T \f$ */
        QuaternionSLERPCurve()
            : QuaternionSLERPCurve(Eigen::Quaterniond(1, 0, 0, 0), Eigen::Quaterniond(1, 0, 0, 0))
        {
        }

        //! Destructor
        virtual ~QuaternionSLERPCurve()
        {
        }

        //! Comparison operator: **equality**
        virtual bool operator==(const QuaternionSLERPCurve& reference) const
        {
            // Compare base class
            if (QuaternionCurve::operator!=(reference))
                return false;

            // Compare control points
            for (size_t i = 0; i < m_controlPoints.size(); i++)
                if (fabs(2.0 * acos(m_controlPoints[i].dot(reference.m_controlPoints[i]))) > 1e-6 /* <- tolerance in [rad] */)
                    return false;

            // Otherwise -> equality
            return true;
        }

        //! Comparison operator: **inequality**
        virtual bool operator!=(const QuaternionSLERPCurve& reference) const { return !(*this == reference); }

        //! Returns type of underlying curve interpolation
        static InterpolationType interpolationType() { return InterpolationType::SLERP; }

        // Member data
        std::array<Eigen::Quaterniond, 2> m_controlPoints; //!< Quaternion control points ([0]=begin-quaternion, [1]=end-quaternion)

        // Evaluation of (derivative of) underlying curve (see base class for details)
        virtual Eigen::Quaterniond evaluate(const double& position, const unsigned int& derivationOrder = 0) const
        {
            // Initialize return value
            Eigen::Quaterniond returnValue(1, 0, 0, 0);

            // Project interpolation parameter to boundaries
            double projectedPosition = position;
            if (projectedPosition < 0)
                projectedPosition = 0;
            if (projectedPosition > 1)
                projectedPosition = 1;

            // Check if we only want the base function (for speed up of computation)
            if (derivationOrder == 0) {
                /*! Evaluation of base function
                 *  ---------------------------
                 * Performs simple quaternion slerp:
                 * \f[ q(x) = SLERP(q_0,\,q_1,\,x) = q_0 \cdot (q_0^{-1} \cdot q_1)^x \f]
                 * with \f$ x \in [0,\,1] \f$
                 */
                returnValue = m_controlPoints[0].slerp(projectedPosition, m_controlPoints[1]);
                returnValue.normalize(); // Normalize resulting quaternion to be sure to have a unit quaternion
            } else {
                /*! Evaluation of derivatives
                 *  -------------------------
                 * Compute derivative with respect to interpolation parameter \f$ x \f$ (**not** time in general).
                 *
                 * SLERP:
                 * \f[ q(x) = q_0 \cdot (q_0^{-1} \cdot q_1)^x = q_0 \cdot q_2^x \f]
                 * with \f$ q_2 := q_0^{-1} \cdot q_1 = const. \f$
                 * \f[ \frac{\partial q(x)}{\partial x} = \underbrace{q_0 \cdot q_2^x}_{q(x)} \cdot \ln(q_2) = q(x) \cdot \ln(q_2) = q(x) \cdot q_3 \f]
                 * with \f$ q_3 := \ln(q_2) = \ln(q_0^{-1} \cdot q_1) = const. \f$
                 * \f[ \frac{\partial^2 q(x)}{\partial x^2} = \frac{\partial q(x)}{\partial x} \cdot q_3 = q(x) \cdot q_3 \cdot q_3 = q(x) \cdot q_3^2 \f]
                 * \f[ \vdots \f]
                 * \f[ \frac{\partial^n q(x)}{\partial x^n} = q(x) \cdot q_3^n \f]
                 */

                // Compute helper variables (note: for unit quaternions the inverse is equal to the conjugate (conjugate is easier to compute))
                Eigen::Quaterniond q0invq1 = m_controlPoints[0].conjugate() * m_controlPoints[1];
                if (q0invq1.x() == 0 && q0invq1.y() == 0 && q0invq1.z() == 0) {
                    // ...start- and end-quaternion describe the same orientation -> derivative of q is zero (no change)!
                    return Eigen::Quaterniond(0, 0, 0, 0);
                }
                Eigen::Quaterniond q3 = core::math::quaternionNaturalLogarithm(q0invq1);

                // Compute final derivative
                returnValue = m_controlPoints[0].slerp(projectedPosition, m_controlPoints[1]);
                returnValue.normalize(); // Normalize resulting quaternion to be sure to have a unit quaternion
                // (no normalization after this point, as the quaternion is not a unit-quaternion anymore!)
                for (unsigned int i = 1; i <= derivationOrder; i++)
                    returnValue = returnValue * q3;
            }

            // Pass back return value
            return returnValue;
        }

        // Encoding
        // --------
        // Encode member data as XML element and add it to the specified stream (see base class for details)
        virtual io::encoding::CharacterStreamSize encodeToXML(io::encoding::CharacterStream& stream, const size_t& XMLIndentationLevel, const size_t& XMLTabWhiteSpaces, const std::string& numericFormat = "%.7g") const
        {
            io::encoding::CharacterStreamSize addedElements = 0;

            // Start XML element
            for (size_t i = 0; i < XMLIndentationLevel * XMLTabWhiteSpaces; i++)
                addedElements += io::encoding::encode(stream, " ");
            addedElements += io::encoding::encode(stream, "<QuaternionSLERPCurve");

            // Write Control point data
            addedElements += io::encoding::encode(stream, " ControlPoints=\"");
            for (size_t i = 0; i < m_controlPoints.size(); i++) {
                if (i > 0)
                    addedElements += io::encoding::encode(stream, ",");
                addedElements += io::encoding::encode(stream, (double)m_controlPoints[i].w(), numericFormat);
                addedElements += io::encoding::encode(stream, ",");
                addedElements += io::encoding::encode(stream, (double)m_controlPoints[i].x(), numericFormat);
                addedElements += io::encoding::encode(stream, ",");
                addedElements += io::encoding::encode(stream, (double)m_controlPoints[i].y(), numericFormat);
                addedElements += io::encoding::encode(stream, ",");
                addedElements += io::encoding::encode(stream, (double)m_controlPoints[i].z(), numericFormat);
            }
            addedElements += io::encoding::encode(stream, "\">");

            // End XML element
            addedElements += io::encoding::encode(stream, "</QuaternionSLERPCurve>\n");

            // Pass back added elements in stream
            return addedElements;
        }

    public:
        EIGEN_MAKE_ALIGNED_OPERATOR_NEW // <-- Proper 128 bit alignment of member data necessary for Eigen vectorization
    };
} // namespace curve
} // namespace broccoli

#endif // HAVE_EIGEN3
