/*
 * This file is part of broccoli.
 * Copyright (C) 2019 Chair of Applied Mechanics, Technical University of Munich
 * https://www.amm.mw.tum.de/
 */

#pragma once // Load this file only once

// This module requires Eigen library
#ifdef HAVE_EIGEN3

#include "../core/math.hpp"
#include "Curve.hpp"
#include "Exponential.hpp"
#include "Polynomial.hpp"
#include "Trigonometric.hpp"
#include <Eigen/Dense>
#include <array>
#include <vector>

namespace broccoli {
namespace curve {
    // Configuration
    const double minimumSplineSegmentProportion = 1e-6; //!< Minimum proportion of spline segments (used to avoid numeric errors)

    //! Specification of spline interpolation methods
    enum class SplineInterpolationMethod : uint8_t {
        POLYNOMIAL_PIECEWISE_CONSTANT = 0, //!< **Piecewise** interpolation with polynomials of **order 0** (piecewise constant) through given points
        POLYNOMIAL_PIECEWISE_LINEAR, //!< **Piecewise** interpolation with polynomials of **order 1** (piecewise linear) through given points
        POLYNOMIAL_PIECEWISE_CUBIC_FIRST_DERIVATIVES, //!< **Piecewise** interpolation with polynomials of **order 3** (piecewise cubic) through given points and with given first order derivative at control points
        POLYNOMIAL_PIECEWISE_CUBIC_SECOND_DERIVATIVES, //!< **Piecewise** interpolation with polynomials of **order 3** (piecewise cubic) through given points and with given second order derivative at control points
        POLYNOMIAL_PIECEWISE_QUINTIC_FIRST_AND_SECOND_DERIVATIVES, //!< **Piecewise** interpolation with polynomials of **order 5** (piecewise quintic) through given points and with given first and second order derivatives at control points
        POLYNOMIAL_SMOOTH_CUBIC_SECOND_DERIVATIVES, //!< **Smooth** interpolation (spline) with polynomials of **order 3** (piecewise cubic) through given points, \f$ C^2 \f$ continuous and with given second order derivative at start and end of spline
        POLYNOMIAL_SMOOTH_QUINTIC_FIRST_AND_SECOND_DERIVATIVES, //!< **Smooth** interpolation (spline) with polynomials of **order 5** (piecewise quintic) through given points, \f$ C^4 \f$ continuous and with given first and second order derivative at start and end of spline
        SPLINEINTERPOLATIONMETHOD_COUNT //!< Total count of elements
    };

    //! Returns the string representation of the given spline interpolation method
    static inline std::string splineInterpolationMethodString(const SplineInterpolationMethod& method)
    {
        // Check method
        switch (method) {
        case SplineInterpolationMethod::POLYNOMIAL_PIECEWISE_CONSTANT:
            return "POLYNOMIAL_PIECEWISE_CONSTANT";
        case SplineInterpolationMethod::POLYNOMIAL_PIECEWISE_LINEAR:
            return "POLYNOMIAL_PIECEWISE_LINEAR";
        case SplineInterpolationMethod::POLYNOMIAL_PIECEWISE_CUBIC_FIRST_DERIVATIVES:
            return "POLYNOMIAL_PIECEWISE_CUBIC_FIRST_DERIVATIVES";
        case SplineInterpolationMethod::POLYNOMIAL_PIECEWISE_CUBIC_SECOND_DERIVATIVES:
            return "POLYNOMIAL_PIECEWISE_CUBIC_SECOND_DERIVATIVES";
        case SplineInterpolationMethod::POLYNOMIAL_PIECEWISE_QUINTIC_FIRST_AND_SECOND_DERIVATIVES:
            return "POLYNOMIAL_PIECEWISE_QUINTIC_FIRST_AND_SECOND_DERIVATIVES";
        case SplineInterpolationMethod::POLYNOMIAL_SMOOTH_CUBIC_SECOND_DERIVATIVES:
            return "POLYNOMIAL_SMOOTH_CUBIC_SECOND_DERIVATIVES";
        case SplineInterpolationMethod::POLYNOMIAL_SMOOTH_QUINTIC_FIRST_AND_SECOND_DERIVATIVES:
            return "POLYNOMIAL_SMOOTH_QUINTIC_FIRST_AND_SECOND_DERIVATIVES";
        default:
            return "UNKNOWN";
        }
    }

    //! Specification of result types of spline algorithms
    enum class SplineResult : uint8_t {
        UNKNOWN = 0, //!< Unknown result (this should **never** happen)
        SUCCESS, //!< Algorithm was successful
        ERROR_INVALID_SPLINE_NO_SEGMENTS, //!< An **error** occured: Spline has to contain at least one segment
        ERROR_INVALID_SPLINE_PROPORTION_COUNT, //!< An **error** occured: Proportion count does **not** match segment count
        ERROR_INVALID_SPLINE_PROPORTION_NEGATIVE, //!< An **error** occured: Negative proportion found
        ERROR_INVALID_SPLINE_PROPORTION_SUM, //!< An **error** occured: Sum of proportions does **not** match 1
        ERROR_INVALID_PARAMETERS, //!< An **error** occured: invalid input parameters
        ERROR_DEGREE_TOO_LOW, //!< An **error** occured: the degree of the underlying polynomial is too low
        ERROR_INVALID_SPLINES, //!< An **error** occured: the list of splines or at least one element of it is invalid
        ERROR_INVALID_CONTROL_POINTS, //!< An **error** occured: the list of control points or at least one element of it is invalid
        ERROR_CURVE_INTERPOLATION_FAILED, //!< An **error** occured: the interpolation of a curve segment failed
        ERROR_NOTIMPLEMENTED, //!< An **error** occured: the desired algorithm is **not** yet implemented
        WARNING_HIGH_NUMERICAL_ERROR, //!< A **warning** occured: a high numerical error is expected
        SPLINERESULT_COUNT //!< Total count of elements
    };

    //! Returns the string representation of the given result type for spline algorithms
    static inline std::string splineResultString(const SplineResult& result)
    {
        // Check result
        switch (result) {
        case SplineResult::UNKNOWN:
            return "UNKNOWN";
        case SplineResult::SUCCESS:
            return "SUCCESS";
        case SplineResult::ERROR_INVALID_SPLINE_NO_SEGMENTS:
            return "ERROR_INVALID_SPLINE_NO_SEGMENTS";
        case SplineResult::ERROR_INVALID_SPLINE_PROPORTION_COUNT:
            return "ERROR_INVALID_SPLINE_PROPORTION_COUNT";
        case SplineResult::ERROR_INVALID_SPLINE_PROPORTION_NEGATIVE:
            return "ERROR_INVALID_SPLINE_PROPORTION_NEGATIVE";
        case SplineResult::ERROR_INVALID_SPLINE_PROPORTION_SUM:
            return "ERROR_INVALID_SPLINE_PROPORTION_SUM";
        case SplineResult::ERROR_INVALID_PARAMETERS:
            return "ERROR_INVALID_PARAMETERS";
        case SplineResult::ERROR_DEGREE_TOO_LOW:
            return "ERROR_DEGREE_TOO_LOW";
        case SplineResult::ERROR_INVALID_SPLINES:
            return "ERROR_INVALID_SPLINE_LIST";
        case SplineResult::ERROR_INVALID_CONTROL_POINTS:
            return "ERROR_INVALID_CONTROL_POINTS";
        case SplineResult::ERROR_CURVE_INTERPOLATION_FAILED:
            return "ERROR_CURVE_INTERPOLATION_FAILED";
        case SplineResult::ERROR_NOTIMPLEMENTED:
            return "ERROR_NOTIMPLEMENTED";
        case SplineResult::WARNING_HIGH_NUMERICAL_ERROR:
            return "WARNING_HIGH_NUMERICAL_ERROR";
        default:
            return "UNKNOWN";
        }
    }

    //! Template class for splines as concatenation of segments (curves)
    /*!
     * Contains an ordered list of spline segments (\ref m_segments). The segment/curve type has to be specified at compile time (\p CurveType). Each segment has an underlying curve
     * function, which is bounded by the *segment-position* \f$ \in [0,\,1] \f$. The complete spline is also bounded by *spline-position* \f$ \in [0,\,1] \f$ where in general
     * *spline-position* is **not** equal to the *segment-position*. Thus there has to be a mapping between *spline-position* and *segment-position* such that we can find out
     * which segment corresponds to which *spline-position*. This is done on a percentage basis (\ref m_segmentProportions). Considering the derivative of the underlying
     * function, the gradient relates to the *spline-coordinates*, **not** to the *segment-coordinates* (mapping is performed automatically).
     *
     * \par Segmentation
     * \code
     * |0--------1|0--------1|0-------1|  <-- segment-coordinates
     *   Segment0 | Segment1 | Segment2
     *   Propor.0 | Propor.1 | Propor.2
     *
     * |0-----------------------------1|  <-- spline-coordinates
     *              Spline
     * \endcode
     */
    template <class CurveType>
    class Spline {
    public:
        //! Constructor
        Spline()
        {
        }

        //! Destructor
        virtual ~Spline()
        {
        }

        //! Comparison operator: **equality**
        virtual bool operator==(const Spline<CurveType>& reference) const
        {
            // Compare segments
            if (m_segments != reference.m_segments)
                return false;

            // Compare segment proportions
            if (m_segmentProportions != reference.m_segmentProportions)
                return false;

            // Otherwise -> equality
            return true;
        }

        //! Comparison operator: **inequality**
        virtual bool operator!=(const Spline<CurveType>& reference) const { return !(*this == reference); }

        // Members
        std::vector<CurveType> m_segments; //!< Ordered list of spline segments (each segment has its own boundaries: *segment-position* \f$ \in [0,\,1] \f$)
        std::vector<double> m_segmentProportions; //!< Proportion of single segments relative to complete spline "duration" (each element has to be \f$ >=0 \f$ and the sum of all elements has to be exactly 1)

        //! Checks, if the spline is properly defined (proper definition of segments and proportions)
        /*!
         * \param [out] result Pointer to flag in which the result should be stored
         * \return `true`, if the spline is valid, `false` otherwise.
         */
        inline bool isValid(SplineResult* const result = nullptr) const
        {
            // Check, if there is at least one segment
            if (m_segments.size() == 0) {
                if (result != nullptr)
                    *result = SplineResult::ERROR_INVALID_SPLINE_NO_SEGMENTS;
                return false;
            }

            // Check, if segment and proportion count are the same
            if (m_segments.size() != m_segmentProportions.size()) {
                if (result != nullptr)
                    *result = SplineResult::ERROR_INVALID_SPLINE_PROPORTION_COUNT;
                return false;
            }

            // Check, proportion entries
            // (each entry has to be >=0 and the sum of all entries has to be 1.0)
            double sumOfProportions = 0;
            for (size_t i = 0; i < m_segmentProportions.size(); i++) {
                if (m_segmentProportions[i] < 0) {
                    if (result != nullptr)
                        *result = SplineResult::ERROR_INVALID_SPLINE_PROPORTION_NEGATIVE;
                    return false;
                }
                sumOfProportions += m_segmentProportions[i];
            }
            if (fabs(sumOfProportions - 1.0) > minimumSplineSegmentProportion) {
                if (result != nullptr)
                    *result = SplineResult::ERROR_INVALID_SPLINE_PROPORTION_SUM;
                return false;
            }

            // Otherwise -> valid
            if (result != nullptr)
                *result = SplineResult::SUCCESS;
            return true;
        }

        //! Computes the segment boundaries (control point positions) represented in **spline coordinates** ([0, 1])
        /*!
         * \warning You may call \ref isValid() before to check, if the spline is valid! If the spline is invalid the behaviour of this method is not defined!
         *
         * \return Ordered list of segment boundaries, or empty list in case of an error.
         *
         * \code
         * |--------S--------| <-- spline
         * |---s1---|---s2---| <-- spline segments
         * B1       B2      B3 <-- segment boundaries
         * 0       0.x       1 <-- returned values
         * \endcode
         */
        inline std::vector<double> getSegmentBoundaries() const
        {
            // Initialize helpers
            std::vector<double> boundaries;

            // Abort, if there are no segments
            if (m_segmentProportions.size() == 0)
                return boundaries;

            // Pre-allocate memory for speedup
            boundaries.reserve(m_segmentProportions.size() + 1);

            // Add "left" boundary of fist segment
            boundaries.push_back(0);

            // Get "right" boundary of all segments
            for (size_t i = 0; i < m_segmentProportions.size(); i++)
                boundaries.push_back(boundaries.back() + m_segmentProportions[i]);

            // Force last boundary position to match exactly 1 (to avoid numeric errors)
            boundaries.back() = 1;

            // Pass back list
            return boundaries;
        }

        //! Removes segments with zero proportion (i.e. no contribution to the spline)
        /*! \return Count of removed segments */
        inline size_t removeEmptySegments()
        {
            // Abort, if there are no segments (or invalid dimensions)
            if (m_segments.size() == 0 || m_segments.size() != m_segmentProportions.size())
                return 0;

            // Iterate through segments
            size_t removedSegments = 0;
            size_t nextFreeSlot = 0; // Next "free" in "new" segment list
            for (size_t i = 0; i < m_segments.size(); i++) {
                // Check proportion
                if (m_segmentProportions[i] > 0) {
                    // Necessary segment -> copy to next free slot (only if not already there)
                    if (i > nextFreeSlot) {
                        m_segments[nextFreeSlot] = m_segments[i];
                        m_segmentProportions[nextFreeSlot] = m_segmentProportions[i];
                    }
                    nextFreeSlot++;
                } else {
                    // Unnecessary segment -> do not copy!
                    removedSegments++;
                }
            }

            // Remove garbage data at end of vectors
            if (removedSegments > 0) {
                m_segments.resize(m_segments.size() - removedSegments);
                m_segmentProportions.resize(m_segmentProportions.size() - removedSegments);
            }

            // Pass back count of removed elements
            return removedSegments;
        }

        //! Returns the index of the segment related to the given spline position
        /*!
         * \warning You may call \ref isValid() before to check, if the spline is valid! If the spline is invalid the behaviour of this method is not defined!
         *
         * \param [in] splinePosition Position in `spline-coordinates` (must be within \f$ [0,\,1] \f$, values are projected otherwise)
         * \param [out] segmentPosition Position within segment corresponding to the given spline position (optional additional output)
         * \return Index of the segment or -1 in case of an error
         */
        inline int getSegmentIndex(const double& splinePosition, double* const segmentPosition = nullptr) const
        {
            // Abort, if segment count and proportion count do not match (to avoid invalid read access)
            if (m_segments.size() != m_segmentProportions.size())
                return -1;

            // Project position to boundaries
            double projectedSplinePosition = splinePosition;
            if (projectedSplinePosition < 0)
                projectedSplinePosition = 0;
            if (projectedSplinePosition > 1)
                projectedSplinePosition = 1;

            // Search for segment index and segment position (optional)
            double startPosition = 0; // Start position of current segment in spline-coordinates
            double endPosition = 0; // End position of current segment in spline-coordinates
            for (size_t i = 0; i < m_segments.size(); i++) {
                // Compute end position
                endPosition = startPosition + m_segmentProportions[i];
                if (i == m_segments.size() - 1)
                    endPosition = 1.0; // Hard-setting to 1.0 in case of the last segment (to avoid numeric errors)

                // Check, if target position is between start and end
                if (projectedSplinePosition >= startPosition && projectedSplinePosition <= endPosition) {
                    // We found the corresponding segment!

                    // Optional: get segment position from coordinate transformation
                    if (segmentPosition != nullptr) {
                        // Compute coordinate transformation (spline-coordinates <-> segment-coordinates)
                        // Linear relationship: segment-position = alpha * spline-position + beta
                        // Derivation: d(segment-position) / d(spline-position) = alpha
                        double alpha = 0;
                        if (endPosition != startPosition)
                            alpha = 1.0 / (endPosition - startPosition);
                        double beta = -alpha * startPosition;

                        // Compute segment position from spline position
                        *segmentPosition = alpha * projectedSplinePosition + beta;
                    }

                    // Stop search and pass back result
                    return (int)i;
                } else {
                    // Otherwise: go to next segment
                    startPosition = endPosition; // Start position of next segment is end position of current segment
                }
            }

            // We could not find the corresponding segment -> error
            return -1;
        }

        //! Evaluation of (derivative of) underlying function
        /*!
         * \warning You may call \ref isValid() before to check, if the spline is valid! If the spline is invalid the behaviour of this method is not defined!
         *
         * \warning The derivative is returned relative to the *spline-coordinates* (**not** relative to *segment-coordinates*)!
         *
         * \param [in] splinePosition Position to evaluate (derivative) at in `spline-coordinates` (must be within \f$ [0,\,1] \f$, values are projected otherwise)
         * \param [in] derivationOrder Order of the derivation (use 0 for base function, i.e. no derivation)
         * \return Value (of derivative) at specified position - gradient is expressed in *spline-coordinates*
         */
        virtual double evaluate(const double& splinePosition, const unsigned int& derivationOrder = 0) const
        {
            // Get segment index and position within this segment
            double segmentPosition = 0;
            int segmentIndex = getSegmentIndex(splinePosition, &segmentPosition);

            // Abort, if we did **not** find the corresponding segment
            if (segmentIndex < 0)
                return 0;

            // Check if we only want the base function (for speed up of computation)
            if (derivationOrder == 0) {
                // ... -> just evaluate underlying function of segment and return this value
                return m_segments[segmentIndex].evaluate(segmentPosition, derivationOrder);
            } else {
                // Compute coordinate transformation (spline-coordinates <-> segment-coordinates)
                // Linear relationship: segment-position = alpha * spline-position + beta
                // Derivation: d(segment-position) / d(spline-position) = alpha
                double alpha = 0;
                if (m_segmentProportions[segmentIndex] > 0)
                    alpha = 1.0 / m_segmentProportions[segmentIndex];

                // Compute gradient relative to segment coordinates
                double segmentGradient = m_segments[segmentIndex].evaluate(segmentPosition, derivationOrder);

                // Compute gradient relative to spline coordinates
                double splineGradient = segmentGradient * alpha;
                for (unsigned int i = 2; i <= derivationOrder; i++)
                    splineGradient *= alpha;

                // Pass back transformed gradient
                return splineGradient;
            }
        }

        //! Evaluation of value (D0) and first derivative (D1) of underlying function (improved performance over calling \ref evaluate() two times)
        /*!
         * \warning You may call \ref isValid() before to check, if the spline is valid! If the spline is invalid the behaviour of this method is not defined!
         *
         * \warning The derivative is returned relative to the *spline-coordinates* (**not** relative to *segment-coordinates*)!
         *
         * \param [in] splinePosition Position to evaluate (derivative) at in `spline-coordinates` (must be within \f$ [0,\,1] \f$, values are projected otherwise)
         * \return Value and first derivative at specified position - gradient is expressed in *spline-coordinates*
         */
        inline std::array<double, 2> evaluateD0D1(const double& splinePosition) const
        {
            // Prepare return value
            std::array<double, 2> returnValue({ { 0 } });

            // Get segment index and position within this segment
            double segmentPosition = 0;
            int segmentIndex = getSegmentIndex(splinePosition, &segmentPosition);

            // Abort, if we did **not** find the corresponding segment
            if (segmentIndex < 0)
                return returnValue;

            // Compute value
            returnValue[0] = m_segments[segmentIndex].evaluate(segmentPosition, 0);

            // Compute coordinate transformation (spline-coordinates <-> segment-coordinates)
            // Linear relationship: segment-position = alpha * spline-position + beta
            // Derivation: d(segment-position) / d(spline-position) = alpha
            double alpha = 0;
            if (m_segmentProportions[segmentIndex] > 0)
                alpha = 1.0 / m_segmentProportions[segmentIndex];

            // Compute first derivative
            returnValue[1] = alpha * m_segments[segmentIndex].evaluate(segmentPosition, 1);

            // Pass back result
            return returnValue;
        }

        //! Evaluation of value (D0), first (D1), and second derivative (D2) of underlying function (improved performance over calling \ref evaluate() three times)
        /*!
         * \warning You may call \ref isValid() before to check, if the spline is valid! If the spline is invalid the behaviour of this method is not defined!
         *
         * \warning The derivatives are returned relative to the *spline-coordinates* (**not** relative to *segment-coordinates*)!
         *
         * \param [in] splinePosition Position to evaluate (derivatives) at in `spline-coordinates` (must be within \f$ [0,\,1] \f$, values are projected otherwise)
         * \return Value, first, and second derivative at specified position - gradients are expressed in *spline-coordinates*
         */
        inline std::array<double, 3> evaluateD0D1D2(const double& splinePosition) const
        {
            // Prepare return value
            std::array<double, 3> returnValue({ { 0 } });

            // Get segment index and position within this segment
            double segmentPosition = 0;
            int segmentIndex = getSegmentIndex(splinePosition, &segmentPosition);

            // Abort, if we did **not** find the corresponding segment
            if (segmentIndex < 0)
                return returnValue;

            // Compute value
            returnValue[0] = m_segments[segmentIndex].evaluate(segmentPosition, 0);

            // Compute coordinate transformation (spline-coordinates <-> segment-coordinates)
            // Linear relationship: segment-position = alpha * spline-position + beta
            // Derivation: d(segment-position) / d(spline-position) = alpha
            double alpha = 0;
            if (m_segmentProportions[segmentIndex] > 0)
                alpha = 1.0 / m_segmentProportions[segmentIndex];

            // Compute first derivative
            returnValue[1] = alpha * m_segments[segmentIndex].evaluate(segmentPosition, 1);

            // Compute second derivative
            returnValue[2] = alpha * alpha * m_segments[segmentIndex].evaluate(segmentPosition, 2);

            // Pass back result
            return returnValue;
        }

        //! Interpolate coefficients of underlying function to fit specific parameters (through-points, derivatives, ...)
        /*!
         * A detailed description on how to set the \p parameters is given below at the corresponding section.
         * \param [in] parameters Interpolation parameters (meaning varies depending on chosen method - for details see corresponding sections below)
         * \param [in] proportions Pointer to segment proportions. If `nullptr`, all segments have the same proportion (uniform distribution), otherwise uses given proportions.
         * \param [in] method Interpolation method to be used (has to be compatible with specified curve type)
         * \param [out] result Pointer to flag in which the result of the algorithm should be stored
         * \return `true` on success, `false` otherwise
         */
        bool interpolate(const std::vector<double>& parameters, const std::vector<double>* const proportions, const SplineInterpolationMethod& method, SplineResult* const result = nullptr)
        {
            // Initialize helpers
            CurveResult curveResult = CurveResult::UNKNOWN;

            // Assume success
            if (result != nullptr)
                *result = SplineResult::SUCCESS;

            // Check, if proportions are specified
            bool proportionsSpecified = false;
            if (proportions != nullptr)
                if (proportions->size() > 0)
                    proportionsSpecified = true;

            // Clear current data
            m_segments.clear();
            if (proportionsSpecified == false)
                m_segmentProportions.clear();
            else
                m_segmentProportions = *proportions;

            /*! Notation:
             *  ---------
             * Symbol        | Code   | Description
             * ------        | -----  | -----------
             * \f$ n \f$     | `n`    | segment count
             * \f$ s_i \f$   | `si`   | spline segment \f$ i \f$
             * \f$ x_i \f$   | `xi`   | *segment-coordinate* of segment \f$ i \f$
             * \f$ h_i \f$   | `hi`   | segment "duration" of segment \f$ i \f$ (= segment proportion)
             * \f$ g_i \f$   | `gi`   | reciprocal of \f$ h_i \f$ (i.e. \f$ g_i = \frac{1}{h_i} \f$)
             * \f$ z \f$     | `z`    | *spline-coordinate* (\f$ z \in [0,\,1] \f$, \f$ z \f$ is **not** the time in general!)
             * \f$ y_i \f$   | `yi`   | value of spline segment \f$ i \f$
             * \f$ y'_i \f$  | `dyi`  | first order derivative (with respect to *spline-coordinate* \f$ z \f$, **not** the time in general!) of spline segment \f$ i \f$
             * \f$ y''_i \f$ | `ddyi` | second order derivative (with respect to *spline-coordinate* \f$ z \f$, **not** the time in general!) of spline segment \f$ i \f$
             *
             * \par Segmentation
             * \code
             * | ---- s0 ---- | ---- s1 ---- | ---- ... ---- | ---- si ---- | ---- ... ---- | ---- sn-1 ---- |
             * 0 ---- x0 --- 1|0 --- x1 --- 1|0 --- ... --- 1|0 --- xi --- 1|0 --- ... --- 1|0 --- xn-1 --- 1|
             * | ---- h0 ---- | ---- h1 ---- | ---- ... ---- | ---- hi ---- | ---- ... ---- | ---- hn-1 ---- |
             * 0 ------------------------------------------- z --------------------------------------------- 1
             * z0 ----------- z1 ----------- z2 --- ... ---- zi ---------- zi+1 -- ... --- zn-1 ------------ zn
             * y0 ----------- y1 ----------- y2 --- ... ---- yi ---------- yi+1 -- ... --- yn-1 ------------ yn
             * dy0 --------- dy1 ---------- dy2 --- ... --- dyi --------- dyi+1 -- ... -- dyn-1 ----------- dyn
             * ddy0 ------- ddy1 --------- ddy2 --- ... -- ddyi -------- ddyi+1 -- ... - ddyn-1 ---------- ddyn
             * \endcode
             *
             * **Note:** \f$ x_{i+1} \f$ will be written as `xip1` ("p" for "plus") and \f$ x_{i-1} \f$ will be written as `xim1` ("m" for "minus")
             */

            /*! Notes on Coordinate Systems
             *  ---------------------------
             * \par Coordinate transformation
             * The transformation from *spline-coordinates* \f$ z \f$ to *segment-coordinates* \f$x_i \f$ is given by \f$ x_i(z) = \alpha_i\,z + \beta_i \f$.
             *
             * With the boundary values:
             * - Start: \f$ x_i(z_i) = \alpha_i \, z_i + \beta_i = 0 \f$
             * - End: \f$ x_i(z_{i+1}) = \alpha_i\,z_{i+1} + \beta_i = 1 \f$
             *
             * we obtain
             * \f[ \alpha_i = \frac{1}{z_{i+1} - z_i} = \frac{1}{h_i} = g_i \f]
             * and
             * \f[ \beta_i = - \frac{z_i}{h_i} \f]
             *
             * \par Derivatives
             * When switching from *spline-coordinates* \f$ z \f$ to *segment-coordinates* \f$x_i \f$ and back one has to consider the transformation of derivatives (i.e. the chain rule).
             *
             * \f[\def\arraystretch{2}
             * \begin{array}{ll}
             * \displaystyle\frac{\partial x_i(z)}{\partial z} &= \displaystyle\alpha_i = \frac{1}{hi} \\
             * \displaystyle\frac{\partial^2 x_i(z)}{\partial z^2} &= \displaystyle0 \\
             * \displaystyle\frac{\partial s_i(z)}{\partial z} &= \displaystyle\frac{\partial s_i(x_i)}{\partial x_i} \cdot \frac{\partial x_i(z)}{\partial z} = \frac{\partial s_i(x_i)}{\partial x_i} \cdot \frac{1}{h_i} \\
             * \displaystyle\frac{\partial^2 s_i(z)}{\partial z^2} &= \displaystyle\frac{\partial^2 s_i(x_i)}{\partial x_i^2} \cdot \left(\frac{\partial x_i(z)}{\partial z}\right)^2 + \frac{\partial s_i(x_i)}{\partial x_i} \cdot \underbrace{\frac{\partial^2 x_i(z)}{\partial z^2}}_{=0} = \frac{\partial^2 s_i(x_i)}{\partial x_i^2} \cdot \left(\frac{1}{h_i}\right)^2 \\
             * \displaystyle\frac{\partial^3 s_i(z)}{\partial z^3} &= \displaystyle\frac{\partial^3 s_i(x_i)}{\partial x_i^3} \cdot \left(\frac{1}{h_i}\right)^3 \\
             * \displaystyle\frac{\partial^4 s_i(z)}{\partial z^4} &= \displaystyle\frac{\partial^4 s_i(x_i)}{\partial x_i^4} \cdot \left(\frac{1}{h_i}\right)^4 \\
             * \displaystyle & \displaystyle\vdots
             * \end{array}
             * \f]
             */

            // Check type of underlying function
            switch (CurveType::functionType()) {
            // ======================================
            // --- Case: POLYNOMIAL interpolation ---
            // ======================================
            case Curve::FunctionType::POLYNOMIAL: {
                // Check interpolation method
                switch (method) {
                case SplineInterpolationMethod::POLYNOMIAL_PIECEWISE_CONSTANT: {
                    /*! Interpolation with POLYNOMIAL_PIECEWISE_CONSTANT
                     *  ------------------------------------------------
                     * **Piecewise** interpolation with polynomials of order 0. The segments are piecewise constant. The spline is **not** \f$ C^0 \f$ continuous.
                     *
                     * \par Parameter-layout
                     * Element            | Description
                     * -------            | -----------
                     * \p parameters[0]   | function value of first segment
                     * ...                | ...
                     * \p parameters[n-1] | function value of last segment
                     */

                    // Check curve type
                    // no test necessary...

                    // Checking parameter set
                    if (parameters.size() == 0 || (proportionsSpecified == true && m_segmentProportions.size() != parameters.size())) {
                        if (result != nullptr)
                            *result = SplineResult::ERROR_INVALID_PARAMETERS;
                        return false;
                    }

                    // Initialize helpers
                    size_t segmentCount = parameters.size();

                    // Allocate memory
                    m_segments.resize(segmentCount);
                    if (proportionsSpecified == false) // use uniform distribution
                        m_segmentProportions.resize(segmentCount, (double)1.0 / segmentCount);

                    // Create segments and set their coefficients
                    for (size_t i = 0; i < segmentCount; i++) {
                        // Prepare indices (mapping in parameter array)
                        const double& yi = parameters[i];

                        // Interpolate curve segment
                        if (m_segments[i].interpolate(std::vector<double>{ yi }, CurveInterpolationMethod::POLYNOMIAL_CONSTANT, &curveResult) == false) {
                            if (result != nullptr)
                                *result = SplineResult::ERROR_CURVE_INTERPOLATION_FAILED;
                            return false;
                        }
                    }
                    break;
                }
                case SplineInterpolationMethod::POLYNOMIAL_PIECEWISE_LINEAR: {
                    /*! Interpolation with POLYNOMIAL_PIECEWISE_LINEAR
                     *  ----------------------------------------------
                     * **Piecewise** interpolation with polynomials of order 1. The segments are piecewise linear. Each segment is linked to its neighbor.
                     * The spline is \f$ C^0 \f$ continuous.
                     *
                     * \par Parameter-layout
                     * Element            | Description
                     * -------            | -----------
                     * \p parameters[0]   | function value at beginning of first segment
                     * ...                | ...
                     * \p parameters[n-1] | function value at beginning of last segment
                     * \p parameters[n]   | function value at end of last segment
                     */

                    // Check curve type
                    if (CurveType::degree() < 1) {
                        if (result != nullptr)
                            *result = SplineResult::ERROR_DEGREE_TOO_LOW;
                        return false;
                    }

                    // Checking parameter set
                    if (parameters.size() <= 1 || (proportionsSpecified == true && m_segmentProportions.size() != parameters.size() - 1)) {
                        if (result != nullptr)
                            *result = SplineResult::ERROR_INVALID_PARAMETERS;
                        return false;
                    }

                    // Initialize helpers
                    size_t segmentCount = parameters.size() - 1;

                    // Allocate memory
                    m_segments.resize(segmentCount);
                    if (proportionsSpecified == false) // use uniform distribution
                        m_segmentProportions.resize(segmentCount, (double)1.0 / segmentCount);

                    // Create segments and set their coefficients
                    for (size_t i = 0; i < segmentCount; i++) {
                        // Prepare indices (mapping in parameter array)
                        const double& yi = parameters[i];
                        const double& yip1 = parameters[i + 1];

                        // Interpolate curve segment
                        if (m_segments[i].interpolate(std::vector<double>{ yi, yip1 }, CurveInterpolationMethod::POLYNOMIAL_LINEAR, &curveResult) == false) {
                            if (result != nullptr)
                                *result = SplineResult::ERROR_CURVE_INTERPOLATION_FAILED;
                            return false;
                        }
                    }
                    break;
                }
                case SplineInterpolationMethod::POLYNOMIAL_PIECEWISE_CUBIC_FIRST_DERIVATIVES: {
                    /*! Interpolation with POLYNOMIAL_PIECEWISE_CUBIC_FIRST_DERIVATIVES
                     *  ---------------------------------------------------------------
                     * **Piecewise** interpolation with polynomials of order 3 (given first order derivatives). The segments are piecewise cubic.
                     * The segments are linked to their neighbors. The spline is \f$ C^1 \f$ continuous. The user can specify the first order
                     * derivatives (with respect to spline coordinates, **not** time in general!) at each control point.
                     *
                     * \par Parameter-layout
                     * Element             | Description
                     * -------             | -----------
                     * \p parameters[0]    | function value at beginning of first segment
                     * ...                 | ...
                     * \p parameters[n-1]  | function value at beginning of last segment
                     * \p parameters[n]    | function value at end of last segment
                     * \p parameters[n+1]  | first order derivative at beginning of first segment (derivative with respect to *spline-coordinates*, **not** time in general!)
                     * ...                 | ...
                     * \p parameters[2n]   | first order derivative at beginning of last segment (derivative with respect to *spline-coordinates*, **not** time in general!)
                     * \p parameters[2n+1] | first order derivative at end of last segment (derivative with respect to *spline-coordinates*, **not** time in general!)
                     */

                    // Check curve type
                    if (CurveType::degree() < 3) {
                        if (result != nullptr)
                            *result = SplineResult::ERROR_DEGREE_TOO_LOW;
                        return false;
                    }

                    // Checking parameter set
                    if (parameters.size() < 4 || (proportionsSpecified == true && m_segmentProportions.size() != parameters.size() / 2 - 1) || parameters.size() % 2 != 0) {
                        if (result != nullptr)
                            *result = SplineResult::ERROR_INVALID_PARAMETERS;
                        return false;
                    }

                    // Initialize helpers
                    size_t segmentCount = parameters.size() / 2 - 1;

                    // Allocate memory
                    m_segments.resize(segmentCount);
                    if (proportionsSpecified == false) // use uniform distribution
                        m_segmentProportions.resize(segmentCount, (double)1.0 / segmentCount);

                    // Create segments and set their coefficients
                    for (size_t i = 0; i < segmentCount; i++) {
                        // Note on derivatives:
                        /* --------------------
                         * The chain rule has to be respected using the proportion of the current segment (hi)!
                         * si(xi=0)           = yi
                         * si(xi=1)           = yip1
                         * dsi(z)/dz|xi=0     = dyi    --> dsi(xi)/dxi|xi=0     = dyi * hi
                         * dsi(z)/dz|xi=1     = dyip1  --> dsi(xi)/dxi|xi=1     = dyip1 * hi
                         */

                        // Prepare indices (mapping in parameter array)
                        const double& yi = parameters[i];
                        const double& yip1 = parameters[i + 1];
                        const double& hi = m_segmentProportions[i];
                        double ddyiXhi = parameters[segmentCount + 1 + i] * hi; // ddyi * hi
                        double ddyip1Xhi = parameters[segmentCount + 1 + i + 1] * hi; // ddyip1 * hi

                        // Interpolate curve segment
                        if (m_segments[i].interpolate(std::vector<double>{ yi, yip1, ddyiXhi, ddyip1Xhi }, CurveInterpolationMethod::POLYNOMIAL_CUBIC_FIRST_DERIVATIVES, &curveResult) == false) {
                            if (result != nullptr)
                                *result = SplineResult::ERROR_CURVE_INTERPOLATION_FAILED;
                            return false;
                        }
                    }
                    break;
                }
                case SplineInterpolationMethod::POLYNOMIAL_PIECEWISE_CUBIC_SECOND_DERIVATIVES: {
                    /*! Interpolation with POLYNOMIAL_PIECEWISE_CUBIC_SECOND_DERIVATIVES
                     *  ----------------------------------------------------------------
                     * **Piecewise** interpolation with polynomials of order 3 (given second order derivatives). The segments are piecewise cubic.
                     * The segments are linked to their neighbors. The spline is \f$ C^0 \f$ continuous. The user can specify the second order
                     * derivatives (with respect to spline coordinates, **not** time in general!) at each control point.
                     *
                     * \par Parameter-layout
                     * Element             | Description
                     * -------             | -----------
                     * \p parameters[0]    | function value at beginning of first segment
                     * ...                 | ...
                     * \p parameters[n-1]  | function value at beginning of last segment
                     * \p parameters[n]    | function value at end of last segment
                     * \p parameters[n+1]  | second order derivative at beginning of first segment (derivative with respect to *spline-coordinates*, **not** time in general!)
                     * ...                 | ...
                     * \p parameters[2n]   | second order derivative at beginning of last segment (derivative with respect to *spline-coordinates*, **not** time in general!)
                     * \p parameters[2n+1] | second order derivative at end of last segment (derivative with respect to *spline-coordinates*, **not** time in general!)
                     */

                    // Check curve type
                    if (CurveType::degree() < 3) {
                        if (result != nullptr)
                            *result = SplineResult::ERROR_DEGREE_TOO_LOW;
                        return false;
                    }

                    // Checking parameter set
                    if (parameters.size() < 4 || (proportionsSpecified == true && m_segmentProportions.size() != parameters.size() / 2 - 1) || parameters.size() % 2 != 0) {
                        if (result != nullptr)
                            *result = SplineResult::ERROR_INVALID_PARAMETERS;
                        return false;
                    }

                    // Initialize helpers
                    size_t segmentCount = parameters.size() / 2 - 1;

                    // Allocate memory
                    m_segments.resize(segmentCount);
                    if (proportionsSpecified == false) // use uniform distribution
                        m_segmentProportions.resize(segmentCount, (double)1.0 / segmentCount);

                    // Create segments and set their coefficients
                    for (size_t i = 0; i < segmentCount; i++) {
                        // Note on derivatives:
                        /* --------------------
                         * The chain rule has to be respected using the proportion of the current segment (hi)!
                         * si(xi=0)           = yi
                         * si(xi=1)           = yip1
                         * dsi(z)/dz|xi=0     = dyi    --> dsi(xi)/dxi|xi=0     = dyi * hi
                         * dsi(z)/dz|xi=1     = dyip1  --> dsi(xi)/dxi|xi=1     = dyip1 * hi
                         * d^2si(z)/dz^2|xi=0 = ddyi   --> d^2si(xi)/dxi^2|xi=0 = ddyi * hi^2
                         * d^2si(z)/dz^2|xi=1 = ddyip1 --> d^2si(xi)/dxi^2|xi=1 = ddyip1 * hi^2
                         */

                        // Prepare indices (mapping in parameter array)
                        const double& yi = parameters[i];
                        const double& yip1 = parameters[i + 1];
                        const double& hi = m_segmentProportions[i];
                        double his = hi * hi; // hi^2 (squared)
                        double ddyiXhis = parameters[segmentCount + 1 + i] * his; // ddyi * hi^2
                        double ddyip1Xhis = parameters[segmentCount + 1 + i + 1] * his; // ddyip1 * hi^2

                        // Interpolate curve segment
                        if (m_segments[i].interpolate(std::vector<double>{ yi, yip1, ddyiXhis, ddyip1Xhis }, CurveInterpolationMethod::POLYNOMIAL_CUBIC_SECOND_DERIVATIVES, &curveResult) == false) {
                            if (result != nullptr)
                                *result = SplineResult::ERROR_CURVE_INTERPOLATION_FAILED;
                            return false;
                        }
                    }
                    break;
                }
                case SplineInterpolationMethod::POLYNOMIAL_PIECEWISE_QUINTIC_FIRST_AND_SECOND_DERIVATIVES: {
                    /*! Interpolation with POLYNOMIAL_PIECEWISE_QUINTIC_FIRST_AND_SECOND_DERIVATIVES
                     *  ----------------------------------------------------------------------------
                     * **Piecewise** interpolation with polynomials of order 5. The segments are piecewise quintic. The segments are linked to their neighbors.
                     * The spline is \f$ C^2 \f$ continuous. The user can specify the first and second order derivatives (with respect to spline coordinates,
                     * **not** time in general!) at each control point.
                     *
                     * \par Parameter-layout
                     * Element             | Description
                     * -------             | -----------
                     * \p parameters[0]    | function value at beginning of first segment
                     * ...                 | ...
                     * \p parameters[n-1]  | function value at beginning of last segment
                     * \p parameters[n]    | function value at end of last segment
                     * \p parameters[n+1]  | first order derivative at beginning of first segment (derivative with respect to *spline-coordinates*, **not** time in general!)
                     * ...                 | ...
                     * \p parameters[2n]   | first order derivative at beginning of last segment (derivative with respect to *spline-coordinates*, **not** time in general!)
                     * \p parameters[2n+1] | first order derivative at end of last segment (derivative with respect to *spline-coordinates*, **not** time in general!)
                     * \p parameters[2n+2] | second order derivative at beginning of first segment (derivative with respect to *spline-coordinates*, **not** time in general!)
                     * ...                 | ...
                     * \p parameters[3n+1] | second order derivative at beginning of last segment (derivative with respect to *spline-coordinates*, **not** time in general!)
                     * \p parameters[3n+2] | second order derivative at end of last segment (derivative with respect to *spline-coordinates*, **not** time in general!)
                     */

                    // Check curve type
                    if (CurveType::degree() < 5) {
                        if (result != nullptr)
                            *result = SplineResult::ERROR_DEGREE_TOO_LOW;
                        return false;
                    }

                    // Checking parameter set
                    if (parameters.size() < 6 || (proportionsSpecified == true && m_segmentProportions.size() != parameters.size() / 3 - 1) || parameters.size() % 3 != 0) {
                        if (result != nullptr)
                            *result = SplineResult::ERROR_INVALID_PARAMETERS;
                        return false;
                    }

                    // Initialize helpers
                    size_t segmentCount = parameters.size() / 3 - 1;

                    // Allocate memory
                    m_segments.resize(segmentCount);
                    if (proportionsSpecified == false) // use uniform distribution
                        m_segmentProportions.resize(segmentCount, (double)1.0 / segmentCount);

                    // Create segments and set their coefficients
                    for (size_t i = 0; i < segmentCount; i++) {
                        // Note on derivatives:
                        /* --------------------
                         * The chain rule has to be respected using the proportion of the current segment (hi)!
                         * si(xi=0)           = yi
                         * si(xi=1)           = yip1
                         * dsi(z)/dz|xi=0     = dyi    --> dsi(xi)/dxi|xi=0     = dyi * hi
                         * dsi(z)/dz|xi=1     = dyip1  --> dsi(xi)/dxi|xi=1     = dyip1 * hi
                         * d^2si(z)/dz^2|xi=0 = ddyi   --> d^2si(xi)/dxi^2|xi=0 = ddyi * hi^2
                         * d^2si(z)/dz^2|xi=1 = ddyip1 --> d^2si(xi)/dxi^2|xi=1 = ddyip1 * hi^2
                         */

                        // Prepare indices (mapping in parameter array)
                        const double& yi = parameters[i];
                        const double& yip1 = parameters[i + 1];
                        const double& hi = m_segmentProportions[i];
                        double his = hi * hi; // hi^2 (squared)
                        double dyiXhi = parameters[segmentCount + 1 + i] * hi; // dyi * hi
                        double dyip1Xhi = parameters[segmentCount + 1 + i + 1] * hi; // dyip1 * hi
                        double ddyiXhis = parameters[2 * segmentCount + 2 + i] * his; // ddyi * hi^2
                        double ddyip1Xhis = parameters[2 * segmentCount + 2 + i + 1] * his; // ddyip1 * hi^2

                        // Interpolate curve segment
                        if (m_segments[i].interpolate(std::vector<double>{ yi, yip1, dyiXhi, dyip1Xhi, ddyiXhis, ddyip1Xhis }, CurveInterpolationMethod::POLYNOMIAL_QUINTIC_FIRST_AND_SECOND_DERIVATIVES, &curveResult) == false) {
                            if (result != nullptr)
                                *result = SplineResult::ERROR_CURVE_INTERPOLATION_FAILED;
                            return false;
                        }
                    }
                    break;
                }
                case SplineInterpolationMethod::POLYNOMIAL_SMOOTH_CUBIC_SECOND_DERIVATIVES: {
                    /*! Interpolation with POLYNOMIAL_SMOOTH_CUBIC_SECOND_DERIVATIVES
                     *  -------------------------------------------------------------
                     * **Smooth** interpolation with polynomials of order 3 (given second order derivatives at start and end of spline).
                     * The segments are piecewise cubic. The segments are linked to their neighbors. The spline is \f$ C^2 \f$ continuous.
                     * The second order derivatives (with respect to *spline-coordinates*, **not** time in general!) at start and end
                     * are fixed to a given value.
                     *
                     * \par Parameter-layout
                     * Element             | Description
                     * -------             | -----------
                     * \p parameters[0]    | function value at beginning of first segment
                     * ...                 | ...
                     * \p parameters[n-1]  | function value at beginning of last segment
                     * \p parameters[n]    | function value at end of last segment
                     * \p parameters[n+1]  | second order derivative at beginning of first segment (derivative with respect to *spline-coordinates*, **not** time in general!)
                     * \p parameters[n+2]  | second order derivative at end of last segment (derivative with respect to *spline-coordinates*, **not** time in general!)
                     */

                    // Check curve type
                    if (CurveType::degree() < 3) {
                        if (result != nullptr)
                            *result = SplineResult::ERROR_DEGREE_TOO_LOW;
                        return false;
                    }

                    // Checking parameter set
                    if (parameters.size() < 4 || (proportionsSpecified == true && m_segmentProportions.size() != parameters.size() - 3)) {
                        if (result != nullptr)
                            *result = SplineResult::ERROR_INVALID_PARAMETERS;
                        return false;
                    }

                    // Initialize helpers
                    size_t segmentCount = parameters.size() - 3;

                    // Allocate memory
                    m_segments.resize(segmentCount);
                    if (proportionsSpecified == false) // use uniform distribution
                        m_segmentProportions.resize(segmentCount, (double)1.0 / segmentCount);

                    // Check, if there is more than one segment
                    if (segmentCount > 1) {
                        // Problem:
                        /* --------
                         * si(xi)          =   ai*xi^3 +   bi*xi^2 + ci*xi + di
                         * dsi(xi)/dxi     = 3*ai*xi^2 + 2*bi*xi   + ci
                         * d^2si(xi)/dxi^2 = 6*ai*xi   + 2*bi
                         *
                         * C0-continuity: (2*n Eqs.)
                         * (I)    si(xi=0)            = yi    (n Eqs.)
                         * (II)   si(xi=1)            = yip1  (n Eqs.)
                         *
                         * C1-continuity: (n-1 Eqs.)
                         * (III)  dsi(z)/dz|xi=1      = dsip1(z)/dz|xip1=0     --> 1/hi * dsi(xi)/dxi|xi=1 = 1/hip1 * dsip1(xip1)/dxip1|xip1=0
                         *
                         * C2-continuity: (n-1 Eqs.)
                         * (IV)   d^2si(z)/dz^2|xi=1  = d^2sip1(z)/dz^2|xip1=0 --> (1/hi)^2 * d^2si(xi)/dxi^2|xi=1 = (1/hip1)^2 * d^2sip1(xip1)/dxip1^2|xip1=0
                         *
                         * Boundary values: (2 Eqs.)
                         * (V)    d^2s0(z)/dz^2|z=0   = ddy0                   --> d^2s0(x0)/dx0^2|x0=0 = ddy0 * h0^2
                         * (VI)   d^2snm1(z)/dz^2|z=1 = ddyn                   --> d^2snm1(xnm1)/dxnm1^2|xnm1=1 = ddyn * hnm1^2
                         */

                        // Solution:
                        /* ---------
                         * ai = hi^2/6*(ddyip1 - ddyi)
                         * bi = hi^2/2*ddyi
                         * ci = yip1 - yi - hi^2/6*(ddyip1 + 2*ddyi)
                         * di = yi
                         *
                         * From C1-continuity: (and inserting ai, bi, ci, di)
                         * -------------------
                         * him1*ddyim1 + 2*(him1 + hi)*ddyi + hi*ddyip1 = 6/hi*(yip1 - yi) - 6/him1*(yi - yim1)  for all i=1...n-1
                         *
                         * (n-1) unknowns: [ddy1, ddy2, ... , ddynm1]
                         * (ddy0 and ddyn are the known boundary values)
                         *
                         * --> can be written into matrix form with tridiagonal matrix A:
                         *
                         * --------------------------------------------   --------   --------
                         * |  d1    u1                            0   |   |  x1  |   |  b1  |
                         * |  l2    d2    u2                          |   |  x2  |   |  b2  |
                         * |        l3    d3    u3                    |   |  x3  |   |  b3  |
                         * |              ..    ..    ..              |   |  ..  |   |  ..  |
                         * |              li    di    ui              | * |  xi  | = |  bi  |
                         * |                    ..    ..    ..        |   |  ..  |   |  ..  |
                         * |                         ln-2  dn-2  un-2 |   | xn-2 |   | bn-2 |
                         * |  0                            ln-1  dn-1 |   | xn-1 |   | bn-1 |
                         * --------------------------------------------   --------   --------
                         *                      |                            |          |
                         *                      A                       *    x     =    b
                         *                [(n-1)x(n-1)]                  [(n-1)x1]  [(n-1)x1]
                         *
                         * with x = [ ddy1, ddy2, ... , ddyi, ... , ddyn-1 ]^T as the unknown solution of the linear system of equations
                         */

                        // Step 1: Setup linear system of equations
                        // ----------------------------------------
                        Eigen::VectorXd lowerDiagonalOfA(segmentCount - 1); // First element is "outside" the matrix A (to be ignored)
                        Eigen::VectorXd diagonalOfA(segmentCount - 1);
                        Eigen::VectorXd upperDiagonalOfA(segmentCount - 1); // Last element is "outside" the matrix A (to be ignored)
                        Eigen::VectorXd b(segmentCount - 1);
                        const double& ddy0 = parameters[segmentCount + 1]; // Second order derivative with respect to *spline-coordinates* at start of spline
                        const double& ddyn = parameters[segmentCount + 2]; // Second order derivative with respect to *spline-coordinates* at end of spline
                        const double& h0 = m_segmentProportions[0]; // Proportion of first segment
                        const double& hnm1 = m_segmentProportions[segmentCount - 1]; // Proportion of last segment
                        for (size_t k = 0; k < segmentCount - 1; k++) {
                            size_t i = k + 1; // Index shift is necessary, since we want to skip i=0 and comply with the previous **not**ation/meaning of index "i"

                            // Create references to current values
                            const double& yim1 = parameters[i - 1];
                            const double& yi = parameters[i];
                            const double& yip1 = parameters[i + 1];
                            const double& hi = m_segmentProportions[i];
                            const double& him1 = m_segmentProportions[i - 1];

                            // Compute lower diagonal values
                            if (i == 1)
                                lowerDiagonalOfA(k) = 0; // First element is "outside" the matrix A (to be ignored)
                            else
                                lowerDiagonalOfA(k) = him1;

                            // Compute diagonal values
                            diagonalOfA(k) = 2.0 * (him1 + hi);

                            // Compute upper diagonal values
                            if (i == segmentCount - 1)
                                upperDiagonalOfA(k) = 0; // Last element is "outside" the matrix A (to be ignored)
                            else
                                upperDiagonalOfA(k) = hi;

                            // Compute values in b
                            b(k) = 6.0 / hi * (yip1 - yi) - 6.0 / him1 * (yi - yim1);
                            if (i == 1)
                                b(k) -= h0 * ddy0;
                            if (i == segmentCount - 1)
                                b(k) -= hnm1 * ddyn;
                        }

                        // Step 2: Solve linear system of equations
                        // ----------------------------------------
                        Eigen::MatrixXd X(segmentCount - 1, 1); // Solution vector x = [ ddy1, ddy2, ... , ddyi, ... , ddyn-1 ]^T
                        core::math::solveTriDiagonalSystemOfEquations(lowerDiagonalOfA, diagonalOfA, upperDiagonalOfA, b, X);

                        // Step 3: Compute coefficients
                        // ----------------------------
                        for (size_t i = 0; i < segmentCount; i++) {
                            // Create references to current values
                            const double& yi = parameters[i];
                            const double& yip1 = parameters[i + 1];
                            const double& hi = m_segmentProportions[i];
                            const double his = hi * hi; // hi^2
                            double ddyi = ddy0;
                            if (i > 0)
                                ddyi = X(i - 1, 0);
                            double ddyip1 = ddyn;
                            if (i < segmentCount - 1)
                                ddyip1 = X(i, 0);

                            // Set coefficients ai, bi, ci, di
                            m_segments[i].m_coefficients[3] = his / 6.0 * (ddyip1 - ddyi);
                            m_segments[i].m_coefficients[2] = his / 2.0 * ddyi;
                            m_segments[i].m_coefficients[1] = yip1 - yi - his / 6.0 * (ddyip1 + 2.0 * ddyi);
                            m_segments[i].m_coefficients[0] = yi;
                        }

                        // Force higher terms to be zero (in case underlying polynomial is of higher degree)
                        if (CurveType::degree() > 3)
                            for (size_t i = 0; i < segmentCount; i++)
                                for (size_t j = 4; j < m_segments[i].m_coefficients.size(); j++)
                                    m_segments[i].m_coefficients[j] = 0;
                    } else {
                        // ...only one segment -> same solution as for piecewise cubic spline (see above for details)

                        // Prepare indices (mapping in parameter array)
                        const double& y0 = parameters[0];
                        const double& y0p1 = parameters[1];
                        const double& h0 = m_segmentProportions[0];
                        double h0s = h0 * h0; // h0^2 (squared)
                        double ddy0Xh0s = parameters[2] * h0s; // ddy0 * h0^2
                        double ddy0p1Xh0s = parameters[3] * h0s; // ddy0p1 * h0^2

                        // Interpolate curve segment
                        if (m_segments[0].interpolate(std::vector<double>{ y0, y0p1, ddy0Xh0s, ddy0p1Xh0s }, CurveInterpolationMethod::POLYNOMIAL_CUBIC_SECOND_DERIVATIVES, &curveResult) == false) {
                            if (result != nullptr)
                                *result = SplineResult::ERROR_CURVE_INTERPOLATION_FAILED;
                            return false;
                        }
                    }

                    break;
                }
                case SplineInterpolationMethod::POLYNOMIAL_SMOOTH_QUINTIC_FIRST_AND_SECOND_DERIVATIVES: {
                    /*! Interpolation with POLYNOMIAL_SMOOTH_QUINTIC_FIRST_AND_SECOND_DERIVATIVES
                     *  -------------------------------------------------------------------------
                     * **Smooth** interpolation with polynomials of order 5 (given first and second order derivatives at start and end of spline).
                     * The segments are piecewise quintic. The segments are linked to their neighbors. The spline is \f$ C^4 \f$ continuous.
                     * The first and second order derivatives (with respect to *spline-coordinates*, **not** time in general!) at start and end
                     * are fixed to a given value.
                     *
                     * \par Parameter-layout
                     * Element             | Description
                     * -------             | -----------
                     * \p parameters[0]    | function value at beginning of first segment
                     * ...                 | ...
                     * \p parameters[n-1]  | function value at beginning of last segment
                     * \p parameters[n]    | function value at end of last segment
                     * \p parameters[n+1]  | first order derivative at beginning of first segment (derivative with respect to *spline-coordinates*, **not** time in general!)
                     * \p parameters[n+2]  | first order derivative at end of last segment (derivative with respect to *spline-coordinates*, **not** time in general!)
                     * \p parameters[n+3]  | second order derivative at beginning of first segment (derivative with respect to *spline-coordinates*, **not** time in general!)
                     * \p parameters[n+4]  | second order derivative at end of last segment (derivative with respect to *spline-coordinates*, **not** time in general!)
                     */

                    // Check curve type
                    if (CurveType::degree() < 5) {
                        if (result != nullptr)
                            *result = SplineResult::ERROR_DEGREE_TOO_LOW;
                        return false;
                    }

                    // Checking parameter set
                    if (parameters.size() < 6 || (proportionsSpecified == true && m_segmentProportions.size() != parameters.size() - 5)) {
                        if (result != nullptr)
                            *result = SplineResult::ERROR_INVALID_PARAMETERS;
                        return false;
                    }

                    // Initialize helpers
                    size_t segmentCount = parameters.size() - 5;

                    // Allocate memory
                    m_segments.resize(segmentCount);
                    if (proportionsSpecified == false) // use uniform distribution
                        m_segmentProportions.resize(segmentCount, (double)1.0 / segmentCount);

                    // Check, if there is more than one segment
                    if (segmentCount > 1) {
                        // Problem:
                        /* --------
                         * si(xi)          =     ai*xi^5 +    bi*xi^4 +   ci*xi^3 +   di*xi^2 + ei*xi^1 + fi
                         * dsi(xi)/dxi     =   5*ai*xi^4 +  4*bi*xi^3 + 3*ci*xi^2 + 2*di*xi^1 + ei
                         * d^2si(xi)/dxi^2 =  20*ai*xi^3 + 12*bi*xi^2 + 6*ci*xi^1 + 2*di
                         * d^3si(xi)/dxi^3 =  60*ai*xi^2 + 24*bi*xi^1 + 6*ci
                         * d^4si(xi)/dxi^4 = 120*ai*xi^1 + 24*bi
                         *
                         * C0-continuity: (2*n Eqs.)
                         * (I)    si(xi=0)            = yi    (n Eqs.)
                         * (II)   si(xi=1)            = yip1  (n Eqs.)
                         *
                         * C1-continuity: (n-1 Eqs.)
                         * (III)  dsi(z)/dz|xi=1      = dsip1(z)/dz|xip1=0     --> 1/hi * dsi(xi)/dxi|xi=1 = 1/hip1 * dsip1(xip1)/dxip1|xip1=0
                         *
                         * C2-continuity: (n-1 Eqs.)
                         * (IV)   d^2si(z)/dz^2|xi=1  = d^2sip1(z)/dz^2|xip1=0 --> (1/hi)^2 * d^2si(xi)/dxi^2|xi=1 = (1/hip1)^2 * d^2sip1(xip1)/dxip1^2|xip1=0
                         *
                         * C3-continuity: (n-1 Eqs.)
                         * (V)    d^3si(z)/dz^3|xi=1  = d^3sip1(z)/dz^3|xip1=0 --> (1/hi)^3 * d^3si(xi)/dxi^3|xi=1 = (1/hip1)^3 * d^3sip1(xip1)/dxip1^3|xip1=0
                         *
                         * C4-continuity: (n-1 Eqs.)
                         * (VI)   d^4si(z)/dz^4|xi=1  = d^4sip1(z)/dz^4|xip1=0 --> (1/hi)^4 * d^4si(xi)/dxi^4|xi=1 = (1/hip1)^4 * d^4sip1(xip1)/dxip1^4|xip1=0
                         *
                         * Boundary values: (4 Eqs.)
                         * (VII)  ds0(z)/dz|z=0   = dy0                        --> ds0(x0)/dx0|x0=0 = dy0 * h0
                         * (VIII) dsnm1(z)/dz|z=1 = dyn                        --> dsnm1(xnm1)/dxnm1|xnm1=1 = dyn * hnm1
                         * (IX)   d^2s0(z)/dz^2|z=0   = ddy0                   --> d^2s0(x0)/dx0^2|x0=0 = ddy0 * h0^2
                         * (X)    d^2snm1(z)/dz^2|z=1 = ddyn                   --> d^2snm1(xnm1)/dxnm1^2|xnm1=1 = ddyn * hnm1^2
                         */

                        // Solution:
                        /* ---------
                         * ai = - 6*yi +  6*yip1 - 3*dyi*hi - 3*dyip1*hi - 0.5*ddyi*hi^2 + 0.5*ddyip1*hi^2
                         * bi =  15*yi - 15*yip1 + 8*dyi*hi + 7*dyip1*hi + 1.5*ddyi*hi^2 -     ddyip1*hi^2
                         * ci = -10*yi + 10*yip1 - 6*dyi*hi - 4*dyip1*hi - 1.5*ddyi*hi^2 + 0.5*ddyip1*hi^2
                         * di =                                            0.5*ddyi*hi^2
                         * ei =                      dyi*hi
                         * fi =     yi
                         *
                         * From C4-continuity: (and inserting ai, bi, ci, di, ei, fi)
                         * -------------------
                         * -56*dyim1*gim1^3 - 8*ddyim1*gim1^2 - 64*dyi*(gim1^3 + gi^3) + 12*ddyi*(gim1^2 - gi^2) -56*dyip1*gi^3 + 8*ddyip1*gi^2 = -120*gi^4*(yip1 - yi) - 120*gim1^4*(yi - yim1)   for all i=1...n-1
                         * (1st line in blocks of linear system of equations below)
                         *
                         * From C3-continuity: (and inserting ai, bi, ci, di, ei, fi)
                         * -------------------
                         * -8*dyim1*gim1^2 - ddyim1*gim1 - 12*dyi*(gim1^2 - gi^2) + 3*ddyi*(gim1 + gi) + 8*dyip1*gi^2 - ddyip1*gi = 20*gi^3*(yip1 - yi) - 20*gim1^3*(yi - yim1)   for all i=1...n-1
                         * (2nd line in blocks of linear system of equations below)
                         *
                         * 2*(n-1) unknowns: [dy1, ddy1, dy2, ddy2, ... , dynm1, ddynm1]
                         * (dy0, ddy0, dyn and ddyn are the known boundary values)
                         *
                         * --> can be written into matrix form with block-tridiagonal matrix A:
                         *
                         * --------------------------------------------   --------   --------
                         * |  D1    U1                            0   |   |  x1  |   |  b1  |
                         * |  L2    D2    U2                          |   |  x2  |   |  b2  |
                         * |        L3    D3    U3                    |   |  x3  |   |  b3  |
                         * |              ..    ..    ..              |   |  ..  |   |  ..  |
                         * |              Li    Di    Ui              | * |  xi  | = |  bi  |
                         * |                    ..    ..    ..        |   |  ..  |   |  ..  |
                         * |                         Ln-2  Dn-2  Un-2 |   | xn-2 |   | bn-2 |
                         * |  0                            Ln-1  Dn-1 |   | xn-1 |   | bn-1 |
                         * --------------------------------------------   --------   --------
                         *                      |                            |          |
                         *                      A                       *    x     =    b
                         *               [2(n-1)x2(n-1)]                [2(n-1)x1] [2(n-1)x1]
                         *
                         * with x = [ -dy1, 1/(alpha*g1)*ddy1, -dy2, 1/(alpha*g2)*ddy2, ... , -dyi, 1/(alpha*gi)*ddyi, ... , -dyn-1, 1/(alpha*gn-1)*ddyn-1 ]^T as the unknown solution of the linear system of equations
                         */

                        // Step 1: Setup linear system of equations
                        // ----------------------------------------
                        const double alpha = sqrt(64.0 / 3.0); // Parameter to improve numerical stability (see Mund1975, p. 280)
                        const double alphasquared = alpha * alpha; // alpha^2
                        std::vector<Eigen::Matrix2d, Eigen::aligned_allocator<Eigen::Matrix2d>> lowerDiagonalOfA;
                        lowerDiagonalOfA.resize(segmentCount - 1); // First element is "outside" the matrix A (to be ignored)
                        std::vector<Eigen::Matrix2d, Eigen::aligned_allocator<Eigen::Matrix2d>> diagonalOfA;
                        diagonalOfA.resize(segmentCount - 1);
                        std::vector<Eigen::Matrix2d, Eigen::aligned_allocator<Eigen::Matrix2d>> upperDiagonalOfA;
                        upperDiagonalOfA.resize(segmentCount - 1); // Last element is "outside" the matrix A (to be ignored)
                        std::vector<Eigen::Vector2d, Eigen::aligned_allocator<Eigen::Vector2d>> b;
                        b.resize(segmentCount - 1);
                        const double& dy0 = parameters[segmentCount + 1]; // First order derivative with respect to *spline-coordinates* at start of spline
                        const double& dyn = parameters[segmentCount + 2]; // First order derivative with respect to *spline-coordinates* at end of spline
                        const double& ddy0 = parameters[segmentCount + 3]; // Second order derivative with respect to *spline-coordinates* at start of spline
                        const double& ddyn = parameters[segmentCount + 4]; // Second order derivative with respect to *spline-coordinates* at end of spline
                        const double& g0 = 1.0 / m_segmentProportions[0]; // Proportion of first segment
                        const double g0s = g0 * g0; // g0^2 (squared)
                        const double g0c = g0s * g0; // g0^3 (cubic)
                        const double& gnm1 = 1.0 / m_segmentProportions[segmentCount - 1]; // Proportion of last segment
                        const double gnm1s = gnm1 * gnm1; // gnm1^2 (squared)
                        const double gnm1c = gnm1s * gnm1; // gnm1^3 (cubic)
                        double gi = g0;
                        double gis = g0s;
                        double gic = g0c;
                        double giq = gic * gi; // gi^4 (quartic)
                        for (size_t k = 0; k < segmentCount - 1; k++) {
                            size_t i = k + 1; // Index shift is necessary, since we want to skip i=0 and comply with the previous **not**ation/meaning of index "i"

                            // Create references to current values
                            const double& yim1 = parameters[i - 1];
                            const double& yi = parameters[i];
                            const double& yip1 = parameters[i + 1];
                            double gim1 = gi;
                            double gim1s = gis; // gim1^2 (squared)
                            double gim1c = gic; // gim1^3 (cubic)
                            double gim1q = giq; // gim1^4 (quartic)
                            gi = 1.0 / m_segmentProportions[i];
                            gis = gi * gi; // gi^2 (squared)
                            gic = gis * gi; // gi^3 (cubic)
                            giq = gic * gi; // gi^4 (quartic)

                            // Compute upper diagonal values
                            if (i == segmentCount - 1)
                                upperDiagonalOfA[k] = Eigen::Matrix2d::Zero(); // Last element is "outside" the matrix A (to be ignored)
                            else {
                                const double gip1 = 1.0 / m_segmentProportions[i + 1];

                                upperDiagonalOfA[k](0, 0) = 56.0 * gic;
                                upperDiagonalOfA[k](0, 1) = 8.0 * alpha * gis * gip1;
                                upperDiagonalOfA[k](1, 0) = -8.0 * alpha * gic;
                                upperDiagonalOfA[k](1, 1) = -alphasquared * gis * gip1;
                            }

                            // Compute diagonal values
                            diagonalOfA[k](0, 0) = 64.0 * (gim1c + gic);
                            diagonalOfA[k](0, 1) = 12.0 * alpha * (gim1s - gis) * gi;
                            diagonalOfA[k](1, 0) = diagonalOfA[k](0, 1);
                            diagonalOfA[k](1, 1) = 3.0 * alphasquared * (gim1 + gi) * gis;

                            // Compute lower diagonal values
                            if (i == 1)
                                lowerDiagonalOfA[k] = Eigen::Matrix2d::Zero(); // First element is "outside" the matrix A (to be ignored)
                            else
                                lowerDiagonalOfA[k] = upperDiagonalOfA[k - 1].transpose();

                            // Compute values in b
                            b[k](0) = -120.0 * (giq * (yip1 - yi) + gim1q * (yi - yim1));
                            b[k](1) = 20.0 * alpha * (giq * (yip1 - yi) - gim1c * gi * (yi - yim1));
                            if (i == 1) {
                                b[k](0) += 56.0 * g0c * dy0 + 8.0 * g0s * ddy0;
                                b[k](1) += 8.0 * alpha * g0s * gi * dy0 + alpha * g0 * gi * ddy0;
                            }
                            if (i == segmentCount - 1) {
                                b[k](0) += 56.0 * gnm1c * dyn - 8.0 * gnm1s * ddyn;
                                b[k](1) += -8.0 * alpha * gnm1c * dyn + alpha * gnm1s * ddyn;
                            }
                        }

                        // Step 2: Solve linear system of equations
                        // ----------------------------------------
                        std::vector<Eigen::Vector2d, Eigen::aligned_allocator<Eigen::Vector2d>> x; // Solution vector x = [ -dy1, 1/(alpha*g1)*ddy1, -dy2, 1/(alpha*g2)*ddy2, ... , -dyi, 1/(alpha*gi)*ddyi, ... , -dyn-1, 1/(alpha*gn-1)*ddyn-1 ]^T
                        x.resize(segmentCount - 1);
                        core::math::solveBlockTriDiagonalSystemOfEquations<Eigen::Matrix2d, Eigen::Vector2d>(lowerDiagonalOfA, diagonalOfA, upperDiagonalOfA, b, x);

#ifndef NDEBUG
                        // Intermediate step: compute numerical error
                        // ------------------------------------------
                        // Compute total error made during solving linear system of equations
                        double totalError = 0;
                        for (size_t k = 0; k < segmentCount - 1; k++) {
                            size_t i = k + 1; // Index shift is necessary, since we want to skip i=0 and comply with the previous **not**ation/meaning of index "i"

                            // Compute bi by matrix vector multiplication
                            Eigen::Vector2d bi(0, 0);
                            if (i > 1)
                                bi += lowerDiagonalOfA[k] * x[k - 1];
                            bi += diagonalOfA[k] * x[k];
                            if (i < segmentCount - 1)
                                bi += upperDiagonalOfA[k] * x[k + 1];

                            // Compute difference between bi of left hand side and bi of right hand side
                            Eigen::Vector2d delta = bi - b[k];
                            totalError += delta.norm();
                        }

                        // Output warning
                        if (totalError > 1e-6)
                            if (result != nullptr)
                                *result = SplineResult::WARNING_HIGH_NUMERICAL_ERROR;
#endif

                        // Step 3: Transform solution
                        // --------------------------
                        // Transform from x -> eta
                        // with   x = [ -dy1, 1/(alpha*g1)*ddy1, -dy2, 1/(alpha*g2)*ddy2, ... , -dyi, 1/(alpha*gi)*ddyi, ... , -dyn-1, 1/(alpha*gn-1)*ddyn-1 ]^T
                        //          = [ x1, x2, ... , xi, ... , xn-1]^T
                        //      eta = [ dy1, ddy1, dy2, ddy2, ... , dyi, ddyi, ... , dyn-1, ddyn-1 ]^T
                        std::vector<double> eta;
                        eta.resize(2.0 * x.size());
                        for (size_t i = 0; i < segmentCount - 1; i++) {
                            eta[2 * i] = -x[i](0);
                            eta[2 * i + 1] = alpha * x[i](1) / m_segmentProportions[i + 1];
                        }

                        // Step 4: Compute coefficients
                        // ----------------------------
                        for (size_t i = 0; i < segmentCount; i++) {
                            // Create references to current values
                            const double& yi = parameters[i];
                            const double& yip1 = parameters[i + 1];
                            const double& hi = m_segmentProportions[i];
                            const double his = hi * hi; // hi^2
                            double dyiXhi = dy0 * hi; // dyi * hi
                            double ddyiXhis = ddy0 * his; // ddyi * hi^2
                            if (i > 0) {
                                dyiXhi = eta[2 * (i - 1)] * hi;
                                ddyiXhis = eta[2 * (i - 1) + 1] * his;
                            }
                            double dyip1Xhi = dyn * hi; // dyip1 * hi
                            double ddyip1Xhis = ddyn * his; // ddyip1 * hi^2
                            if (i < segmentCount - 1) {
                                dyip1Xhi = eta[2 * i] * hi;
                                ddyip1Xhis = eta[2 * i + 1] * his;
                            }

                            // Set coefficients ai, bi, ci, di, ei, fi
                            m_segments[i].m_coefficients[5] = -6.0 * yi + 6.0 * yip1 - 3.0 * dyiXhi - 3.0 * dyip1Xhi - 0.5 * ddyiXhis + 0.5 * ddyip1Xhis;
                            m_segments[i].m_coefficients[4] = 15.0 * yi - 15.0 * yip1 + 8.0 * dyiXhi + 7.0 * dyip1Xhi + 1.5 * ddyiXhis - ddyip1Xhis;
                            m_segments[i].m_coefficients[3] = -10.0 * yi + 10.0 * yip1 - 6.0 * dyiXhi - 4.0 * dyip1Xhi - 1.5 * ddyiXhis + 0.5 * ddyip1Xhis;
                            m_segments[i].m_coefficients[2] = 0.5 * ddyiXhis;
                            m_segments[i].m_coefficients[1] = dyiXhi;
                            m_segments[i].m_coefficients[0] = yi;
                        }

                        // Force higher terms to be zero (in case underlying polynomial is of higher degree)
                        if (CurveType::degree() > 5)
                            for (size_t i = 0; i < segmentCount; i++)
                                for (size_t j = 6; j < m_segments[i].m_coefficients.size(); j++)
                                    m_segments[i].m_coefficients[j] = 0;
                    } else {
                        // ...only one segment -> same solution as for piecewise quintic spline (see above for details)

                        // Prepare indices (mapping in parameter array)
                        const double& y0 = parameters[0];
                        const double& y1 = parameters[1];
                        const double& h0 = m_segmentProportions[0];
                        double h0s = h0 * h0; // h0^2 (squared)
                        double dy0Xh0 = parameters[2] * h0; // dy0 * h0
                        double dy1Xh0 = parameters[3] * h0; // dy1 * h0
                        double ddy0Xh0s = parameters[4] * h0s; // ddy0 * h0^2
                        double ddy1Xh0s = parameters[5] * h0s; // ddy1 * h0^2

                        // Interpolate curve segment
                        if (m_segments[0].interpolate(std::vector<double>{ y0, y1, dy0Xh0, dy1Xh0, ddy0Xh0s, ddy1Xh0s }, CurveInterpolationMethod::POLYNOMIAL_QUINTIC_FIRST_AND_SECOND_DERIVATIVES, &curveResult) == false) {
                            if (result != nullptr)
                                *result = SplineResult::ERROR_CURVE_INTERPOLATION_FAILED;
                            return false;
                        }
                    }
                    break;
                }
                default: {
                    // No interpolation methods implemented for now
                    if (result != nullptr)
                        *result = SplineResult::ERROR_NOTIMPLEMENTED;
                    return false;
                }
                }
                break;
            }
            // ========================================================
            // --- Case: TRIGONOMETRIC or EXPONENTIAL interpolation ---
            // ========================================================
            case Curve::FunctionType::TRIGONOMETRIC:
            case Curve::FunctionType::EXPONENTIAL: {
                // No interpolation methods implemented for now
                if (result != nullptr)
                    *result = SplineResult::ERROR_NOTIMPLEMENTED;
                return false;
            }
            default: {
                // Unknown function type
                if (result != nullptr)
                    *result = SplineResult::ERROR_NOTIMPLEMENTED;
                return false;
            }
            }

            // Success
            return true;
        }

        //! Splits **this** spline ("source") into multiple "target" splines (of the same type). The interconnection ("chaining") of the target splines preserves the shape of the source spline.
        /*!
         * \param [out] targetSplines List of pointers to target splines (list is constant, (values of) elements are modified)
         * \param [in] targetProportions Pointer to proportions of target splines. If `nullptr`, all target splines have the same proportion (uniform distribution, "split into equal parts"), otherwise uses given proportions.
         * \param [out] result Pointer to flag in which the result of the algorithm should be stored
         * \return `true` on success, `false` otherwise
         *
         * \par Notes on Implementation
         * In order to avoid loss of information one has to be careful when computing the segmentation of the target splines. It is **not** sufficient to evaluate the source spline at the
         * boundaries of the target splines (and use one segment per target spline), since there might be control points of the source spline within the boundaries of the target spline. Thus
         * we have to use control points at the boundaries of the target splines **and** the boundaries of the source spline segments (which may split a target spline into further segments.
         * This way we do not "miss" any control points and the original shape of the source spline is preserved. Moreover, we evaluate the source spline at the "left" and "right" boundaries
         * of the control points, since the source spline might not be continuous (probably jumps at control points).
         *
         * \code
         * |------------------S-------------------| <-- source spline (known)
         * |-----s1-----|-----s2-----|-----s3-----| <-- source spline segments (known)
         * C1           C2           C4          C5 <-- control points originating from source spline segment boundaries
         *
         * C1                 C3                 C5 <-- control points originating from target spline boundaries
         * |--------T1--------|---------T2--------| <-- target splines (partitioning known through user defined target spline proportions (optional))
         * |-----t11----|-t12-|--t21-|----t22-----| <-- target spline segments (unknown - connect control points)
         * C1           C2    C3     C4          C5 <-- control points (define segmentation of target splines and target spline segments)
         * \endcode
         */
        bool split(const std::vector<Spline<CurveType>*>& targetSplines, const std::vector<double>* const targetProportions, SplineResult* const result = nullptr) const
        {
            // Definition of control points (only used internally)
            /*
             * Helper class to store as much information as possible during setup of segmentation -> evaluation of source spline is then faster
             *
             * ...left source segment] [right source segment...
             *  ----------------------|------------------------
             *                        Ci
             *
             * Note: left and right source segment index are the same at control points which do not coincide with source spline segment boundaries ("inside" of a source spline segment)
             */
            class ControlPoint {
            public:
                // Default constructor
                ControlPoint()
                    : m_sourceSplinePosition(0)
                    , m_leftSourceSegmentIndex(0)
                    , m_rightSourceSegmentIndex(0)
                    , m_leftSourceSegmentPosition(0)
                    , m_rightSourceSegmentPosition(0)
                    , m_isSourceSegmentBoundary(false)
                {
                }

                // Specialized constructor
                ControlPoint(const double& sourceSplinePosition, const int& leftSourceSegmentIndex, const int& rightSourceSegmentIndex, const double& leftSourceSegmentPosition, const double& rightSourceSegmentPosition, const bool& isSourceSegmentBoundary)
                    : m_sourceSplinePosition(sourceSplinePosition)
                    , m_leftSourceSegmentIndex(leftSourceSegmentIndex)
                    , m_rightSourceSegmentIndex(rightSourceSegmentIndex)
                    , m_leftSourceSegmentPosition(leftSourceSegmentPosition)
                    , m_rightSourceSegmentPosition(rightSourceSegmentPosition)
                    , m_isSourceSegmentBoundary(isSourceSegmentBoundary)
                {
                }

                // Members
                double m_sourceSplinePosition; // Position of control point in source spline (described in source spline coordinates)
                int m_leftSourceSegmentIndex; // Index of source spline segment on "left" side of control point
                int m_rightSourceSegmentIndex; // Index of source spline segment on "right" side of control point
                double m_leftSourceSegmentPosition; // Position of control point in "left" source spline segment (described in left source spline segment coordinates)
                double m_rightSourceSegmentPosition; // Position of control point in "right" source spline segment (described in left source spline segment coordinates)
                bool m_isSourceSegmentBoundary; // Flag indicating, if control point originates from source spline segment boundary
            };

            // Initialize helpers
            std::vector<double> targetSplineBoundaryPositions; // Positions of target spline boundaries in source spline coordinates (start and end of each target spline when "cut" out of the source spline)
            std::vector<ControlPoint> controlPoints; // Total list of control points
            std::vector<size_t> targetSplineBeginControlPointIndices; // Index of control point at beginning of target spline "i"
            std::vector<size_t> targetSplineEndControlPointIndices; // Index of control point at end of target spline "i"

            // Pre-allocate memory for speedup
            targetSplineBoundaryPositions.reserve(targetSplines.size() + 1);
            targetSplineBeginControlPointIndices.reserve(targetSplines.size()); // One begin control point for each target spline
            targetSplineEndControlPointIndices.reserve(targetSplines.size()); // One end control point for each target spline

            // Assume success
            if (result != nullptr)
                *result = SplineResult::SUCCESS;

            // Check input for validity
            // ------------------------
            // Check, if source spline is valid
            if (isValid(result) == false)
                return false;

            // Check if list of target splines is empty
            if (targetSplines.size() == 0) {
                if (result != nullptr)
                    *result = SplineResult::ERROR_INVALID_SPLINES;
                return false;
            }

            // Check if any pointer in list of target splines is invalid
            for (size_t i = 0; i < targetSplines.size(); i++) {
                // Check pointer
                if (targetSplines[i] == nullptr) {
                    if (result != nullptr)
                        *result = SplineResult::ERROR_INVALID_SPLINES;
                    return false;
                }
            }

            // Check, if target spline proportions are specified, otherwise use uniform proportions
            std::vector<double> targetSplineProportions;
            if (targetProportions != nullptr) {
                // ...yes, proportions are specified -> check, if dimension of lists match
                if (targetProportions->size() == targetSplines.size()) {
                    // Dimensions do match -> use specified proportions
                    targetSplineProportions = *targetProportions;

                    // Check, proportion entries (each entry has to be >=0 and the sum of all entries has to be 1.0)
                    double sumOfProportions = 0;
                    for (size_t i = 0; i < targetSplineProportions.size(); i++) {
                        if (targetSplineProportions[i] < 0) {
                            if (result != nullptr)
                                *result = SplineResult::ERROR_INVALID_SPLINE_PROPORTION_NEGATIVE;
                            return false;
                        }
                        sumOfProportions += targetSplineProportions[i];
                    }
                    if (fabs(sumOfProportions - 1.0) > minimumSplineSegmentProportion) {
                        // Sum of proportions is not 1 -> error
                        if (result != nullptr)
                            *result = SplineResult::ERROR_INVALID_SPLINE_PROPORTION_SUM;
                        return false;
                    }
                } else {
                    // Dimensions do not match -> error!
                    if (result != nullptr)
                        *result = SplineResult::ERROR_INVALID_SPLINE_PROPORTION_COUNT;
                    return false;
                }
            } else {
                // ...no, proportions are not specified -> use uniform partitioning instead
                targetSplineProportions.resize(targetSplines.size(), (double)1.0 / targetSplines.size());
            }

            // Preprocessing
            // -------------
            // Create copy of source spline WITHOUT zero-proportion segments (these have no contribution to the target splines anyway and could lead to errors in computing derivatives)
            Spline<CurveType> preprocessedSourceSpline;
            preprocessedSourceSpline.m_segments.reserve(m_segments.size());
            preprocessedSourceSpline.m_segmentProportions.reserve(m_segmentProportions.size());
            for (size_t i = 0; i < m_segments.size(); i++) {
                if (m_segmentProportions[i] > 0) {
                    preprocessedSourceSpline.m_segments.push_back(m_segments[i]);
                    preprocessedSourceSpline.m_segmentProportions.push_back(m_segmentProportions[i]);
                }
            }

            // Compute control points
            // ----------------------
            // Get positions of source segment boundaries in source spline coordinates
            std::vector<double> sourceSegmentBoundaryPositions = preprocessedSourceSpline.getSegmentBoundaries();

            // Get positions of target spline boundaries in source spline coordinates
            targetSplineBoundaryPositions.push_back(0); // Add "left" boundary of first target spline
            for (size_t i = 0; i < targetSplineProportions.size(); i++)
                targetSplineBoundaryPositions.push_back(targetSplineBoundaryPositions.back() + targetSplineProportions[i]);
            targetSplineBoundaryPositions.back() = 1; // Force last control point to match exactly 1 (to avoid numeric errors)

            // Compute a SORTED list of control points WITHOUT DUPLICATES from combining list of source segment boundaries and list of target spline boundaries
            controlPoints.reserve(sourceSegmentBoundaryPositions.size() + targetSplineBoundaryPositions.size()); // Allocate memory for best performance (over-estimate, since there might be duplicates)
            controlPoints.push_back(ControlPoint(0, 0, 0, 0, 0, true)); // Add initial control point ("left" boundary)
            size_t currentSourceSegmentBoundaryIndex = 1; // Currently investigated element in list of source segment boundaries (skip first element which is always zero and already in list)
            size_t currentTargetSplineBoundaryIndex = 1; // Currently investigated element in list of target spline boundaries (skip first element which is always zero and already in list)
            while (currentSourceSegmentBoundaryIndex < sourceSegmentBoundaryPositions.size() || currentTargetSplineBoundaryIndex < targetSplineBoundaryPositions.size()) {
                // Initialize helpers
                bool isSourceSegmentBoundary = false; // Flag indicating, if the control point originates from a source spline segment boundary (otherwise it originates from a target spline boundary)

                // Check, from which list we should pick the new control point
                if (currentSourceSegmentBoundaryIndex < sourceSegmentBoundaryPositions.size()) {
                    // List of source segment boundaries is non-empty -> check, if list of target spline boundaries is also non-empty
                    if (currentTargetSplineBoundaryIndex < targetSplineBoundaryPositions.size()) {
                        // Both lists are non-empty -> check, which list element is closer to zero (leftmost)
                        if (sourceSegmentBoundaryPositions[currentSourceSegmentBoundaryIndex] <= targetSplineBoundaryPositions[currentTargetSplineBoundaryIndex])
                            isSourceSegmentBoundary = true;
                    } else {
                        // List of target spline boundaries is empty -> use next source segment boundary as this is our only option
                        isSourceSegmentBoundary = true;
                    }
                }
                // Otherwise: list of source segment boundaries is empty -> use next target spline boundary as this is our only option

                // Fill data for control point
                if (isSourceSegmentBoundary == true) {
                    // Candidate originates from source segment boundary
                    ControlPoint newControlPoint = ControlPoint(sourceSegmentBoundaryPositions[currentSourceSegmentBoundaryIndex], currentSourceSegmentBoundaryIndex - 1, currentSourceSegmentBoundaryIndex, 1, 0, true);

                    // Project source segment information to boundaries of source spline
                    if (newControlPoint.m_leftSourceSegmentIndex < 0) {
                        // Project to beginning of first source segment
                        newControlPoint.m_leftSourceSegmentIndex = 0;
                        newControlPoint.m_leftSourceSegmentPosition = 0;
                    }
                    if (newControlPoint.m_rightSourceSegmentIndex >= (int)preprocessedSourceSpline.m_segments.size()) {
                        // Project to end of last source segment
                        newControlPoint.m_rightSourceSegmentIndex = preprocessedSourceSpline.m_segments.size() - 1;
                        newControlPoint.m_rightSourceSegmentPosition = 1;
                    }

                    // Add new control point to list (only if it differs from previous control point)
                    if (newControlPoint.m_sourceSplinePosition - controlPoints.back().m_sourceSplinePosition > minimumSplineSegmentProportion)
                        controlPoints.push_back(newControlPoint);
                    else {
                        // Control points coincide -> replace old one if old one originates from target spline boundary (to use better information on left/right segment)
                        if (controlPoints.back().m_isSourceSegmentBoundary == false)
                            controlPoints.back() = newControlPoint;
                    }

                    // "Remove" control point from list
                    currentSourceSegmentBoundaryIndex++;
                } else {
                    // Candidate originates from target spline boundary -> assume that this control point lies WITHIN a source spline segment and does NOT coincide with source segment boundary
                    // (this assumption might be wrong, but control point will be overwritten by source segment boundary control point in case we were wrong)
                    double sourceSegmentPosition = 0;
                    int sourceSegmentIndex = getSegmentIndex(targetSplineBoundaryPositions[currentTargetSplineBoundaryIndex], &sourceSegmentPosition);
                    if (sourceSegmentIndex < 0) {
                        if (result != nullptr)
                            *result = SplineResult::ERROR_INVALID_PARAMETERS;
                        return false;
                    }
                    ControlPoint newControlPoint = ControlPoint(targetSplineBoundaryPositions[currentTargetSplineBoundaryIndex], sourceSegmentIndex, sourceSegmentIndex, sourceSegmentPosition, sourceSegmentPosition, false);

                    // Add new control point to list (only if it differs from previous control point)
                    if (newControlPoint.m_sourceSplinePosition - controlPoints.back().m_sourceSplinePosition > minimumSplineSegmentProportion)
                        controlPoints.push_back(newControlPoint);
                    // Otherwise: control points coincide -> do NOT replace old one (to avoid overwriting control points originating from source segment boundaries)

                    // "Remove" control point from list
                    currentTargetSplineBoundaryIndex++;
                }

                // Check, if this control point is related to a target spline boundary
                if (isSourceSegmentBoundary == false) {
                    // Remember, that the current target spline ended at this control point
                    targetSplineEndControlPointIndices.push_back(controlPoints.size() - 1);
                }
            }

            // Force last control point to match end of spline (to avoid numeric errors)
            controlPoints.back() = ControlPoint(1, preprocessedSourceSpline.m_segments.size() - 1, preprocessedSourceSpline.m_segments.size() - 1, 1, 1, true);

            // Compute begin control point indices for target splines
            targetSplineBeginControlPointIndices.push_back(0); // Always start with zero
            for (size_t i = 0; i < targetSplineEndControlPointIndices.size() - 1; i++)
                targetSplineBeginControlPointIndices.push_back(targetSplineEndControlPointIndices[i]); // Begin control point is end control point of previous target spline

            // Check list of control points
            // ----------------------------
            // There have to be at least two control points...
            if (controlPoints.size() < 2) {
                if (result != nullptr)
                    *result = SplineResult::ERROR_INVALID_CONTROL_POINTS;
                return false;
            }

            // Interpolate target splines
            // --------------------------
            // Iterate over target splines
            for (size_t i = 0; i < targetSplines.size(); i++) {
                // Compute segment count for current target spline
                size_t segmentCount = targetSplineEndControlPointIndices[i] - targetSplineBeginControlPointIndices[i];
                if (segmentCount == 0)
                    segmentCount = 1; // Force always at least one segment

                // Compute segmentation of target spline
                targetSplines[i]->m_segmentProportions.clear();
                targetSplines[i]->m_segmentProportions.reserve(segmentCount);
                if (segmentCount == 1)
                    targetSplines[i]->m_segmentProportions.push_back(1.0);
                else
                    for (size_t j = targetSplineBeginControlPointIndices[i]; j < targetSplineEndControlPointIndices[i]; j++)
                        targetSplines[i]->m_segmentProportions.push_back((controlPoints[j + 1].m_sourceSplinePosition - controlPoints[j].m_sourceSplinePosition) / targetSplineProportions[i]);

                // Reallocate segments
                targetSplines[i]->m_segments.clear();
                targetSplines[i]->m_segments.resize(segmentCount);

                // Iterate through segments of target spline
                size_t targetSegmentIndex = 0; // Index of currently investigated segment in target spline
                for (size_t j = targetSplineBeginControlPointIndices[i]; j < targetSplineEndControlPointIndices[i]; j++) {
                    // Get reference to target spline and target segment
                    Spline<CurveType>& targetSpline = *targetSplines[i];
                    CurveType& targetSegment = targetSpline.m_segments[targetSegmentIndex];

                    // Get references to corresponding begin and end control point
                    size_t& beginControlPointIndex = j;
                    size_t endControlPointIndex = j + 1;
                    if (endControlPointIndex > targetSplineEndControlPointIndices[i])
                        endControlPointIndex = targetSplineEndControlPointIndices[i];
                    const ControlPoint& beginControlPoint = controlPoints[beginControlPointIndex];
                    const ControlPoint& endControlPoint = controlPoints[endControlPointIndex];

                    // Get references to corresponding begin and end source segment to evaluate at
                    const CurveType& beginSourceSegment = preprocessedSourceSpline.m_segments[beginControlPoint.m_rightSourceSegmentIndex];
                    const CurveType& endSourceSegment = preprocessedSourceSpline.m_segments[endControlPoint.m_leftSourceSegmentIndex];

                    // Get reference to corresponding begin and end source segment proportion
                    const double& beginSourceSegmentProportion = preprocessedSourceSpline.m_segmentProportions[beginControlPoint.m_rightSourceSegmentIndex];
                    const double& endSourceSegmentProportion = preprocessedSourceSpline.m_segmentProportions[endControlPoint.m_leftSourceSegmentIndex];

                    // Get reference to corresponding begin and end source segment position to evaluate at
                    const double& beginSourceSegmentPosition = beginControlPoint.m_rightSourceSegmentPosition;
                    const double& endSourceSegmentPosition = endControlPoint.m_leftSourceSegmentPosition;

                    // Get reference to duration of current source segments
                    // WARNING: "duration" relative to SOURCE spline coordinates (complete time domain)
                    const double& beginSourceSegmentDuration = beginSourceSegmentProportion;
                    const double& endSourceSegmentDuration = endSourceSegmentProportion;

                    // Get reference to duration of current target segment
                    // WARNING: "duration" relative to SOURCE spline coordinates (complete time domain)
                    const double& targetSegmentDuration = targetSpline.m_segmentProportions[targetSegmentIndex] * targetSplineProportions[i];
                    const double targetSegmentDurationSquared = targetSegmentDuration * targetSegmentDuration;

                    // Check type of underlying curve
                    switch (CurveType::functionType()) {
                    // ========================
                    // --- Case: POLYNOMIAL ---
                    // ========================
                    case Curve::FunctionType::POLYNOMIAL: {
                        // Select interpolation method depending on degree of polynomial
                        switch (CurveType::degree()) {
                        case 0: {
                            // Piecewise constant interpolation

                            // Compute left boundary
                            const double y0 = beginSourceSegment.evaluate(beginSourceSegmentPosition, 0);

                            // Interpolate
                            if (targetSegment.interpolate(std::vector<double>{ y0 }, CurveInterpolationMethod::POLYNOMIAL_CONSTANT) == false) {
                                if (result != nullptr)
                                    *result = SplineResult::ERROR_CURVE_INTERPOLATION_FAILED;
                                return false;
                            }
                            break;
                        }
                        case 1: {
                            // Piecewise linear interpolation

                            // Compute left boundary
                            const double y0 = beginSourceSegment.evaluate(beginSourceSegmentPosition, 0);

                            // Compute right boundary
                            const double y1 = endSourceSegment.evaluate(endSourceSegmentPosition, 0);

                            // Interpolate
                            if (targetSegment.interpolate(std::vector<double>{ y0, y1 }, CurveInterpolationMethod::POLYNOMIAL_LINEAR) == false) {
                                if (result != nullptr)
                                    *result = SplineResult::ERROR_CURVE_INTERPOLATION_FAILED;
                                return false;
                            }
                            break;
                        }
                        case 3: {
                            // Piecewise cubic interpolation

                            // Compute left boundary
                            const double y0 = beginSourceSegment.evaluate(beginSourceSegmentPosition, 0);
                            double alpha0 = 0;
                            if (beginSourceSegmentDuration > 0)
                                alpha0 = 1.0 / beginSourceSegmentDuration;
                            const double dy0 = beginSourceSegment.evaluate(beginSourceSegmentPosition, 1) * alpha0 * targetSegmentDuration; // Derivative relative to target segment coordinates

                            // Compute right boundary
                            const double y1 = endSourceSegment.evaluate(endSourceSegmentPosition, 0);
                            double alpha1 = 0;
                            if (endSourceSegmentDuration > 0)
                                alpha1 = 1.0 / endSourceSegmentDuration;
                            const double dy1 = endSourceSegment.evaluate(endSourceSegmentPosition, 1) * alpha1 * targetSegmentDuration; // Derivative relative to target segment coordinates

                            // Interpolate
                            if (targetSegment.interpolate(std::vector<double>{ y0, y1, dy0, dy1 }, CurveInterpolationMethod::POLYNOMIAL_CUBIC_FIRST_DERIVATIVES) == false) {
                                if (result != nullptr)
                                    *result = SplineResult::ERROR_CURVE_INTERPOLATION_FAILED;
                                return false;
                            }
                            break;
                        }
                        case 5: {
                            // Piecewise quintic interpolation

                            // Compute left boundary
                            const double y0 = beginSourceSegment.evaluate(beginSourceSegmentPosition, 0);
                            double alpha0 = 0;
                            if (beginSourceSegmentDuration > 0)
                                alpha0 = 1.0 / beginSourceSegmentDuration;
                            const double alpha0Squared = alpha0 * alpha0;
                            const double dy0 = beginSourceSegment.evaluate(beginSourceSegmentPosition, 1) * alpha0 * targetSegmentDuration; // Derivative relative to target segment coordinates
                            const double ddy0 = beginSourceSegment.evaluate(beginSourceSegmentPosition, 2) * alpha0Squared * targetSegmentDurationSquared; // Derivative relative to target segment coordinates

                            // Compute right boundary
                            const double y1 = endSourceSegment.evaluate(endSourceSegmentPosition, 0);
                            double alpha1 = 0;
                            if (endSourceSegmentDuration > 0)
                                alpha1 = 1.0 / endSourceSegmentDuration;
                            const double alpha1Squared = alpha1 * alpha1;
                            const double dy1 = endSourceSegment.evaluate(endSourceSegmentPosition, 1) * alpha1 * targetSegmentDuration; // Derivative relative to target segment coordinates
                            const double ddy1 = endSourceSegment.evaluate(endSourceSegmentPosition, 2) * alpha1Squared * targetSegmentDurationSquared; // Derivative relative to target segment coordinates

                            // Interpolate
                            if (targetSegment.interpolate(std::vector<double>{ y0, y1, dy0, dy1, ddy0, ddy1 }, CurveInterpolationMethod::POLYNOMIAL_QUINTIC_FIRST_AND_SECOND_DERIVATIVES) == false) {
                                if (result != nullptr)
                                    *result = SplineResult::ERROR_CURVE_INTERPOLATION_FAILED;
                                return false;
                            }
                            break;
                        }
                        default: {
                            // No appropriate interpolation method for this specific degree implemented -> we can not interpolate without loosing information -> error
                            if (result != nullptr)
                                *result = SplineResult::ERROR_NOTIMPLEMENTED;
                            return false;
                        }
                        }
                        break;
                    }
                    // ==========================================
                    // --- Case: TRIGONOMETRIC or EXPONENTIAL ---
                    // ==========================================
                    case Curve::FunctionType::TRIGONOMETRIC:
                    case Curve::FunctionType::EXPONENTIAL: {
                        // No interpolation methods implemented for now
                        if (result != nullptr)
                            *result = SplineResult::ERROR_NOTIMPLEMENTED;
                        return false;
                    }
                    default: {
                        // Unknown function type
                        if (result != nullptr)
                            *result = SplineResult::ERROR_NOTIMPLEMENTED;
                        return false;
                    }
                    }

                    // Update target segment index
                    targetSegmentIndex++;
                }
            }

            // Success!
            return true;
        }

        //! Constructs **this** spline ("target") by joining multiple "source" splines (of the same type). The interconnection ("chaining") of the source splines defines the shape of the target spline.
        /*!
         * \param [in] sourceSplines List of pointers to source splines (list is constant, elements are constant too)
         * \param [in] sourceProportions Pointer to proportions of source splines. If `nullptr`, all source splines have the same proportion (uniform distribution, "join as equal parts"), otherwise uses given proportions.
         * \param [out] result Pointer to flag in which the result of the algorithm should be stored
         * \return `true` on success, `false` otherwise
         *
         * \par Notes on Implementation
         * \code
         * |-----------S1----------|-----------S2----------| <-- source splines (known)
         * |----s11----|----s12----|----s21----|----s22----| <-- source spline segments (known)
         * C1          C2          C3          C4         C5 <-- control points originating from source spline segment boundaries
         *
         * |-----------------------T-----------------------| <-- target spline (unknown)
         * |-----t1----|-----t2----|-----t3----|-----t4----| <-- target spline segments (unknown)
         * C1          C2          C3          C4         C5 <-- control points (define segmentation of target spline)
         * \endcode
         */
        bool join(const std::vector<Spline<CurveType> const*>& sourceSplines, const std::vector<double>* const sourceProportions, SplineResult* const result = nullptr)
        {
            // Definition of control points (only used internally)
            /*
             * Helper class to store as much information as possible during setup of segmentation -> evaluation of source splines is then faster
             *
             * ... left source spline} {right source spline ...
             * ...left source segment] [right source segment...
             *  ----------------------|------------------------
             *                        Ci
             */
            class ControlPoint {
            public:
                // Default constructor
                ControlPoint()
                    : m_targetSplinePosition(0)
                    , m_leftSourceSplineIndex(0)
                    , m_rightSourceSplineIndex(0)
                    , m_leftSourceSegmentIndex(0)
                    , m_rightSourceSegmentIndex(0)
                    , m_leftSourceSegmentPosition(0)
                    , m_rightSourceSegmentPosition(0)
                {
                }

                // Specialized constructor
                ControlPoint(const double& targetSplinePosition, const int& leftSourceSplineIndex, const int& rightSourceSplineIndex, const int& leftSourceSegmentIndex, const int& rightSourceSegmentIndex, const double& leftSourceSegmentPosition, const double& rightSourceSegmentPosition)
                    : m_targetSplinePosition(targetSplinePosition)
                    , m_leftSourceSplineIndex(leftSourceSplineIndex)
                    , m_rightSourceSplineIndex(rightSourceSplineIndex)
                    , m_leftSourceSegmentIndex(leftSourceSegmentIndex)
                    , m_rightSourceSegmentIndex(rightSourceSegmentIndex)
                    , m_leftSourceSegmentPosition(leftSourceSegmentPosition)
                    , m_rightSourceSegmentPosition(rightSourceSegmentPosition)
                {
                }

                // Members
                double m_targetSplinePosition; // Position of control point in target spline (described in target spline coordinates)
                int m_leftSourceSplineIndex; // Index of source spline on "left" side of control point
                int m_rightSourceSplineIndex; // Index of source spline on "right" side of control point
                int m_leftSourceSegmentIndex; // Index of source spline segment on "left" side of control point
                int m_rightSourceSegmentIndex; // Index of source spline segment on "right" side of control point
                double m_leftSourceSegmentPosition; // Position of control point in "left" source spline segment (described in left source spline segment coordinates)
                double m_rightSourceSegmentPosition; // Position of control point in "right" source spline segment (described in left source spline segment coordinates)
            };

            // Initialize helpers
            std::vector<ControlPoint> controlPoints; // Total list of control points

            // Assume success
            if (result != nullptr)
                *result = SplineResult::SUCCESS;

            // Check input for validity
            // ------------------------
            // Check if list of source splines is empty
            if (sourceSplines.size() == 0) {
                if (result != nullptr)
                    *result = SplineResult::ERROR_INVALID_SPLINES;
                return false;
            }

            // Check if any pointer in list of source splines is invalid, or if source splines are invalid themselves
            for (size_t i = 0; i < sourceSplines.size(); i++) {
                // Check pointer
                if (sourceSplines[i] == nullptr) {
                    if (result != nullptr)
                        *result = SplineResult::ERROR_INVALID_SPLINES;
                    return false;
                } else {
                    // Check spline validity
                    if (sourceSplines[i]->isValid() == false) {
                        if (result != nullptr)
                            *result = SplineResult::ERROR_INVALID_SPLINES;
                        return false;
                    }
                }
            }

            // Check, if source spline proportions are specified, otherwise use uniform proportions
            std::vector<double> sourceSplineProportions;
            if (sourceProportions != nullptr) {
                // ...yes, proportions are specified -> check, if dimension of lists match
                if (sourceProportions->size() == sourceSplines.size()) {
                    // Dimensions do match -> use specified proportions
                    sourceSplineProportions = *sourceProportions;

                    // Check, proportion entries (each entry has to be >=0 and the sum of all entries has to be 1.0)
                    double sumOfProportions = 0;
                    for (size_t i = 0; i < sourceSplineProportions.size(); i++) {
                        if (sourceSplineProportions[i] < 0) {
                            if (result != nullptr)
                                *result = SplineResult::ERROR_INVALID_SPLINE_PROPORTION_NEGATIVE;
                            return false;
                        }
                        sumOfProportions += sourceSplineProportions[i];
                    }
                    if (fabs(sumOfProportions - 1.0) > minimumSplineSegmentProportion) {
                        // Sum of proportions is not 1 -> error
                        if (result != nullptr)
                            *result = SplineResult::ERROR_INVALID_SPLINE_PROPORTION_SUM;
                        return false;
                    }
                } else {
                    // Dimensions do not match -> error!
                    if (result != nullptr)
                        *result = SplineResult::ERROR_INVALID_SPLINE_PROPORTION_COUNT;
                    return false;
                }
            } else {
                // ...no, proportions are not specified -> use uniform partitioning instead
                sourceSplineProportions.resize(sourceSplines.size(), (double)1.0 / sourceSplines.size());
            }

            // Preprocessing
            // -------------
            // Create copy of source spline list WITHOUT zero-proportion splines and zero-proportion segments (these have no contribution to the target spline anyway and could lead to errors in computing derivatives)
            std::vector<Spline<CurveType>> preprocessedSourceSplines;
            std::vector<double> preprocessedSourceSplineProportions;
            preprocessedSourceSplines.reserve(sourceSplines.size());
            preprocessedSourceSplineProportions.reserve(sourceSplineProportions.size());
            for (size_t i = 0; i < sourceSplines.size(); i++) {
                // Check if spline proportion is non-zero
                if (sourceSplineProportions[i] > 0) {
                    // ...yes -> add this spline to list
                    preprocessedSourceSplines.push_back(Spline<CurveType>());
                    preprocessedSourceSplineProportions.push_back(sourceSplineProportions[i]);
                    preprocessedSourceSplines.back().m_segments.reserve(sourceSplines[i]->m_segments.size());
                    preprocessedSourceSplines.back().m_segmentProportions.reserve(sourceSplines[i]->m_segmentProportions.size());

                    // Iterate through segments
                    for (size_t j = 0; j < sourceSplines[i]->m_segments.size(); j++) {
                        // Check if segment proportion is non-zero
                        if (sourceSplines[i]->m_segmentProportions[j] > 0) {
                            // ...yes -> add this segment to list
                            preprocessedSourceSplines.back().m_segments.push_back(sourceSplines[i]->m_segments[j]);
                            preprocessedSourceSplines.back().m_segmentProportions.push_back(sourceSplines[i]->m_segmentProportions[j]);
                        }
                    }
                }
            }

            // Compute control points
            // ----------------------
            // Get estimate for count of control points (over-estimate)
            size_t estimatedControlPointCount = 1;
            for (size_t i = 0; i < preprocessedSourceSplines.size(); i++)
                estimatedControlPointCount += preprocessedSourceSplines[i].m_segments.size();

            // Pre-allocate memory for best performance
            controlPoints.reserve(estimatedControlPointCount);

            // Add initial control point
            controlPoints.push_back(ControlPoint(0, 0, 0, 0, 0, 0, 0));

            // Create SORTED list of control points WITHOUT DUPLICATES
            double targetSplinePosition = 0;
            for (size_t i = 0; i < preprocessedSourceSplines.size(); i++) {
                // Add control point for every source segment boundary
                for (size_t j = 0; j < preprocessedSourceSplines[i].m_segmentProportions.size(); j++) {
                    // Update target spline position
                    targetSplinePosition += preprocessedSourceSplines[i].m_segmentProportions[j] * preprocessedSourceSplineProportions[i];

                    // Add new control point only, if there does not exist an control point at this position already
                    if (targetSplinePosition - controlPoints.back().m_targetSplinePosition > minimumSplineSegmentProportion) {
                        // Check, if this is the last segment in this spline
                        if (j < preprocessedSourceSplines[i].m_segmentProportions.size() - 1) {
                            // ...no, not last segment in current spline
                            controlPoints.push_back(ControlPoint(targetSplinePosition, i, i, j, j + 1, 1, 0));
                        } else {
                            // ...yes, last segment in current spline
                            controlPoints.push_back(ControlPoint(targetSplinePosition, i, i + 1, j, 0, 1, 0));
                        }
                    }
                }
            }

            // Force last control point to match end of spline (to avoid numeric errors)
            controlPoints.back() = ControlPoint(1, preprocessedSourceSplines.size() - 1, preprocessedSourceSplines.size() - 1, preprocessedSourceSplines.back().m_segments.size() - 1, preprocessedSourceSplines.back().m_segments.size() - 1, 1, 1);

            // Check list of control points
            // ----------------------------
            // There have to be at least two control points...
            if (controlPoints.size() < 2) {
                if (result != nullptr)
                    *result = SplineResult::ERROR_INVALID_CONTROL_POINTS;
                return false;
            }

            // Interpolate target spline
            // -------------------------
            // Compute segmentation of target spline
            const size_t targetSegmentCount = controlPoints.size() - 1;
            m_segmentProportions.clear();
            m_segmentProportions.reserve(targetSegmentCount);
            for (size_t i = 0; i < targetSegmentCount; i++)
                m_segmentProportions.push_back(controlPoints[i + 1].m_targetSplinePosition - controlPoints[i].m_targetSplinePosition);

            // Reallocate segments
            m_segments.clear();
            m_segments.resize(targetSegmentCount);

            // Iterate through target spline segments
            for (size_t i = 0; i < m_segments.size(); i++) {
                // Get reference to target spline and target segment
                CurveType& targetSegment = m_segments[i];

                // Get references to corresponding begin and end control point
                const ControlPoint& beginControlPoint = controlPoints[i];
                const ControlPoint& endControlPoint = controlPoints[i + 1];

                // Get references to corresponding begin and end source spline to evaluate at
                const Spline<CurveType>& beginSourceSpline = preprocessedSourceSplines[beginControlPoint.m_rightSourceSplineIndex];
                const Spline<CurveType>& endSourceSpline = preprocessedSourceSplines[endControlPoint.m_leftSourceSplineIndex];

                // Get references to corresponding begin and end source segment to evaluate at
                const CurveType& beginSourceSegment = beginSourceSpline.m_segments[beginControlPoint.m_rightSourceSegmentIndex];
                const CurveType& endSourceSegment = endSourceSpline.m_segments[endControlPoint.m_leftSourceSegmentIndex];

                // Get reference to corresponding begin and end source segment proportion
                const double& beginSourceSegmentProportion = beginSourceSpline.m_segmentProportions[beginControlPoint.m_rightSourceSegmentIndex];
                const double& endSourceSegmentProportion = endSourceSpline.m_segmentProportions[endControlPoint.m_leftSourceSegmentIndex];

                // Get reference to corresponding begin and end source segment position to evaluate at
                const double& beginSourceSegmentPosition = beginControlPoint.m_rightSourceSegmentPosition;
                const double& endSourceSegmentPosition = endControlPoint.m_leftSourceSegmentPosition;

                // Get reference to duration of current source segments
                // WARNING: "duration" relative to TARGET spline coordinates (complete time domain)
                const double& beginSourceSegmentDuration = beginSourceSegmentProportion * preprocessedSourceSplineProportions[beginControlPoint.m_rightSourceSplineIndex];
                const double& endSourceSegmentDuration = endSourceSegmentProportion * preprocessedSourceSplineProportions[endControlPoint.m_leftSourceSplineIndex];

                // Get reference to duration of current target segment
                // WARNING: "duration" relative to TARGET spline coordinates (complete time domain)
                const double& targetSegmentDuration = m_segmentProportions[i];
                const double targetSegmentDurationSquared = targetSegmentDuration * targetSegmentDuration;

                // Check type of underlying curve
                switch (CurveType::functionType()) {
                // ========================
                // --- Case: POLYNOMIAL ---
                // ========================
                case Curve::FunctionType::POLYNOMIAL: {
                    // Select interpolation method depending on degree of polynomial
                    switch (CurveType::degree()) {
                    case 0: {
                        // Piecewise constant interpolation

                        // Compute left boundary
                        const double y0 = beginSourceSegment.evaluate(beginSourceSegmentPosition, 0);

                        // Interpolate
                        if (targetSegment.interpolate(std::vector<double>{ y0 }, CurveInterpolationMethod::POLYNOMIAL_CONSTANT) == false) {
                            if (result != nullptr)
                                *result = SplineResult::ERROR_CURVE_INTERPOLATION_FAILED;
                            return false;
                        }
                        break;
                    }
                    case 1: {
                        // Piecewise linear interpolation

                        // Compute left boundary
                        const double y0 = beginSourceSegment.evaluate(beginSourceSegmentPosition, 0);

                        // Compute right boundary
                        const double y1 = endSourceSegment.evaluate(endSourceSegmentPosition, 0);

                        // Interpolate
                        if (targetSegment.interpolate(std::vector<double>{ y0, y1 }, CurveInterpolationMethod::POLYNOMIAL_LINEAR) == false) {
                            if (result != nullptr)
                                *result = SplineResult::ERROR_CURVE_INTERPOLATION_FAILED;
                            return false;
                        }
                        break;
                    }
                    case 3: {
                        // Piecewise cubic interpolation

                        // Compute left boundary
                        const double y0 = beginSourceSegment.evaluate(beginSourceSegmentPosition, 0);
                        double alpha0 = 0;
                        if (beginSourceSegmentDuration > 0)
                            alpha0 = 1.0 / beginSourceSegmentDuration;
                        const double dy0 = beginSourceSegment.evaluate(beginSourceSegmentPosition, 1) * alpha0 * targetSegmentDuration; // Derivative relative to target segment coordinates

                        // Compute right boundary
                        const double y1 = endSourceSegment.evaluate(endSourceSegmentPosition, 0);
                        double alpha1 = 0;
                        if (endSourceSegmentDuration > 0)
                            alpha1 = 1.0 / endSourceSegmentDuration;
                        const double dy1 = endSourceSegment.evaluate(endSourceSegmentPosition, 1) * alpha1 * targetSegmentDuration; // Derivative relative to target segment coordinates

                        // Interpolate
                        if (targetSegment.interpolate(std::vector<double>{ y0, y1, dy0, dy1 }, CurveInterpolationMethod::POLYNOMIAL_CUBIC_FIRST_DERIVATIVES) == false) {
                            if (result != nullptr)
                                *result = SplineResult::ERROR_CURVE_INTERPOLATION_FAILED;
                            return false;
                        }
                        break;
                    }
                    case 5: {
                        // Piecewise quintic interpolation

                        // Compute left boundary
                        const double y0 = beginSourceSegment.evaluate(beginSourceSegmentPosition, 0);
                        double alpha0 = 0;
                        if (beginSourceSegmentDuration > 0)
                            alpha0 = 1.0 / beginSourceSegmentDuration;
                        const double alpha0Squared = alpha0 * alpha0;
                        const double dy0 = beginSourceSegment.evaluate(beginSourceSegmentPosition, 1) * alpha0 * targetSegmentDuration; // Derivative relative to target segment coordinates
                        const double ddy0 = beginSourceSegment.evaluate(beginSourceSegmentPosition, 2) * alpha0Squared * targetSegmentDurationSquared; // Derivative relative to target segment coordinates

                        // Compute right boundary
                        const double y1 = endSourceSegment.evaluate(endSourceSegmentPosition, 0);
                        double alpha1 = 0;
                        if (endSourceSegmentDuration > 0)
                            alpha1 = 1.0 / endSourceSegmentDuration;
                        const double alpha1Squared = alpha1 * alpha1;
                        const double dy1 = endSourceSegment.evaluate(endSourceSegmentPosition, 1) * alpha1 * targetSegmentDuration; // Derivative relative to target segment coordinates
                        const double ddy1 = endSourceSegment.evaluate(endSourceSegmentPosition, 2) * alpha1Squared * targetSegmentDurationSquared; // Derivative relative to target segment coordinates

                        // Interpolate
                        if (targetSegment.interpolate(std::vector<double>{ y0, y1, dy0, dy1, ddy0, ddy1 }, CurveInterpolationMethod::POLYNOMIAL_QUINTIC_FIRST_AND_SECOND_DERIVATIVES) == false) {
                            if (result != nullptr)
                                *result = SplineResult::ERROR_CURVE_INTERPOLATION_FAILED;
                            return false;
                        }
                        break;
                    }
                    default: {
                        // No appropriate interpolation method for this specific degree implemented -> we can not interpolate without loosing information -> error
                        if (result != nullptr)
                            *result = SplineResult::ERROR_NOTIMPLEMENTED;
                        return false;
                    }
                    }
                    break;
                }
                // ==========================================
                // --- Case: TRIGONOMETRIC or EXPONENTIAL ---
                // ==========================================
                case Curve::FunctionType::TRIGONOMETRIC:
                case Curve::FunctionType::EXPONENTIAL: {
                    // No interpolation methods implemented for now
                    if (result != nullptr)
                        *result = SplineResult::ERROR_NOTIMPLEMENTED;
                    return false;
                }
                default: {
                    // Unknown function type
                    if (result != nullptr)
                        *result = SplineResult::ERROR_NOTIMPLEMENTED;
                    return false;
                }
                }
            }

            // Success!
            return true;
        }

        // Encoding
        // --------
        //! Encode member data as XML element and add it to the specified stream
        /*!
         * \param [in,out] stream The stream to which the member data should be appended to.
         * \param [in] XMLIndentationLevel Level of indentation in XML format (used to add spaces)
         * \param [in] XMLTabWhiteSpaces Count of whitespaces corresponding to a tab.
         * \param [in] numericFormat C-format specifier for numeric values.
         * \return Count of appended characters (elements in stream).
         */
        virtual io::encoding::CharacterStreamSize encodeToXML(io::encoding::CharacterStream& stream, const size_t& XMLIndentationLevel, const size_t& XMLTabWhiteSpaces, const std::string& numericFormat = "%.7g") const
        {
            io::encoding::CharacterStreamSize addedElements = 0;

            // Start XML element
            for (size_t i = 0; i < XMLIndentationLevel * XMLTabWhiteSpaces; i++)
                addedElements += io::encoding::encode(stream, " ");
            addedElements += io::encoding::encode(stream, "<Spline");

            // Write attributes
            addedElements += io::encoding::encode(stream, " SegmentProportions=\"");
            for (size_t i = 0; i < m_segmentProportions.size(); i++) {
                if (i > 0)
                    addedElements += io::encoding::encode(stream, ",");
                addedElements += io::encoding::encode(stream, (double)m_segmentProportions[i], numericFormat);
            }
            addedElements += io::encoding::encode(stream, "\">\n");

            // Write segments
            for (size_t i = 0; i < m_segments.size(); i++)
                addedElements += m_segments[i].encodeToXML(stream, XMLIndentationLevel + 1, XMLTabWhiteSpaces, numericFormat);

            // End XML element
            for (size_t i = 0; i < XMLIndentationLevel * XMLTabWhiteSpaces; i++)
                addedElements += io::encoding::encode(stream, " ");
            addedElements += io::encoding::encode(stream, "</Spline>\n");

            // Pass back added elements in stream
            return addedElements;
        }

    public:
        EIGEN_MAKE_ALIGNED_OPERATOR_NEW // <-- Proper 128 bit alignment of member data necessary for Eigen vectorization
    }; // namespace planning

} // namespace curve
} // namespace broccoli

#endif // HAVE_EIGEN3
