/*
 * This file is part of broccoli.
 * Copyright (C) 2019 Chair of Applied Mechanics, Technical University of Munich
 * https://www.amm.mw.tum.de/
 */

#pragma once // Load this file only once

#include "io/Joystick.hpp"
#include "io/JoystickAction.hpp"
#include "io/JoystickEvent.hpp"
#include "io/JoystickInformation.hpp"
#include "io/JoystickSignal.hpp"
#include "io/JoystickState.hpp"
#include "io/LogFile.hpp"
#include "io/LogFileBase.hpp"
#include "io/LogFileData.hpp"
#include "io/LogFileDataArray.hpp"
#include "io/LogFileDataString.hpp"
#include "io/LogFileSignal.hpp"
#include "io/Logger.hpp"
#include "io/LoggerAction.hpp"
#include "io/LoggerSignal.hpp"
#include "io/NetworkSocket.hpp"
#include "io/NetworkSocketSignal.hpp"
#include "io/SerializableData.hpp"
#include "io/SerializableDataString.hpp"
#include "io/encoding.hpp"
#include "io/filesystem.hpp"
#include "io/serialization.hpp"

/** \defgroup broccoli_io_module broccoli IO Module
  * The *io* module provides classes and methods for abstracting input- and output-operations.
  * IO-operations for reading/writing files, logging, binary serialization and network
  * transfer are provided. Furthermore interaction with external devices (e.g. joysticks /
  * game controllers) is implemented.
  * Include all functionalities of this module with
  * \code
  * #include <broccoli/io.hpp>
  * \endcode
  *
  * In order to keep compilation time low it is suggested that you include only the files you really need:
  *
  * General
  * -------
  * File | Description
  * ---- | -----------
  * \code #include "io/filesystem.hpp" \endcode | Defines \ref broccoli::io::filesystem - \copybrief broccoli::io::filesystem
  * \code #include "io/encoding.hpp" \endcode | Defines \ref broccoli::io::encoding - \copybrief broccoli::io::encoding
  * \code #include "io/serialization.hpp" \endcode | Defines \ref broccoli::io::serialization - \copybrief broccoli::io::serialization
  * \code #include "io/SerializableData.hpp" \endcode | Defines \ref broccoli::io::SerializableData - \copybrief broccoli::io::SerializableData
  * \code #include "io/SerializableDataString.hpp" \endcode | Defines \ref broccoli::io::SerializableDataString - \copybrief broccoli::io::SerializableDataString
  *
  * Joystick
  * --------
  * File | Description
  * ---- | -----------
  * \code #include "io/JoystickSignal.hpp" \endcode | Defines \ref broccoli::io::JoystickSignal - \copybrief broccoli::io::JoystickSignal
  * \code #include "io/JoystickInformation.hpp" \endcode | Defines \ref broccoli::io::JoystickInformation - \copybrief broccoli::io::JoystickInformation
  * \code #include "io/JoystickState.hpp" \endcode | Defines \ref broccoli::io::JoystickState - \copybrief broccoli::io::JoystickState
  * \code #include "io/JoystickEvent.hpp" \endcode | Defines \ref broccoli::io::JoystickEvent - \copybrief broccoli::io::JoystickEvent
  * \code #include "io/JoystickAction.hpp" \endcode | Defines \ref broccoli::io::JoystickAction - \copybrief broccoli::io::JoystickAction
  * \code #include "io/Joystick.hpp" \endcode | Defines \ref broccoli::io::Joystick - \copybrief broccoli::io::Joystick
  *
  * Logging
  * -------
  * File | Description
  * ---- | -----------
  * \code #include "io/LogFileData.hpp" \endcode | Defines \ref broccoli::io::LogFileData - \copybrief broccoli::io::LogFileData
  * \code #include "io/LogFileDataString.hpp" \endcode | Defines \ref broccoli::io::LogFileDataString - \copybrief broccoli::io::LogFileDataString
  * \code #include "io/LogFileDataArray.hpp" \endcode | Defines \ref broccoli::io::LogFileDataArray - \copybrief broccoli::io::LogFileDataArray
  * \code #include "io/LogFileSignal.hpp" \endcode | Defines \ref broccoli::io::LogFileSignal - \copybrief broccoli::io::LogFileSignal
  * \code #include "io/LogFileBase.hpp" \endcode | Defines \ref broccoli::io::LogFileBase - \copybrief broccoli::io::LogFileBase
  * \code #include "io/LogFile.hpp" \endcode | Defines \ref broccoli::io::LogFile - \copybrief broccoli::io::LogFile
  * \code #include "io/LoggerSignal.hpp" \endcode | Defines \ref broccoli::io::LoggerSignal - \copybrief broccoli::io::LoggerSignal
  * \code #include "io/LoggerAction.hpp" \endcode | Defines \ref broccoli::io::LoggerAction - \copybrief broccoli::io::LoggerAction
  * \code #include "io/Logger.hpp" \endcode | Defines \ref broccoli::io::Logger - \copybrief broccoli::io::Logger
  *
  * Network
  * -------
  * File | Description
  * ---- | -----------
  * \code #include "io/NetworkSocketSignal.hpp" \endcode | Defines \ref broccoli::io::NetworkSocketSignal - \copybrief broccoli::io::NetworkSocketSignal
  * \code #include "io/NetworkSocket.hpp" \endcode | Defines \ref broccoli::io::NetworkSocket - \copybrief broccoli::io::NetworkSocket
  */

// Documentation of namespace
namespace broccoli {
//! Namespace of \ref broccoli_io_module
namespace io {
}
}
