/*
 * This file is part of broccoli.
 * Copyright (C) 2019 Chair of Applied Mechanics, Technical University of Munich
 * https://www.amm.mw.tum.de/
 */

#pragma once // Load this file only once

#include "LogFileData.hpp"
#include "encoding.hpp"
#include <array>

namespace broccoli {
namespace io {
    //! Container template for log data which can be written as array
    /*!
     * Spezialization of generic \ref LogFileData. May be used as reference implementation to
     * derive from the base class \ref LogFileData. The members are represented by a single
     * array of elements which will be encoded one after the other for each dataset. In the
     * default configuration, the logfile will have the following tabular structure:
     *
     * \code
     * <Header>
     * ...
     * ...            element-delimiter               array-delimiter
     * ...                   |                             |
     * ...   ------------------------------- ...           |
     * ...   |      |      |      |      |                 |
     * ...   v      v      v      v      v                 v
     * +1.234 -5.678 +1.234 -5.678 +1.234 -5.678 ... -5.678    <--     i-th LogFileDataArray instance
     * +1.234 -5.678 +1.234 -5.678 +1.234 -5.678 ... -5.678    <-- (i+1)-th LogFileDataArray instance
     * ...
     * ...
     * <Footer>
     * \endcode
     *
     * \par Configuration
     *
     * Template parameters are used to specialize the class:
     * - \p ElementType specifies the datatype for the elements in the array (e.g. `double`)
     * - \p Dimension specifies the count of elements in the array
     *
     * Furthermore reimplementing the following predefined virtual functions allows to specify the output further:
     * - \ref encodeElement()
     * - \ref encodeElementDelimiter()
     * - \ref encodeArrayDelimiter()
     */
    template <class ElementType, unsigned int Dimension>
    class LogFileDataArray : public LogFileData {
    public:
        //! Default constructor
        LogFileDataArray()
        {
        }

        //! Destructor
        virtual ~LogFileDataArray()
        {
        }

        // Members
        std::array<ElementType, Dimension> m_data; //!< Array of data elements of the specified \p ElementType

        // Encodes member data to logging stream (see base class for details)
        virtual LogStreamSize encodeDataToLogStream(LogStream& stream) const
        {
            // Initialize helpers
            LogStreamSize addedElements = 0;

            // Iterate over all elements
            for (size_t i = 0; i < m_data.size(); i++) {
                // Write element data
                addedElements += encodeElement(stream, i);

                // Write delimiters
                if (i < m_data.size() - 1) {
                    // Write element delimiter
                    addedElements += encodeElementDelimiter(stream, i);
                } else {
                    // Write array delimiter
                    addedElements += encodeArrayDelimiter(stream);
                }
            }

            // Pass back count of elements added to the stream
            return addedElements;
        }

    protected:
        //! Encoding of single elments
        /*!
         * Uses the default encoding method for the specified \p ElementType.
         *
         * \param [in,out] stream The logging stream to which the element should be appended to.
         * \param [in] index Index of the element to encode
         * \return Count of appended characters (elements added to stream).
         */
        virtual LogStreamSize encodeElement(LogStream& stream, const size_t& index) const
        {
            // Index unused per default (same format for all elements)
            (void)index;

            // Use default encoding for the specified data type
            return encoding::encode(stream, m_data[index]);
        }

        //! Encoding of element delimiter
        /*!
         * Uses a single whitespace character by default.
         *
         * \param [in,out] stream The logging stream to which the element delimiter should be appended to.
         * \param [in] index Index of the **preceding** element
         * \return Count of appended characters (elements added to stream).
         */
        virtual LogStreamSize encodeElementDelimiter(LogStream& stream, const size_t& index) const
        {
            // Index unused per default (same delimiter for all elements)
            (void)index;

            // Use whitespace character by default
            return encoding::encode(stream, (char)' ');
        }

        //! Encoding of array delimiter
        /*!
         * Uses a single newline character by default.
         *
         * \param [in,out] stream The logging stream to which the array delimiter should be appended to.
         * \return Count of appended characters (elements added to stream).
         */
        virtual LogStreamSize encodeArrayDelimiter(LogStream& stream) const
        {
            // Use newline character by default
            return encoding::encode(stream, (char)'\n');
        }
    };
} // namespace io
} // namespace broccoli
