/*
 * This file is part of broccoli.
 * Copyright (C) 2019 Chair of Applied Mechanics, Technical University of Munich
 * https://www.amm.mw.tum.de/
 */

#pragma once // Load this file only once

#include "serialization.hpp"
#ifdef HAVE_EIGEN3
#include <Eigen/StdVector> // Necessary for 128bit alginnment and STL containers (for vectorization)
#endif // HAVE_EIGEN3

namespace broccoli {
namespace io {
    //! Abstract base class for serializable data containers
    /*!
     * This class is intended to be used as base class for data exchange containers.
     *
     * One has to implement the functions:
     * - \ref serializePayload()
     * - \ref deSerializePayload()
     *
     * The functions \ref serializePayload() and \ref deSerializePayload() can be used by a background thread for efficient
     * multi-threaded (de-)serialization of the payload, where "payload" describes the actual data, i.e. the container members.
     * Note that \ref serializePayload() and \ref deSerializePayload() are indirectly triggered by calling \ref serialize()
     * and \ref deSerialize(). This automatically handles a **leading header** which contains the payload size. The header is
     * necessary to determine the end of a serialized data object in a continous data stream.
     *
     * \par Byte Mapping:
     * Size | Content
     * ---- | -------
     * \ref serialization::BinaryStreamSize | Payload size in bytes (=header)
     * ... | Actual payload, i.e. member data
     */
    class SerializableData {
    public:
        //! Constructor
        SerializableData()
        {
        }

        //! Destructor
        virtual ~SerializableData()
        {
        }

        //! Serialization of complete data container (header + payload)
        /*!
         * Serializes header and payload and appends both to the given stream.
         * \param [in,out] stream Stream to be extended by the header and payload.
         * \return Size of the binary represenation of header and payload (count of bytes added to stream).
         */
        virtual serialization::BinaryStreamSize serialize(serialization::BinaryStream& stream) const
        {
            // Remember starting index of header (to overwrite it later on)
            serialization::BinaryStreamSize headerStartIndex = stream.size();

            // Add dummy header (containing wrong information about size) (used as placeholder)
            serialization::serialize(stream, (serialization::BinaryStreamSize)0);

            // Add actual payload and remember its real size
            serialization::BinaryStreamSize realPayloadSize = serializePayload(stream);

            // Create auxillary binary stream only for header and write real size of payload to it
            serialization::BinaryStream headerStream;
            serialization::serialize(headerStream, (serialization::BinaryStreamSize)realPayloadSize);

            // Overwrite dummy header with correct information for payload size
            for (uint64_t i = 0; i < sizeof(serialization::BinaryStreamSize); i++)
                stream[headerStartIndex + i] = headerStream[i];

            // Return count of added bytes
            return sizeof(serialization::BinaryStreamSize) + realPayloadSize;
        }

        //! Deserialization of complete data container (header + payload)
        /*!
         * Deserializes header and payload from a given binary stream.
         * \param [in] stream The stream from which the data should be extracted.
         * \param [in] index The starting index of the data in the stream.
         * \return Size of binary representation (header + payload) in case of success or 0 on failure
         */
        virtual serialization::BinaryStreamSize deSerialize(const serialization::BinaryStream& stream, const serialization::BinaryStreamSize& index)
        {
            // Extract payload size
            serialization::BinaryStreamSize payloadSize = 0;
            if (serialization::deSerialize(stream, index, payloadSize) == 0) // Fails, if stream is not big enough for header
                return 0;

            // Check, if stream is big enough for header and payload
            serialization::BinaryStreamSize targetSize = sizeof(serialization::BinaryStreamSize) + payloadSize;
            if (stream.size() < index + targetSize)
                return 0;

            // Extract payload (throw error, if reported size does not match size in header)
            if (deSerializePayload(stream, index + sizeof(serialization::BinaryStreamSize), payloadSize) != payloadSize)
                return 0;

            // Return size of binary representation (header + payload) within stream
            return targetSize;
        }

    protected:
        //! Serialization of payload
        /*!
         * Encodes all members (=payload) and appends them to a given binary stream.
         * \return Size of the binary represenation of payload (count of bytes added to stream).
         */
        virtual serialization::BinaryStreamSize serializePayload(serialization::BinaryStream& stream) const = 0;

        //! Deserialization of payload
        /*!
         * Decodes given binary stream to get member data (payload)
         * \param [in] stream The stream from which the payload should be extracted.
         * \param [in] index The starting index of the payload in the stream.
         * \param [in] payloadSize Exact size of the payload data (important for decoding complex data containers of varying size - may be ignored for simple data containers).
         * \return Size of binary representation of the payload in case of success or 0 on failure
         */
        virtual serialization::BinaryStreamSize deSerializePayload(const serialization::BinaryStream& stream, const serialization::BinaryStreamSize& index, const serialization::BinaryStreamSize& payloadSize) = 0;

#ifdef HAVE_EIGEN3
    public:
        EIGEN_MAKE_ALIGNED_OPERATOR_NEW // <-- Proper 128 bit alignment of member data necessary for Eigen vectorization
#endif // HAVE_EIGEN3
    };
} // namespace io
} // namespace broccoli
