/*
 * This file is part of broccoli.
 * Copyright (C) 2019 Chair of Applied Mechanics, Technical University of Munich
 * https://www.amm.mw.tum.de/
 */

#pragma once // Load this file only once

#include <sstream>
#include <stdint.h>
#include <stdio.h>
#include <string>
#include <vector>

namespace broccoli {
namespace io {
    //! Methods for encoding and decoding basic data types to/from **character streams**
    /*!
     * This collection of functions is used to encode and decode data to/from **character streams**.
     *
     * \par Notes on Efficiency
     * Character streams are represented by `std::vectors` which are of dynamic size (allocated on heap).
     * This is necessary, since in some cases one does not know the final size of the stream in advance.
     * Unfortunately dynamic memory allocation is slow. However to get the maximum performance we use
     * `std::vector::reserve()` to allocate memory in advance in cases where we know how big the new data is.
     */
    namespace encoding {
        // Type definitions
        typedef std::vector<char> CharacterStream; //!< Specifies the used type for storing character stream data
        typedef size_t CharacterStreamSize; //!< Specifies the used type for describing the size of a character stream

        // Encoding
        // --------
        //! Appends specified character to a given stream
        /*!
         * \param [in,out] stream The stream to which the character should be appended to.
         * \param [in] character The character to append.
         * \return Count of appended characters (elements in stream).
         */
        static inline CharacterStreamSize encode(CharacterStream& stream, const char& character)
        {
            // Push single character to end of stream
            stream.push_back(character);

            // Pass back length of added data
            return 1;
        }

        //! Encodes specified unsigned integer value and appends it to a given stream
        /*!
         * \param [in,out] stream The stream to which the value should be appended to.
         * \param [in] value The value to encode.
         * \param [in] format C-format specifier.
         * \return Count of appended characters (elements in stream).
         */
        static inline CharacterStreamSize encode(CharacterStream& stream, const uint64_t& value, const std::string& format = "%llu")
        {
            // Initialize helpers
            CharacterStreamSize oldSize = stream.size(); // Remember old size of stream
            CharacterStreamSize maxLengthNewData = 20; // Define maximum length of new data (without null-termination)

            // Fill up stream with placeholders (FAST, if stream capacity is already big enough -> no memory allocation necessary)
            stream.resize(oldSize + maxLengthNewData + 1, 0); // +1 since '\0' (null-termination) is always written to the end of the new data

            // Write data and remember its total length
#ifdef __QNXNTO__
            CharacterStreamSize realLengthNewData = sprintf(&stream[oldSize], format.c_str(), value);
#else
            CharacterStreamSize realLengthNewData = snprintf(&stream[oldSize], stream.size() - oldSize, format.c_str(), value);
#endif // __QNXNTO__

            // Shrink stream to its actual size (remove placeholders and null-termination)
            stream.resize(oldSize + realLengthNewData);

            // Pass back length of added data
            return realLengthNewData;
        }

        //! Encodes specified signed integer value and appends it to a given stream
        /*!
         * \param [in,out] stream The stream to which the value should be appended to.
         * \param [in] value The value to encode.
         * \param [in] format C-format specifier.
         * \return Count of appended characters (elements in stream).
         */
        static inline CharacterStreamSize encode(CharacterStream& stream, const int64_t& value, const std::string& format = "%+lld")
        {
            // Initialize helpers
            CharacterStreamSize oldSize = stream.size(); // Remember old size of stream
            CharacterStreamSize maxLengthNewData = 20; // Define maximum length of new data (without null-termination)

            // Fill up stream with placeholders (FAST, if stream capacity is already big enough -> no memory allocation necessary)
            stream.resize(oldSize + maxLengthNewData + 1, 0); // +1 since '\0' (null-termination) is always written to the end of the new data

            // Write data and remember its total length
#ifdef __QNXNTO__
            CharacterStreamSize realLengthNewData = sprintf(&stream[oldSize], format.c_str(), value);
#else
            CharacterStreamSize realLengthNewData = snprintf(&stream[oldSize], stream.size() - oldSize, format.c_str(), value);
#endif // __QNXNTO__

            // Shrink stream to its actual size (remove placeholders and null-termination)
            stream.resize(oldSize + realLengthNewData);

            // Pass back length of added data
            return realLengthNewData;
        }

        //! Encodes specified floating point value and appends it to a given stream
        /*!
         * \param [in,out] stream The stream to which the value should be appended to.
         * \param [in] value The value to encode.
         * \param [in] format C-format specifier.
         * \return Count of appended characters (elements in stream).
         */
        static inline CharacterStreamSize encode(CharacterStream& stream, const double& value, const std::string& format = "%+.16e")
        {
            // Initialize helpers
            CharacterStreamSize oldSize = stream.size(); // Remember old size of stream
            CharacterStreamSize maxLengthNewData = 24; // Define maximum length of new data (without null-termination)

            // Fill up stream with placeholders (FAST, if stream capacity is already big enough -> no memory allocation necessary)
            stream.resize(oldSize + maxLengthNewData + 1, 0); // +1 since '\0' (null-termination) is always written to the end of the new data

            // Write data and remember its total length
#ifdef __QNXNTO__
            CharacterStreamSize realLengthNewData = sprintf(&stream[oldSize], format.c_str(), value);
#else
            CharacterStreamSize realLengthNewData = snprintf(&stream[oldSize], stream.size() - oldSize, format.c_str(), value);
#endif // __QNXNTO__

            // Shrink stream to its actual size (remove placeholders and null-termination)
            stream.resize(oldSize + realLengthNewData);

            // Pass back length of added data
            return realLengthNewData;
        }

        //! Appends specified string to a given stream
        /*!
         * \warning Does **not** add the null-termination character to the stream!
         *
         * \param [in,out] stream The stream to which the string should be appended to.
         * \param [in] string The string to append.
         * \return Count of appended characters (elements in stream).
         */
        static inline CharacterStreamSize encode(CharacterStream& stream, const std::string& string)
        {
            // Directly copy string data to stream (FAST, if stream capacity is already big enough -> no memory allocation necessary)
            std::copy(string.begin(), string.end(), std::back_inserter(stream));

            // Pass back length of added data
            return string.size();
        }

        // Decoding
        // --------
        //! Decodes a single character from the given stream
        /*!
         * \param [in] stream The stream from which the value should be extracted.
         * \param [in] index The starting index of the encoded value in the stream (=index of the **first element**).
         * \param [out] value Reference to returned value. The result of the decoding will be stored in this parameter.
         * \return Size of encoded value in case of success or 0 on failure
         */
        static inline CharacterStreamSize decode(const CharacterStream& stream, const CharacterStreamSize& index, char& value)
        {
            // Check, if stream is big enough
            if (stream.size() < index + 1)
                return 0; // Error -> stream is not big enough

            // Extract character from stream
            value = stream[index];

            // Return size of encoded value in stream
            return 1;
        }

        //! Decodes an unsigned 64-bit integer from the given stream
        /*!
         * \param [in] stream The stream from which the value should be extracted.
         * \param [in] index The starting index of the encoded value in the stream (=index of the **first element**).
         * \param [in] length The length of the encoded value in the stream (count of characters).
         * \param [out] value Reference to returned value. The result of the decoding will be stored in this parameter.
         * \return Size of encoded value in case of success or 0 on failure
         */
        static inline CharacterStreamSize decode(const CharacterStream& stream, const CharacterStreamSize& index, const CharacterStreamSize& length, uint64_t& value)
        {
            // Check if length is valid
            if (length == 0)
                return 0; // Error -> invalid length!

            // Check, if stream is big enough
            if (stream.size() < index + length)
                return 0; // Error -> stream is not big enough

            // Decode (catch low level exceptions)
            try {
#ifdef __QNXNTO__
                if (!(std::istringstream(std::string(stream.begin() + index, stream.begin() + (index + length))) >> value))
                    return 0; // Error -> conversion failed
#else
                value = std::stoull(std::string(stream.begin() + index, stream.begin() + (index + length)));
#endif // __QNXNTO__
            } catch (...) {
                return 0; // Error -> Decoding went wrong
            }

            // Return size of encoded value in stream
            return length;
        }

        //! Decodes a signed 64-bit integer from the given stream
        /*!
         * \param [in] stream The stream from which the value should be extracted.
         * \param [in] index The starting index of the encoded value in the stream (=index of the **first element**).
         * \param [in] length The length of the encoded value in the stream (count of characters).
         * \param [out] value Reference to returned value. The result of the decoding will be stored in this parameter.
         * \return Size of encoded value in case of success or 0 on failure
         */
        static inline CharacterStreamSize decode(const CharacterStream& stream, const CharacterStreamSize& index, const CharacterStreamSize& length, int64_t& value)
        {
            // Check if length is valid
            if (length == 0)
                return 0; // Error -> invalid length!

            // Check, if stream is big enough
            if (stream.size() < index + length)
                return 0; // Error -> stream is not big enough

            // Decode (catch low level exceptions)
            try {
#ifdef __QNXNTO__
                if (!(std::istringstream(std::string(stream.begin() + index, stream.begin() + (index + length))) >> value))
                    return 0; // Error -> conversion failed
#else
                value = std::stoll(std::string(stream.begin() + index, stream.begin() + (index + length)));
#endif // __QNXNTO__
            } catch (...) {
                return 0; // Error -> Decoding went wrong
            }

            // Return size of encoded value in stream
            return length;
        }

        //! Decodes a floating point number from the given stream
        /*!
         * \param [in] stream The stream from which the value should be extracted.
         * \param [in] index The starting index of the encoded value in the stream (=index of the **first element**).
         * \param [in] length The length of the encoded value in the stream (count of characters).
         * \param [out] value Reference to returned value. The result of the decoding will be stored in this parameter.
         * \return Size of encoded value in case of success or 0 on failure
         */
        static inline CharacterStreamSize decode(const CharacterStream& stream, const CharacterStreamSize& index, const CharacterStreamSize& length, double& value)
        {
            // Check if length is valid
            if (length == 0)
                return 0; // Error -> invalid length!

            // Check, if stream is big enough
            if (stream.size() < index + length)
                return 0; // Error -> stream is not big enough

            // Decode (catch low level exceptions)
            try {
#ifdef __QNXNTO__
                if (!(std::istringstream(std::string(stream.begin() + index, stream.begin() + (index + length))) >> value))
                    return 0; // Error -> conversion failed
#else
                value = std::stod(std::string(stream.begin() + index, stream.begin() + (index + length)));
#endif // __QNXNTO__
            } catch (...) {
                return 0; // Error -> Decoding went wrong
            }

            // Return size of encoded value in stream
            return length;
        }

        //! Decodes a string from the given stream
        /*!
         * \param [in] stream The stream from which the value should be extracted.
         * \param [in] index The starting index of the encoded value in the stream (=index of the **first element**).
         * \param [in] length The length of the encoded value in the stream (count of characters).
         * \param [out] value Reference to returned value. The result of the decoding will be stored in this parameter.
         * \return Size of encoded value in case of success or 0 on failure
         */
        static inline CharacterStreamSize decode(const CharacterStream& stream, const CharacterStreamSize& index, const CharacterStreamSize& length, std::string& value)
        {
            // Check if length is valid
            if (length == 0)
                return 0; // Error -> invalid length!

            // Check, if stream is big enough
            if (stream.size() < index + length)
                return 0; // Error -> stream is not big enough

            // Extract string from stream
            value = std::string(stream.begin() + index, stream.begin() + (index + length));

            // Return size of encoded value in stream
            return length;
        }
    } // namespace encoding
} // namespace io
} // namespace broccoli
