/*
 * This file is part of broccoli.
 * Copyright (C) 2019 Chair of Applied Mechanics, Technical University of Munich
 * https://www.amm.mw.tum.de/
 */

#pragma once // Load this file only once

#include <fstream>
#include <iostream>
#include <stdio.h>
#include <stdlib.h>
#include <string>
#include <sys/stat.h>
#include <sys/types.h>
#include <unistd.h>

namespace broccoli {
namespace io {
    //! Collection of methods related to the filesystem
    /*!
     * This collection of functions provides convenience methods for dealing with the file system.
    */
    namespace filesystem {
        //! Checks, if a specified file exists
        /*!
         * \remark In case the requested information cannot be obtained, `false` is returned (assume that it does not exist).
         *
         * \remark In case of an error the underlying lowlevel functions set the variable `errno` which can be used to further
         * classify the error.
         *
         * \param [in] filePath Path to file to test (absolute path or relative to executable path).
         * \return `true` if specified file exists, `false` if not (or path corresponds not to a file but a directory instead)
         */
        static inline bool fileExists(const std::string& filePath)
        {
            // Catch lowlevel exceptions
            try {
                // Try to obtain info about file system element
                struct ::stat elementInfo;
                if (::stat(filePath.c_str(), &elementInfo) != 0) {
                    // ...could not obtain info -> file does not exist
                    return false;
                }

                // We got the element info -> check if element is a regular file
                if ((elementInfo.st_mode & S_IFMT) == S_IFREG)
                    return true; // Element is a regular file
                else
                    return false; // Element is not a regular file (maybe directory, socket, etc. )
            } catch (...) {
                return false; // ...could not obtain info -> something went wrong
            }
        }

        //! Deletes the specified file
        /*!
         * \remark In case of an error the underlying lowlevel functions set the variable `errno` which can be used to further
         * classify the error.
         *
         * \param [in] filePath Path to file to remove (absolute path or relative to executable path).
         * \return `true` if specified file was removed, `false` if an error occured
         */
        static inline bool removeFile(const std::string& filePath)
        {
            // Catch lowlevel exceptions
            try {
                // Try to remove file
                if (::remove(filePath.c_str()) != 0)
                    return false; // ...something went wrong
                else
                    return true; // ...success!
            } catch (...) {
                return false; // ...something went wrong
            }
        }

        //! Checks, if a specified directory exists
        /*!
         * \remark In case the requested information cannot be obtained, `false` is returned (assume that it does not exist).
         *
         * \remark In case of an error the underlying lowlevel functions set the variable `errno` which can be used to further
         * classify the error.
         *
         * \param [in] directoryPath Path to directory to test (absolute path or relative to executable path).
         * \return `true` if specified directory exists, `false` if not (or path corresponds not to a directory but a file instead)
         */
        static inline bool directoryExists(const std::string& directoryPath)
        {
            // Catch lowlevel exceptions
            try {
                // Try to obtain info about file system element
                struct ::stat elementInfo;
                if (::stat(directoryPath.c_str(), &elementInfo) != 0) {
                    // ...could not obtain info -> directory does not exist
                    return false;
                }

                // We got the element info -> check if element is a directory
                if ((elementInfo.st_mode & S_IFMT) == S_IFDIR)
                    return true; // Element is a directory
                else
                    return false; // Element is not a directory (maybe file, socket, etc. )
            } catch (...) {
                return false; // ...could not obtain info -> something went wrong
            }
        }

        //! Tries to create a directory in the specified path
        /*!
         * \attention The parent directory has to exist for this method to work!
         *
         * \remark In case of an error the underlying lowlevel functions set the variable `errno` which can be used to further
         * classify the error.
         *
         * \param [in] directoryPath Path to directory to create (absolute path or relative to executable path).
         * \param [in] accessMode Permission of directory
         * \return `true` if specified directory was created, `false` if not
         */
        static inline bool createDirectory(const std::string& directoryPath, const mode_t& accessMode = S_IRWXU | S_IRWXG | S_IRWXO)
        {
            // Catch lowlevel exceptions
            try {
                if (::mkdir(directoryPath.c_str(), accessMode) != 0)
                    return false; // Could not create directory (error during creation or directory already exists)
                else
                    return true; // Directory successfully created
            } catch (...) {
                return false; // ...something went wrong
            }
        }

        //! Deletes the specified **empty** directory
        /*!
         * \warning The specified directory has to be **empty** for this method to work!
         *
         * \remark In case of an error the underlying lowlevel functions set the variable `errno` which can be used to further
         * classify the error.
         *
         * \param [in] directoryPath Path to directory to remove (absolute path or relative to executable path).
         * \return `true` if specified directory was removed, `false` if an error occured
         */
        static inline bool removeDirectory(const std::string& directoryPath)
        {
            // Catch lowlevel exceptions
            try {
                // Try to remove directory
                if (::rmdir(directoryPath.c_str()) != 0)
                    return false; // ...something went wrong
                else
                    return true; // ...success!
            } catch (...) {
                return false; // ...something went wrong
            }
        }
    } // namespace filesystem
} // namespace io
} // namespace broccoli
