/*
 * This file is part of broccoli.
 * Copyright (C) 2019 Chair of Applied Mechanics, Technical University of Munich
 * https://www.amm.mw.tum.de/
 */

#pragma once // Load this file only once

#include "../core/Time.hpp"
#include <array>
#include <cstring> // For using memcpy (float and double conversion only)
#include <stdint.h>
#include <string>
#include <vector>

namespace broccoli {
namespace io {
    //! Methods for serializing and deserializing basic data types to/from **binary streams**
    /*!
     * This collection of functions is used to serialize non POD (<b>P</b>lain-<b>O</b>ld-<b>D</b>ata)
     * structures to/from **binary streams** containing complex types (strings, vectors, ...). A
     * simple `memcpy` would not work here as it would only copy the addresses of the pointers. Note
     * that bitwise operations do **not** depend on endianess, thus the following code works for both
     * big- and little-endian platforms. However the current implementation of float and double
     * conversions assumes that the machine uses IEEE 754 notation for floating point numbers. Further
     * it is assumed, that floats are 32-bit and doubles are 64-bit.
     *
     * \par Notes on Efficiency
     * Binary streams are represented by `std::vectors` which are of dynamic size (allocated on heap).
     * This is necessary, since in some cases one does not know the final size of the stream in advance.
     * Unfortunately dynamic memory allocation is slow. However to get the maximum performance we use
     * `std::vector::reserve()` to allocate memory in advance in cases where we know how big the new data is.
     */
    namespace serialization {
        // Type definitions
        typedef std::vector<uint8_t> BinaryStream; //!< Specifies the used type for storing binary stream data
        typedef uint32_t BinaryStreamSize; //!< Specifies the used type for describing the size of a binary stream (will be packed into messages in case of data hierarchies)

        // Serialization
        // -------------
        //! Append binary stream to another binary stream
        /*!
         * \param [in,out] stream Stream which should be extended by "streamToAppend".
         * \param [in] streamToAppend Stream which is appended to "stream". "streamToAppend" is **not** changed (true copy).
         * \return Count of elements added to binary stream.
         */
        static inline BinaryStreamSize serialize(BinaryStream& stream, const BinaryStream& streamToAppend)
        {
            // Allocate memory first for maximum efficiency (does nothing if capacity of stream is already big enough)
            stream.reserve(stream.size() + streamToAppend.size());

            // Copy data into new allocated memory
            stream.insert(stream.end(), streamToAppend.begin(), streamToAppend.end());

            // Return count of added bytes
            return streamToAppend.size();
        }

        //! Encodes a boolean to its binary represenation
        /*!
         * \param [in,out] stream Reference to stream we want to extend by the value.
         * \param [in] value Value to serialize.
         * \return Size of the binary represenation of the value (count of bytes added to stream).
         */
        static inline BinaryStreamSize serialize(BinaryStream& stream, const bool& value)
        {
            // Allocate additional memory (does nothing if capacity of stream is already big enough)
            stream.reserve(stream.size() + sizeof(uint8_t));

            // Add value to stream
            stream.push_back((uint8_t)value); // Converts boolean to uint8_t first (stored in 1 byte (smallest unit))

            // Return count of added bytes
            return sizeof(uint8_t);
        }

        //! Encodes a 8-bit unsigned integer to its binary represenation
        /*!
         * \param [in,out] stream Reference to stream we want to extend by the value.
         * \param [in] value Value to serialize.
         * \return Size of the binary represenation of the value (count of bytes added to stream).
         */
        static inline BinaryStreamSize serialize(BinaryStream& stream, const uint8_t& value)
        {
            // Allocate additional memory (does nothing if capacity of stream is already big enough)
            stream.reserve(stream.size() + sizeof(value));

            // Add value to stream
            stream.push_back((uint8_t)value);

            // Return count of added bytes
            return sizeof(value);
        }

        //! Encodes a 16-bit unsigned integer to its binary represenation
        /*!
         * \param [in,out] stream Reference to stream we want to extend by the value.
         * \param [in] value Value to serialize.
         * \return Size of the binary represenation of the value (count of bytes added to stream).
         */
        static inline BinaryStreamSize serialize(BinaryStream& stream, const uint16_t& value)
        {
            // Allocate additional memory (does nothing if capacity of stream is already big enough)
            stream.reserve(stream.size() + sizeof(value));

            // Add value to stream
            stream.push_back((uint8_t)(value >> 8));
            stream.push_back((uint8_t)(value & 0xFF));

            // Return count of added bytes
            return sizeof(value);
        }

        //! Encodes a 32-bit unsigned integer to its binary represenation
        /*!
         * \param [in,out] stream Reference to stream we want to extend by the value.
         * \param [in] value Value to serialize.
         * \return Size of the binary represenation of the value (count of bytes added to stream).
         */
        static inline BinaryStreamSize serialize(BinaryStream& stream, const uint32_t& value)
        {
            // Allocate additional memory (does nothing if capacity of stream is already big enough)
            stream.reserve(stream.size() + sizeof(value));

            // Add value to stream
            stream.push_back((uint8_t)(value >> 24));
            stream.push_back((uint8_t)((value >> 16) & 0xFF));
            stream.push_back((uint8_t)((value >> 8) & 0xFF));
            stream.push_back((uint8_t)(value & 0xFF));

            // Return count of added bytes
            return sizeof(value);
        }

        //! Encodes a 64-bit unsigned integer to its binary represenation
        /*!
         * \param [in,out] stream Reference to stream we want to extend by the value.
         * \param [in] value Value to serialize.
         * \return Size of the binary represenation of the value (count of bytes added to stream).
         */
        static inline BinaryStreamSize serialize(BinaryStream& stream, const uint64_t& value)
        {
            // Allocate additional memory (does nothing if capacity of stream is already big enough)
            stream.reserve(stream.size() + sizeof(value));

            // Add value to stream
            stream.push_back((uint8_t)(value >> 56));
            stream.push_back((uint8_t)((value >> 48) & 0xFF));
            stream.push_back((uint8_t)((value >> 40) & 0xFF));
            stream.push_back((uint8_t)((value >> 32) & 0xFF));
            stream.push_back((uint8_t)((value >> 24) & 0xFF));
            stream.push_back((uint8_t)((value >> 16) & 0xFF));
            stream.push_back((uint8_t)((value >> 8) & 0xFF));
            stream.push_back((uint8_t)(value & 0xFF));

            // Return count of added bytes
            return sizeof(value);
        }

        //! Encodes a 8-bit signed integer to its binary represenation
        /*!
         * \remark Bit-shift of negative integers is undefined, thus we have to handle the sign bit on our own!
         *
         * \param [in,out] stream Reference to stream we want to extend by the value.
         * \param [in] value Value to serialize.
         * \return Size of the binary represenation of the value (count of bytes added to stream).
         */
        static inline BinaryStreamSize serialize(BinaryStream& stream, const int8_t& value)
        {
            // Check sign
            if (value < 0) {
                // ...negative value
                BinaryStreamSize mostSignificantByte = stream.size(); // Get byte which should contain the sign bit (=first new added byte)
                (void)serialize(stream, (uint8_t)-value); // Serialize unsigned value
                stream[mostSignificantByte] = stream[mostSignificantByte] | 0x80; // Add sign-bit (0b10000000)
            } else {
                // ...positive value
                (void)serialize(stream, (uint8_t)value);
            }

            // Return count of added bytes
            return sizeof(value);
        }

        //! Encodes a 16-bit signed integer to its binary represenation
        /*!
         * \remark Bit-shift of negative integers is undefined, thus we have to handle the sign bit on our own!
         *
         * \param [in,out] stream Reference to stream we want to extend by the value.
         * \param [in] value Value to serialize.
         * \return Size of the binary represenation of the value (count of bytes added to stream).
         */
        static inline BinaryStreamSize serialize(BinaryStream& stream, const int16_t& value)
        {
            // Check sign
            if (value < 0) {
                // ...negative value
                BinaryStreamSize mostSignificantByte = stream.size(); // Get byte which should contain the sign bit (=first new added byte)
                (void)serialize(stream, (uint16_t)-value); // Serialize unsigned value
                stream[mostSignificantByte] = stream[mostSignificantByte] | 0x80; // Add sign-bit (0b10000000)
            } else {
                // ...positive value
                (void)serialize(stream, (uint16_t)value);
            }

            // Return count of added bytes
            return sizeof(value);
        }

        //! Encodes a 32-bit signed integer to its binary represenation
        /*!
         * \remark Bit-shift of negative integers is undefined, thus we have to handle the sign bit on our own!
         *
         * \param [in,out] stream Reference to stream we want to extend by the value.
         * \param [in] value Value to serialize.
         * \return Size of the binary represenation of the value (count of bytes added to stream).
         */
        static inline BinaryStreamSize serialize(BinaryStream& stream, const int32_t& value)
        {
            // Check sign
            if (value < 0) {
                // ...negative value
                BinaryStreamSize mostSignificantByte = stream.size(); // Get byte which should contain the sign bit (=first new added byte)
                (void)serialize(stream, (uint32_t)-value); // Serialize unsigned value
                stream[mostSignificantByte] = stream[mostSignificantByte] | 0x80; // Add sign-bit (0b10000000)
            } else {
                // ...positive value
                (void)serialize(stream, (uint32_t)value);
            }

            // Return count of added bytes
            return sizeof(value);
        }

        //! Encodes a 64-bit signed integer to its binary represenation
        /*!
         * \remark Bit-shift of negative integers is undefined, thus we have to handle the sign bit on our own!
         *
         * \param [in,out] stream Reference to stream we want to extend by the value.
         * \param [in] value Value to serialize.
         * \return Size of the binary represenation of the value (count of bytes added to stream).
         */
        static inline BinaryStreamSize serialize(BinaryStream& stream, const int64_t& value)
        {
            // Check sign
            if (value < 0) {
                // ...negative value
                BinaryStreamSize mostSignificantByte = stream.size(); // Get byte which should contain the sign bit (=first new added byte)
                (void)serialize(stream, (uint64_t)-value); // Serialize unsigned value
                stream[mostSignificantByte] = stream[mostSignificantByte] | 0x80; // Add sign-bit (0b10000000)
            } else {
                // ...positive value
                (void)serialize(stream, (uint64_t)value);
            }

            // Return count of added bytes
            return sizeof(value);
        }

        //! Encodes a 32-bit floating point number to its binary represenation
        /*!
         * \remark Bit-shift of negative integers is undefined, thus we have to handle the sign bit on our own!
         *
         * \param [in,out] stream Reference to stream we want to extend by the value.
         * \param [in] value Value to serialize.
         * \return Size of the binary represenation of the value (count of bytes added to stream).
         */
        static inline BinaryStreamSize serialize(BinaryStream& stream, const float& value)
        {
            // Convert to 32-bit unsigned integer representation
            uint32_t intRepresentation = 0;
            memcpy((void*)&intRepresentation, (void*)&value, sizeof(value));

            // Serialize as if it was a 32-bit unsigned integer
            return serialize(stream, intRepresentation);
        }

        //! Encodes a 64-bit floating point number to its binary represenation
        /*!
         * \remark Bit-shift of negative integers is undefined, thus we have to handle the sign bit on our own!
         *
         * \param [in,out] stream Reference to stream we want to extend by the value.
         * \param [in] value Value to serialize.
         * \return Size of the binary represenation of the value (count of bytes added to stream).
         */
        static inline BinaryStreamSize serialize(BinaryStream& stream, const double& value)
        {
            // Convert to 64-bit unsigned integer representation
            uint64_t intRepresentation = 0;
            memcpy((void*)&intRepresentation, (void*)&value, sizeof(value));

            // Serialize as if it was a 64-bit unsigned integer
            return serialize(stream, intRepresentation);
        }

        //! Encodes a \ref broccoli::core::Time instance to its binary representation
        /*!
         * Serializes seconds and nanoseconds as two consecutive values.
         * \param [in,out] stream Reference to stream we want to extend by the value.
         * \param [in] value Value to serialize.
         * \return Size of the binary represenation of the value (count of bytes added to stream).
         */
        static inline BinaryStreamSize serialize(BinaryStream& stream, const core::Time& value)
        {
            // Allocate memory for maximum performance (does nothing if capacity of stream is already big enough)
            stream.reserve(stream.size() + sizeof(value.m_seconds) + sizeof(value.m_nanoSeconds));

            // Serialize seconds and nanoseconds
            serialize(stream, value.m_seconds);
            serialize(stream, value.m_nanoSeconds);

            // Return count of added bytes
            return sizeof(value.m_seconds) + sizeof(value.m_nanoSeconds);
        }

        //! Encodes an array (fixed-size N) of basic datatypes \p T to its binary representation
        /*!
         * Serializes all elements of the array (of type \p T) preserving the array order.
         * \param [in,out] stream Reference to stream we want to extend by the array.
         * \param [in] array Array to serialize.
         * \return Size of the binary represenation of the array (count of bytes added to stream).
         */
        template <typename T, size_t N>
        static inline BinaryStreamSize serializeArray(BinaryStream& stream, const std::array<T, N>& array)
        {
            // Allocate memory for maximum performance (does nothing if capacity of stream is already big enough)
            stream.reserve(stream.size() + sizeof(T) * N);

            // Iterate through all elements and serialize them one after the other
            for (size_t i = 0; i < N; i++)
                serialize(stream, array[i]);

            // Return count of added bytes
            return sizeof(T) * N;
        }

        //! Encodes a string to its binary representation
        /*!
         * Adds leading size specification (size of string). Although usually strings are null-terminated it is helpful to know the size in advance (especially during deserialization).
         * The string is serialized without trailing null-terminator.
         * \param [in,out] stream Reference to stream we want to extend by the value.
         * \param [in] value String to serialize.
         * \return Size of the binary represenation of the value (incl. header) (count of bytes added to stream).
         */
        static inline BinaryStreamSize serializeString(BinaryStream& stream, const std::string& value)
        {
            // Allocate memory for maximum performance (does nothing if capacity of stream is already big enough)
            stream.reserve(stream.size() + sizeof(BinaryStreamSize) + value.size());

            // Encode size of string as header (this allows easier deserialization afterwards if size is known in advance)
            (void)serialize(stream, (BinaryStreamSize)value.size());

            // Encode string as "payload" and add to the end of the stream
            std::copy(value.begin(), value.end(), std::back_inserter(stream));

            // Return count of added bytes
            return sizeof(BinaryStreamSize) + value.size();
        }

        // Deserialization
        // ---------------
        //! Decodes the binary representation of a boolean
        /*!
         * \param [in] stream The stream from which the value should be extracted.
         * \param [in] index The starting index of the value in the stream.
         * \param [out] value Reference to returned value. The result of the deserialization will be stored in this parameter.
         * \return Size of binary representation of the value in case of success or 0 on failure
         */
        static inline BinaryStreamSize deSerialize(const BinaryStream& stream, const BinaryStreamSize& index, bool& value)
        {
            BinaryStreamSize targetSize = sizeof(uint8_t);

            // Check if stream is big enough
            if (stream.size() < index + targetSize)
                return 0;

            // Extract value
            value = (bool)stream[index];

            // Return size of binary representation within stream
            return targetSize;
        }

        //! Decodes the binary representation of a 8-bit unsigned integer
        /*!
         * \param [in] stream The stream from which the value should be extracted.
         * \param [in] index The starting index of the value in the stream.
         * \param [out] value Reference to returned value. The result of the deserialization will be stored in this parameter.
         * \return Size of binary representation of the value in case of success or 0 on failure
         */
        static inline BinaryStreamSize deSerialize(const BinaryStream& stream, const BinaryStreamSize& index, uint8_t& value)
        {
            BinaryStreamSize targetSize = sizeof(uint8_t);

            // Check if stream is big enough
            if (stream.size() < index + targetSize)
                return 0;

            // Extract value
            value = stream[index];

            // Return size of binary representation within stream
            return targetSize;
        }

        //! Decodes the binary representation of a 16-bit unsigned integer
        /*!
         * \param [in] stream The stream from which the value should be extracted.
         * \param [in] index The starting index of the value in the stream.
         * \param [out] value Reference to returned value. The result of the deserialization will be stored in this parameter.
         * \return Size of binary representation of the value in case of success or 0 on failure
         */
        static inline BinaryStreamSize deSerialize(const BinaryStream& stream, const BinaryStreamSize& index, uint16_t& value)
        {
            BinaryStreamSize targetSize = sizeof(uint16_t);

            // Check if stream is big enough
            if (stream.size() < index + targetSize)
                return 0;

            // Extract value
            value = (((uint16_t)stream[index]) << 8) + (uint16_t)stream[index + 1];

            // Return size of binary representation within stream
            return targetSize;
        }

        //! Decodes the binary representation of a 32-bit unsigned integer
        /*!
         * \param [in] stream The stream from which the value should be extracted.
         * \param [in] index The starting index of the value in the stream.
         * \param [out] value Reference to returned value. The result of the deserialization will be stored in this parameter.
         * \return Size of binary representation of the value in case of success or 0 on failure
         */
        static inline BinaryStreamSize deSerialize(const BinaryStream& stream, const BinaryStreamSize& index, uint32_t& value)
        {
            BinaryStreamSize targetSize = sizeof(uint32_t);

            // Check if stream is big enough
            if (stream.size() < index + targetSize)
                return 0;

            // Extract value
            value = (((uint32_t)stream[index]) << 24) + (((uint32_t)stream[index + 1]) << 16) + (((uint32_t)stream[index + 2]) << 8) + (uint32_t)stream[index + 3];

            // Return size of binary representation within stream
            return targetSize;
        }

        //! Decodes the binary representation of a 64-bit unsigned integer
        /*!
         * \param [in] stream The stream from which the value should be extracted.
         * \param [in] index The starting index of the value in the stream.
         * \param [out] value Reference to returned value. The result of the deserialization will be stored in this parameter.
         * \return Size of binary representation of the value in case of success or 0 on failure
         */
        static inline BinaryStreamSize deSerialize(const BinaryStream& stream, const BinaryStreamSize& index, uint64_t& value)
        {
            BinaryStreamSize targetSize = sizeof(uint64_t);

            // Check if stream is big enough
            if (stream.size() < index + targetSize)
                return 0;

            // Extract value
            value = (((uint64_t)stream[index]) << 56) + (((uint64_t)stream[index + 1]) << 48) + (((uint64_t)stream[index + 2]) << 40) + (((uint64_t)stream[index + 3]) << 32) + (((uint64_t)stream[index + 4]) << 24) + (((uint64_t)stream[index + 5]) << 16) + (((uint64_t)stream[index + 6]) << 8) + (uint64_t)stream[index + 7];

            // Return size of binary representation within stream
            return targetSize;
        }

        //! Decodes the binary representation of a 8-bit signed integer
        /*!
         * \remark Bit-shift of negative integers is undefined, thus we have to handle the sign bit on our own!
         *
         * \param [in] stream The stream from which the value should be extracted.
         * \param [in] index The starting index of the value in the stream.
         * \param [out] value Reference to returned value. The result of the deserialization will be stored in this parameter.
         * \return Size of binary representation of the value in case of success or 0 on failure
         */
        static inline BinaryStreamSize deSerialize(const BinaryStream& stream, const BinaryStreamSize& index, int8_t& value)
        {
            // Get unsigned representation
            uint8_t unsignedValue = 0;
            if (deSerialize(stream, index, unsignedValue) == 0) // Fails if stream is not big enough for value
                return 0;

            // Check, if sign bit is set
            if ((unsignedValue >> 7) == 0) {
                // ...positive value
                value = (int8_t)unsignedValue;
            } else {
                // ...negative value
                value = -(int8_t)(unsignedValue & (~(((uint8_t)1) << 7))); // First remove sign bit, then convert to signed value, then negate
            }

            // Return size of binary representation within stream
            return sizeof(value);
        }

        //! Decodes the binary representation of a 16-bit signed integer
        /*!
         * \remark Bit-shift of negative integers is undefined, thus we have to handle the sign bit on our own!
         *
         * \param [in] stream The stream from which the value should be extracted.
         * \param [in] index The starting index of the value in the stream.
         * \param [out] value Reference to returned value. The result of the deserialization will be stored in this parameter.
         * \return Size of binary representation of the value in case of success or 0 on failure
         */
        static inline BinaryStreamSize deSerialize(const BinaryStream& stream, const BinaryStreamSize& index, int16_t& value)
        {
            // Get unsigned representation
            uint16_t unsignedValue = 0;
            if (deSerialize(stream, index, unsignedValue) == 0) // Fails if stream is not big enough for value
                return 0;

            // Check, if sign bit is set
            if ((unsignedValue >> 15) == 0) {
                // ...positive value
                value = (int16_t)unsignedValue;
            } else {
                // ...negative value
                value = -(int16_t)(unsignedValue & (~(((uint16_t)1) << 15))); // First remove sign bit, then convert to signed value, then negate
            }

            // Return size of binary representation within stream
            return sizeof(value);
        }

        //! Decodes the binary representation of a 32-bit signed integer
        /*!
         * \remark Bit-shift of negative integers is undefined, thus we have to handle the sign bit on our own!
         *
         * \param [in] stream The stream from which the value should be extracted.
         * \param [in] index The starting index of the value in the stream.
         * \param [out] value Reference to returned value. The result of the deserialization will be stored in this parameter.
         * \return Size of binary representation of the value in case of success or 0 on failure
         */
        static inline BinaryStreamSize deSerialize(const BinaryStream& stream, const BinaryStreamSize& index, int32_t& value)
        {
            // Get unsigned representation
            uint32_t unsignedValue = 0;
            if (deSerialize(stream, index, unsignedValue) == 0) // Fails if stream is not big enough for value
                return 0;

            // Check, if sign bit is set
            if ((unsignedValue >> 31) == 0) {
                // ...positive value
                value = (int32_t)unsignedValue;
            } else {
                // ...negative value
                value = -(int32_t)(unsignedValue & (~(((uint32_t)1) << 31))); // First remove sign bit, then convert to signed value, then negate
            }

            // Return size of binary representation within stream
            return sizeof(value);
        }

        //! Decodes the binary representation of a 64-bit signed integer
        /*!
         * \remark Bit-shift of negative integers is undefined, thus we have to handle the sign bit on our own!
         *
         * \param [in] stream The stream from which the value should be extracted.
         * \param [in] index The starting index of the value in the stream.
         * \param [out] value Reference to returned value. The result of the deserialization will be stored in this parameter.
         * \return Size of binary representation of the value in case of success or 0 on failure
         */
        static inline BinaryStreamSize deSerialize(const BinaryStream& stream, const BinaryStreamSize& index, int64_t& value)
        {
            // Get unsigned representation
            uint64_t unsignedValue = 0;
            if (deSerialize(stream, index, unsignedValue) == 0) // Fails if stream is not big enough for value
                return 0;

            // Check, if sign bit is set
            if ((unsignedValue >> 63) == 0) {
                // ...positive value
                value = (int64_t)unsignedValue;
            } else {
                // ...negative value
                value = -(int64_t)(unsignedValue & (~(((uint64_t)1) << 63))); // First remove sign bit, then convert to signed value, then negate
            }

            // Return size of binary representation within stream
            return sizeof(value);
        }

        //! Decodes the binary representation of a 32-bit floating point number
        /*!
         * \remark Bit-shift of negative integers is undefined, thus we have to handle the sign bit on our own!
         *
         * \param [in] stream The stream from which the value should be extracted.
         * \param [in] index The starting index of the value in the stream.
         * \param [out] value Reference to returned value. The result of the deserialization will be stored in this parameter.
         * \return Size of binary representation of the value in case of success or 0 on failure
         */
        static inline BinaryStreamSize deSerialize(const BinaryStream& stream, const BinaryStreamSize& index, float& value)
        {
            // Get integer representation
            uint32_t intRepresentation = 0;
            if (deSerialize(stream, index, intRepresentation) == 0) // Fails if stream is not big enough for value
                return 0;

            // Convert to 32-bit float
            memcpy((void*)&value, (void*)&intRepresentation, sizeof(value));

            // Return size of binary representation within stream
            return sizeof(value);
        }

        //! Decodes the binary representation of a 64-bit floating point number
        /*!
         * \remark Bit-shift of negative integers is undefined, thus we have to handle the sign bit on our own!
         *
         * \param [in] stream The stream from which the value should be extracted.
         * \param [in] index The starting index of the value in the stream.
         * \param [out] value Reference to returned value. The result of the deserialization will be stored in this parameter.
         * \return Size of binary representation of the value in case of success or 0 on failure
         */
        static inline BinaryStreamSize deSerialize(const BinaryStream& stream, const BinaryStreamSize& index, double& value)
        {
            // Get integer representation
            uint64_t intRepresentation = 0;
            if (deSerialize(stream, index, intRepresentation) == 0) // Fails if stream is not big enough for value
                return 0;

            // Convert to 64-bit float (=double)
            memcpy((void*)&value, (void*)&intRepresentation, sizeof(value));

            // Return size of binary representation within stream
            return sizeof(value);
        }

        //! Decodes the binary representation of a \ref broccoli::core::Time instance
        /*!
         * \param [in] stream The stream from which the value should be extracted.
         * \param [in] index The starting index of the value in the stream.
         * \param [out] value Reference to returned value. The result of the deserialization will be stored in this parameter.
         * \return Size of binary representation of the value in case of success or 0 on failure
         */
        static inline BinaryStreamSize deSerialize(const BinaryStream& stream, const BinaryStreamSize& index, core::Time& value)
        {
            BinaryStreamSize targetSize = sizeof(value.m_seconds) + sizeof(value.m_nanoSeconds);

            // Check if stream is big enough
            if (stream.size() < index + targetSize)
                return 0;

            // Deserialize seconds
            if (deSerialize(stream, index, value.m_seconds) == 0)
                return 0; // Should never happen, since size is checked in advance

            // Deserialize nanoseconds
            if (deSerialize(stream, index + sizeof(value.m_seconds), value.m_nanoSeconds) == 0)
                return 0; // Should never happen, since size is checked in advance

            // Return size of binary representation within stream
            return targetSize;
        }

        //! Decodes the binary representation of an array of basic datatypes (type \p T, size \p N)
        /*!
         * \param [in] stream The binary stream from which the array should be extracted.
         * \param [in] index The starting index of the array in the binary stream.
         * \param [out] array Reference to returned array. The result of the deserialization will be stored in this parameter.
         * \return Size of binary representation of the array in case of success or 0 on failure
         */
        template <typename T, size_t N>
        static inline BinaryStreamSize deSerializeArray(const BinaryStream& stream, const BinaryStreamSize& index, std::array<T, N>& array)
        {
            BinaryStreamSize targetSize = sizeof(T) * N;

            // Check if stream is big enough
            if (stream.size() < index + targetSize)
                return 0;

            // Iterate through all elements and deserialize one after the other
            for (size_t i = 0; i < N; i++) {
                T newElement;
                if (deSerialize(stream, index + i * sizeof(T), newElement) == 0)
                    return 0; // Should never happen, since size is checked in advance
                else
                    array[i] = newElement;
            }

            // Return size of binary representation within stream
            return targetSize;
        }

        //! Decodes the binary representation of a string
        /*!
         * \remark Reads and checks the size from the header of the stream.
         *
         * \param [in] stream The stream from which the value should be extracted.
         * \param [in] index The starting index of the value in the stream.
         * \param [out] value Reference to returned value. The result of the deserialization will be stored in this parameter.
         * \return Size of binary representation of the value (header + string) in case of success or 0 on failure
         */
        static inline BinaryStreamSize deSerializeString(const BinaryStream& stream, const BinaryStreamSize& index, std::string& value)
        {
            // Extract header
            BinaryStreamSize stringLength = 0;
            if (deSerialize(stream, index, stringLength) == 0) // Fails if stream is not big enough for header
                return 0;

            // Check, if stream is long enough for header and "payload" (=string)
            BinaryStreamSize targetSize = sizeof(BinaryStreamSize) + stringLength;
            if (stream.size() < index + targetSize)
                return 0;

            // Extract string from stream
            value = std::string(stream.begin() + index + sizeof(BinaryStreamSize), stream.begin() + index + targetSize);

            // Return size of binary representation within stream
            return targetSize;
        }
    } // namespace serialization
} // namespace io
} // namespace broccoli
