/*
 * This file is part of broccoli.
 * Copyright (C) 2019 Chair of Applied Mechanics, Technical University of Munich
 * https://www.amm.mw.tum.de/
 */

#include "broccoli/io/encoding.hpp"
#include "gtest/gtest.h"

using namespace broccoli;
using namespace io;

//! Check encoding and decoding of basic datatypes
TEST(encoding, BasicDataTypes)
{
    // Initialize helpers
    encoding::CharacterStream stream;
    std::vector<encoding::CharacterStreamSize> partSizes; // Sizes of added parts
    char testChar = 'A';
    uint64_t testUint64_t = 123;
    int64_t testInt64_t = -456;
    double testDouble = -1.23456;
    std::string testString = "test";

    // Test encoding
    // -------------
    // Create stream of datatypes
    partSizes.push_back(encoding::encode(stream, testChar)); // Test character
    partSizes.push_back(encoding::encode(stream, testUint64_t)); // Test uint64_t
    partSizes.push_back(encoding::encode(stream, testInt64_t)); // Test int64_t
    partSizes.push_back(encoding::encode(stream, testDouble, "%+.4f")); // Test double
    partSizes.push_back(encoding::encode(stream, testString)); // Test string

    // Check single sizes
    ASSERT_EQ(partSizes[0], 1);
    ASSERT_EQ(partSizes[1], 3);
    ASSERT_EQ(partSizes[2], 4);
    ASSERT_EQ(partSizes[3], 7);
    ASSERT_EQ(partSizes[4], 4);

    // Check total size
    encoding::CharacterStreamSize totalSize = 0;
    for (size_t i = 0; i < partSizes.size(); i++)
        totalSize += partSizes[i];
    ASSERT_EQ(totalSize, stream.size());

    // Test decoding
    // -------------
    encoding::CharacterStreamSize currentIndex = 0;

    // Decode char
    char decodedChar = '0';
    currentIndex = 0;
    ASSERT_EQ(encoding::decode(stream, currentIndex, decodedChar), partSizes[0]);
    ASSERT_EQ(decodedChar, testChar);
    currentIndex += partSizes[0];

    // Decode uint64_t
    uint64_t decodedUint64_t = 0;
    ASSERT_EQ(encoding::decode(stream, currentIndex, partSizes[1], decodedUint64_t), partSizes[1]);
    ASSERT_EQ(decodedUint64_t, testUint64_t);
    currentIndex += partSizes[1];

    // Decode int64_t
    int64_t decodedInt64_t = 0;
    ASSERT_EQ(encoding::decode(stream, currentIndex, partSizes[2], decodedInt64_t), partSizes[2]);
    ASSERT_EQ(decodedInt64_t, testInt64_t);
    currentIndex += partSizes[2];

    // Decode double
    double decodedDouble = 0;
    ASSERT_EQ(encoding::decode(stream, currentIndex, partSizes[3], decodedDouble), partSizes[3]);
    ASSERT_EQ(decodedDouble, -1.2346);
    currentIndex += partSizes[3];

    // Decode string
    std::string decodedString = "";
    ASSERT_EQ(encoding::decode(stream, currentIndex, partSizes[4], decodedString), partSizes[4]);
    ASSERT_EQ(decodedString, testString);
    currentIndex += partSizes[4];

    // Test index length
    ASSERT_EQ(currentIndex, stream.size());

    // Test invalid index
    ASSERT_EQ(encoding::decode(stream, stream.size(), decodedChar), 0);

    // Test invalid lengths
    ASSERT_EQ(encoding::decode(stream, 0, 0, decodedDouble), 0);
    ASSERT_EQ(encoding::decode(stream, 0, stream.size() + 1, decodedDouble), 0);
}
