/*
 * This file is part of broccoli.
 * Copyright (C) 2019 Chair of Applied Mechanics, Technical University of Munich
 * https://www.amm.mw.tum.de/
 */

#include "broccoli/io/serialization.hpp"
#include "gtest/gtest.h"

using namespace broccoli;
using namespace io;

//! Template for serialization and deserialization of various datatypes
template <class T>
void TestBasicDatatype(const T& value)
{
    // Initialize helpers
    serialization::BinaryStream stream;
    serialization::BinaryStreamSize inputStreamSize, outputStreamSize;
    T outputValue = T();

    // Serialize
    inputStreamSize = serialization::serialize(stream, value);

    // Deserialize
    outputStreamSize = serialization::deSerialize(stream, 0, outputValue);

    // Check value
    ASSERT_TRUE(outputValue == value);

    // Check stream sizes
    ASSERT_TRUE(outputStreamSize == inputStreamSize);
    ASSERT_TRUE(outputStreamSize == stream.size());
}

//! Check serialization and deserialization of basic datatypes
TEST(serialization, BasicDatatypes)
{
    // Booleans
    TestBasicDatatype<bool>(false);
    TestBasicDatatype<bool>(true);

    // Unsigned integers
    TestBasicDatatype<uint8_t>(0);
    TestBasicDatatype<uint8_t>(120);
    TestBasicDatatype<uint8_t>(255);
    TestBasicDatatype<uint16_t>(0);
    TestBasicDatatype<uint16_t>(30000);
    TestBasicDatatype<uint16_t>(65535);
    TestBasicDatatype<uint32_t>(0UL);
    TestBasicDatatype<uint32_t>(1111111111UL);
    TestBasicDatatype<uint32_t>(4294967295UL);
    TestBasicDatatype<uint64_t>(0ULL);
    TestBasicDatatype<uint64_t>(123456789123457869ULL);
    TestBasicDatatype<uint64_t>(18446744073709551614ULL);

    // Signed integers
    TestBasicDatatype<int8_t>(-127);
    TestBasicDatatype<int8_t>(-60);
    TestBasicDatatype<int8_t>(0);
    TestBasicDatatype<int8_t>(60);
    TestBasicDatatype<int8_t>(127);
    TestBasicDatatype<int16_t>(-32767);
    TestBasicDatatype<int16_t>(-3000);
    TestBasicDatatype<int16_t>(0);
    TestBasicDatatype<int16_t>(3000);
    TestBasicDatatype<int16_t>(32767);
    TestBasicDatatype<int32_t>(-2147483647L);
    TestBasicDatatype<int32_t>(-1111111L);
    TestBasicDatatype<int32_t>(0L);
    TestBasicDatatype<int32_t>(1111111L);
    TestBasicDatatype<int32_t>(2147483647L);
    TestBasicDatatype<int64_t>(-9223372036854775807LL);
    TestBasicDatatype<int64_t>(-123456789457869LL);
    TestBasicDatatype<int64_t>(0LL);
    TestBasicDatatype<int64_t>(123456789457869LL);
    TestBasicDatatype<int64_t>(9223372036854775807LL);

    // Floating point numbers
    TestBasicDatatype<float>(-12.345);
    TestBasicDatatype<float>(0);
    TestBasicDatatype<float>(12.345);
    TestBasicDatatype<double>(-12.3456789123);
    TestBasicDatatype<double>(0);
    TestBasicDatatype<double>(12.3456789123);
}

//! Check serialization and deserialization of times
TEST(serialization, Time)
{
    // Initialize helpers
    serialization::BinaryStream stream;
    serialization::BinaryStreamSize inputStreamSize, outputStreamSize;
    core::Time inputValue(1, -23456789);
    core::Time outputValue(0, 0);

    // Serialize
    inputStreamSize = serialization::serialize(stream, inputValue);

    // Deserialize
    outputStreamSize = serialization::deSerialize(stream, 0, outputValue);

    // Check value
    ASSERT_TRUE(outputValue == inputValue);

    // Check stream sizes
    ASSERT_TRUE(outputStreamSize == inputStreamSize);
    ASSERT_TRUE(outputStreamSize == stream.size());
}

//! Check serialization and deserialization of arrays
TEST(serialization, Arrays)
{
    // Initialize helpers
    serialization::BinaryStream stream;
    serialization::BinaryStreamSize inputStreamSize, outputStreamSize;
    std::array<double, 3> inputValue({ { 0.123, -0.123, 12.34 } });
    std::array<double, 3> outputValue({ { 0, 0, 0 } });

    // Serialize
    inputStreamSize = serialization::serializeArray<double, 3>(stream, inputValue);

    // Deserialize
    outputStreamSize = serialization::deSerializeArray<double, 3>(stream, 0, outputValue);

    // Check value
    ASSERT_TRUE(outputValue == inputValue);

    // Check stream sizes
    ASSERT_TRUE(outputStreamSize == inputStreamSize);
    ASSERT_TRUE(outputStreamSize == stream.size());
}

//! Check serialization and deserialization of strings
TEST(serialization, Strings)
{
    // Initialize helpers
    serialization::BinaryStream stream;
    serialization::BinaryStreamSize inputStreamSize, outputStreamSize;
    std::string inputValue = "asdf";
    std::string outputValue = "";

    // Serialize
    inputStreamSize = serialization::serializeString(stream, inputValue);

    // Deserialize
    outputStreamSize = serialization::deSerializeString(stream, 0, outputValue);

    // Check value
    ASSERT_TRUE(outputValue == inputValue);

    // Check stream sizes
    ASSERT_TRUE(outputStreamSize == inputStreamSize);
    ASSERT_TRUE(outputStreamSize == stream.size());
}

//! Check concatenation of binary streams
TEST(serialization, StreamConcatenation)
{
    // Initialize helpers
    serialization::BinaryStream stream({ 'a', 'b', 'c' });
    serialization::BinaryStream streamToAppend({ 'd', 'e', 'f' });

    // Concatenate
    serialization::BinaryStreamSize addedElements = serialization::serialize(stream, streamToAppend);

    // Check value
    ASSERT_TRUE(stream == serialization::BinaryStream({ 'a', 'b', 'c', 'd', 'e', 'f' }));

    // Check stream sizes
    ASSERT_TRUE(addedElements == streamToAppend.size());
}
