/*
 * This file is part of broccoli.
 * Copyright (C) 2019 Chair of Applied Mechanics, Technical University of Munich
 * https://www.amm.mw.tum.de/
 */

#include "broccoli/parallel/BackgroundWorker.hpp"
#include "gtest/gtest.h"
#include <string>

using namespace broccoli;
using namespace parallel;

//! Derived background worker class for basic tests
class BasicWorker : public BackgroundWorker {
public:
    BasicWorker(const std::string& name, const bool& multiThreaded, const int& threadPriority, const core::Time& minimumCycleTime, const bool& blockingWaiting)
        : BackgroundWorker(name, multiThreaded, threadPriority, minimumCycleTime, blockingWaiting)
        , m_testCounter(0)
    {
    }

private:
    //! Copy constructor (internal) (**not** thread-safe -> should only be called by the thread-safe wrapper)
    /*! \param [in] original Reference to original object. */
    BasicWorker(const BasicWorker& original, const int& /* <- trick used for locking mutex */)
        : BackgroundWorker(original)
        , m_testCounter(original.m_testCounter)
    {
    }

public:
    //! Copy constructor (wrapper) (**thread-safe**)
    /*!
     * Locks the mutex of the original object for reading before calling the internal (protected) constructor. Unlocks original object afterwards.
     * \param [in] original Reference to original object.
     */
    BasicWorker(const BasicWorker& original)
        : BasicWorker(original, pthread_rwlock_rdlock(&original.m_mutex) /* <- lock mutex of original object first (for reading since also using "const") */)
    {
        pthread_rwlock_unlock(&original.m_mutex); // Unlock mutex of original object after internal (protected) constructor has been called
    }

    //! Copy assignment operator (**thread-safe**)
    /*!
     * Uses own mutex and mutex of the reference object to guarantee thread-safe copying of members.
     * \param [in] reference Reference to reference object.
     * \return Pointer to this instance.
     */
    BasicWorker& operator=(const BasicWorker& reference)
    {
        // Avoid self-assignment
        if (this == &reference)
            return *this;

        // Try to lock ourselves and reference (while avoiding deadlocks)
        while (true) { // spinning
            pthread_rwlock_wrlock(&m_mutex); // Lock ourselves for writing (blocking)
            if (pthread_rwlock_tryrdlock(&reference.m_mutex) == 0) // Try to lock reference for reading
                break; // ...success -> stop spinning
            else
                pthread_rwlock_unlock(&m_mutex); // ...fail -> unlock ourselves to allow other threads to access us and thus resolve possible deadlocks
        }

        // Copy data
        BackgroundWorker::operator=(reference);
        m_testCounter = reference.m_testCounter;

        // Unlock reference object and ourselves
        pthread_rwlock_unlock(&reference.m_mutex);
        pthread_rwlock_unlock(&m_mutex);

        return *this;
    }

    virtual void execute()
    {
        // Increment counter by 1
        int currentCounter = testCounter();
        currentCounter++;
        setTestCounter(currentCounter);

        // Output info to console
        std::cout << "execute() of worker '" << name() << "' triggered! Counter: " << currentCounter << std::endl;

        // Stop after a certain count of executions
        if (currentCounter >= 10)
            stop();
    }

    // Protected worker data
    int m_testCounter;

    // Getters
    int testCounter() const { return getProtectedData<int>(&m_testCounter); }

    // Setters
    void setTestCounter(const int& newValue) { setProtectedData<int>(newValue, &m_testCounter); }
};

//! Derived background worker class for test of abortion
class AbortWorker : public BackgroundWorker {
public:
    AbortWorker(const std::string& name, const bool& multiThreaded, const int& threadPriority, const core::Time& minimumCycleTime, const bool& blockingWaiting)
        : BackgroundWorker(name, multiThreaded, threadPriority, minimumCycleTime, blockingWaiting)
        , m_enableEmergencyExit(false)
    {
    }

private:
    //! Copy constructor (internal) (**not** thread-safe -> should only be called by the thread-safe wrapper)
    /*! \param [in] original Reference to original object. */
    AbortWorker(const AbortWorker& original, const int& /* <- trick used for locking mutex */)
        : BackgroundWorker(original)
        , m_enableEmergencyExit(original.m_enableEmergencyExit)
    {
    }

public:
    //! Copy constructor (wrapper) (**thread-safe**)
    /*!
     * Locks the mutex of the original object for reading before calling the internal (protected) constructor. Unlocks original object afterwards.
     * \param [in] original Reference to original object.
     */
    AbortWorker(const AbortWorker& original)
        : AbortWorker(original, pthread_rwlock_rdlock(&original.m_mutex) /* <- lock mutex of original object first (for reading since also using "const") */)
    {
        pthread_rwlock_unlock(&original.m_mutex); // Unlock mutex of original object after internal (protected) constructor has been called
    }

    //! Copy assignment operator (**thread-safe**)
    /*!
     * Uses own mutex and mutex of the reference object to guarantee thread-safe copying of members.
     * \param [in] reference Reference to reference object.
     * \return Pointer to this instance.
     */
    AbortWorker& operator=(const AbortWorker& reference)
    {
        // Avoid self-assignment
        if (this == &reference)
            return *this;

        // Try to lock ourselves and reference (while avoiding deadlocks)
        while (true) { // spinning
            pthread_rwlock_wrlock(&m_mutex); // Lock ourselves for writing (blocking)
            if (pthread_rwlock_tryrdlock(&reference.m_mutex) == 0) // Try to lock reference for reading
                break; // ...success -> stop spinning
            else
                pthread_rwlock_unlock(&m_mutex); // ...fail -> unlock ourselves to allow other threads to access us and thus resolve possible deadlocks
        }

        // Copy data
        BackgroundWorker::operator=(reference);
        m_enableEmergencyExit = reference.m_enableEmergencyExit;

        // Unlock reference object and ourselves
        pthread_rwlock_unlock(&reference.m_mutex);
        pthread_rwlock_unlock(&m_mutex);

        return *this;
    }

    virtual void execute()
    {
        // Simulate heavy load
        for (uint64_t i = 0; i < 10000; i++) {
            // Emergency exit (if enabled)
            if (enableEmergencyExit() == true && stopTriggered() == true)
                break;
            // else: do nothing
        }
    }

    // Protected worker data
    bool m_enableEmergencyExit;

    // Getters
    int enableEmergencyExit() const { return getProtectedData<bool>(&m_enableEmergencyExit); }

    // Setters
    void setEnableEmergencyExit(const bool& newValue) { setProtectedData<bool>(newValue, &m_enableEmergencyExit); }
};

//! Test setters and getters
TEST(BackgroundWorker, SettersAndGetters)
{
    // Create instance
    BasicWorker worker("Worker", false, 0, core::Time(0.1), false);

    // Setters and getters
    // ...no public setters

    // Getters only
    std::string name = worker.name();
    ASSERT_EQ(name, "Worker");
    bool multiThreaded = worker.multiThreaded();
    ASSERT_EQ(multiThreaded, false);
    int threadPriority = worker.threadPriority();
    ASSERT_EQ(threadPriority, 0);
    core::Time minimumCycleTime = worker.minimumCycleTime();
    ASSERT_EQ(minimumCycleTime, core::Time(0.1));
    bool blockingWaiting = worker.blockingWaiting();
    ASSERT_EQ(blockingWaiting, false);
    bool threadRunning = worker.threadRunning();
    ASSERT_EQ(threadRunning, false);
    bool startTriggered = worker.startTriggered();
    ASSERT_EQ(startTriggered, false);
    bool stopTriggered = worker.stopTriggered();
    ASSERT_EQ(stopTriggered, false);
    bool joinTriggered = worker.joinTriggered();
    ASSERT_EQ(joinTriggered, false);
    bool pauseActive = worker.pauseActive();
    ASSERT_EQ(pauseActive, false);
    core::Time lastCycleTime = worker.lastCycleTime();
    (void)lastCycleTime;
    core::Time lastExecutionCallTime = worker.lastExecutionCallTime();
    (void)lastExecutionCallTime;
    core::Time totalExecutionRunTime = worker.totalExecutionRunTime();
    (void)totalExecutionRunTime;
    unsigned long long executionCalls = worker.executionCalls();
    ASSERT_EQ(executionCalls, 0);
}

//! Check creation, execution and stopping of background workers in single-threaded mode
TEST(BackgroundWorker, SingleThreaded)
{
    // Create workers
    BasicWorker workerA("WorkerASingleThreaded", false, 0, core::Time(0.1), false);
    BasicWorker workerB("WorkerBSingleThreaded", false, 0, core::Time(0.1), false);

    // Copy constructor
    BasicWorker workerCopy(workerA);
    ASSERT_TRUE(workerCopy.name() == workerA.name());

    // Assignment operator
    workerCopy = workerB;
    ASSERT_TRUE(workerCopy.name() == workerB.name());

    // Check execution counter
    ASSERT_EQ(workerA.executionCalls(), 0);
    ASSERT_EQ(workerB.executionCalls(), 0);

    // Check member access
    ASSERT_EQ(workerA.testCounter(), 0);
    ASSERT_EQ(workerB.testCounter(), 0);

    // Start workers
    ASSERT_EQ(workerA.start(), true);
    ASSERT_EQ(workerB.start(), true);
    core::Time runTimeA0 = workerA.lastExecutionRunTime();
    core::Time runTimeB0 = workerB.lastExecutionRunTime();

    // Check execution counter
    ASSERT_EQ(workerA.executionCalls(), 1);
    ASSERT_EQ(workerB.executionCalls(), 1);

    // Check member access
    ASSERT_EQ(workerA.testCounter(), 1);
    ASSERT_EQ(workerB.testCounter(), 1);

    // Trigger execution and get runtimes
    workerA.triggerManualExecution();
    workerB.triggerManualExecution();
    core::Time runTimeA1 = workerA.lastExecutionRunTime();
    core::Time runTimeB1 = workerB.lastExecutionRunTime();

    // Check pause
    workerA.pause();
    workerB.pause();
    workerA.triggerManualExecution();
    workerB.triggerManualExecution();
    ASSERT_EQ(workerA.executionCalls(), 2);
    ASSERT_EQ(workerB.executionCalls(), 2);

    // Check resume
    workerA.resume();
    workerB.resume();
    workerA.triggerManualExecution();
    workerB.triggerManualExecution();
    core::Time runTimeA2 = workerA.lastExecutionRunTime();
    core::Time runTimeB2 = workerB.lastExecutionRunTime();

    // Check execution counter
    ASSERT_EQ(workerA.executionCalls(), 3);
    ASSERT_EQ(workerB.executionCalls(), 3);

    // Check total runtime
    core::Time runTimeAT = workerA.totalExecutionRunTime();
    core::Time runTimeBT = workerB.totalExecutionRunTime();
    ASSERT_EQ(runTimeAT, runTimeA0 + runTimeA1 + runTimeA2);
    ASSERT_EQ(runTimeBT, runTimeB0 + runTimeB1 + runTimeB2);

    // Check member access
    ASSERT_EQ(workerA.testCounter(), 3);
    ASSERT_EQ(workerB.testCounter(), 3);

    // Check stopping
    ASSERT_EQ(workerA.stop(), true);
    ASSERT_EQ(workerB.stop(), true);

    // Trigger execution
    workerA.triggerManualExecution();
    workerB.triggerManualExecution();

    // Check execution counter
    ASSERT_EQ(workerA.executionCalls(), 3);
    ASSERT_EQ(workerB.executionCalls(), 3);

    // Check joining
    ASSERT_EQ(workerA.join(core::Time(0)), true);
    ASSERT_EQ(workerB.join(core::Time(0)), true);

    // Output runtimes
    std::cout << "Total runtime of worker '" << workerA.name() << "' execute(): " << workerA.totalExecutionRunTime().encodeToDurationString(9) << "s" << std::endl;
    std::cout << "Total runtime of worker '" << workerB.name() << "' execute(): " << workerB.totalExecutionRunTime().encodeToDurationString(9) << "s" << std::endl;
}

//! Check creation, execution and stopping of background workers in multi-threaded mode
TEST(BackgroundWorker, MultiThreadedBasic)
{
    // Create workers
    BasicWorker workerA("WorkerAMultiThreaded", true, 0, core::Time(0.010), false);
    BasicWorker workerB("WorkerBMultiThreaded", true, 0, core::Time(0.025), false);

    // Check execution counter
    ASSERT_EQ(workerA.executionCalls(), 0);
    ASSERT_EQ(workerB.executionCalls(), 0);

    // Check member access
    ASSERT_EQ(workerA.testCounter(), 0);
    ASSERT_EQ(workerB.testCounter(), 0);
    workerB.setTestCounter(8); // Start already at 8

    // Start workers
    core::Time startTime = core::Time::currentTime();
    ASSERT_EQ(workerA.start(), true);
    core::Time::sleep(0.001); // Wait some time to simulate different starting times
    ASSERT_EQ(workerB.start(), true);

    // Wait until both threads have finished their work (or a timeout has happend)
    bool timeoutHit = false;
    while (workerA.testCounter() < 10 || workerB.testCounter() < 10) {
        // Check for timeout
        if (core::Time::currentTime() - startTime > 3) {
            timeoutHit = true;
            break;
        }
    }

    // Check, if timeout was hit
    ASSERT_EQ(timeoutHit, false);

    // Try to join counters
    ASSERT_EQ(workerA.join(core::Time(1)), true);
    ASSERT_EQ(workerB.join(core::Time(1)), true);

    // Check execution counter
    ASSERT_EQ(workerA.executionCalls(), 10);
    ASSERT_EQ(workerB.executionCalls(), 2);

    // Check member access
    ASSERT_EQ(workerA.testCounter(), 10);
    ASSERT_EQ(workerB.testCounter(), 10);

    // Output runtimes
    std::cout << "Total runtime of worker '" << workerA.name() << "' execute(): " << workerA.totalExecutionRunTime().encodeToDurationString(9) << "s" << std::endl;
    std::cout << "Total runtime of worker '" << workerB.name() << "' execute(): " << workerB.totalExecutionRunTime().encodeToDurationString(9) << "s" << std::endl;
}

//! Check aborting worker
TEST(BackgroundWorker, MultiThreadedAbort)
{
    // Create workers
    AbortWorker workerA("WorkerAWithEmergencyExit", true, 0, core::Time(0), true);
    AbortWorker workerB("WorkerBWithoutEmergencyExit", true, 0, core::Time(0), true);

    // Check aborting with usual emergency exit (should be implemented by user)
    workerA.setEnableEmergencyExit(true);
    ASSERT_EQ(workerA.start(), true);
    core::Time::sleep(0.01); // Wait some time and then try to stop
    ASSERT_EQ(workerA.join(core::Time(1)), true);

    // Check aborting without usual emergency exit (however with reasonable execution time)
    workerB.setEnableEmergencyExit(false);
    ASSERT_EQ(workerB.start(), true);
    core::Time::sleep(0.01); // Wait some time and then try to stop
    ASSERT_EQ(workerB.join(core::Time(1)), true);

    // Output calls of execute()
    std::cout << "Total calls of worker '" << workerA.name() << "' execute(): " << workerA.executionCalls() << std::endl;
    std::cout << "Total calls of worker '" << workerB.name() << "' execute(): " << workerB.executionCalls() << std::endl;
}
